Package: Dune
Title: Improving replicability in single-cell RNA-Seq cell type
        discovery
Version: 1.23.0
Authors@R: c(person("Hector", "Roux de Bezieux", role = c("aut", "cre"),
                   email = "hector.rouxdebezieux@berkeley.edu",
		   comment = c(ORCID = "0000-0002-1489-8339")),
	    person("Kelly","Street", role = "aut", email = "street.kelly@gmail.com"))
Description: Given a set of clustering labels, Dune merges pairs of clusters to increase mean ARI between labels, improving replicability. 
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.1.1
Imports: BiocParallel, SummarizedExperiment, utils, ggplot2, dplyr,
        tidyr, RColorBrewer, magrittr, gganimate, purrr, aricode
Suggests: knitr, rmarkdown, testthat (>= 2.1.0)
VignetteBuilder: knitr
Depends: R (>= 3.6)
biocViews: Clustering, GeneExpression, RNASeq, Software, SingleCell,
        Transcriptomics, Visualization
git_url: https://git.bioconductor.org/packages/Dune
git_branch: devel
git_last_commit: 9318153
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-10-31
NeedsCompilation: no
Packaged: 2025-10-31 21:36:39 UTC; biocbuild
Author: Hector Roux de Bezieux [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1489-8339>),
  Kelly Street [aut]
Maintainer: Hector Roux de Bezieux <hector.rouxdebezieux@berkeley.edu>

# Generated by roxygen2: do not edit by hand

export(ARIImp)
export(ARIs)
export(ARItrend)
export(ConfusionEvolution)
export(ConfusionPlot)
export(Dune)
export(NMIImp)
export(NMIs)
export(NMItrend)
export(clusterConversion)
export(functionTracking)
export(intermediateMat)
export(plotARIs)
export(plotNMIs)
export(plotPrePost)
export(whenToStop)
exportMethods(Dune)
import(RColorBrewer)
import(dplyr)
import(ggplot2)
import(tidyr)
import(utils)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(RColorBrewer,brewer.pal)
importFrom(SummarizedExperiment,colData)
importFrom(aricode,ARI)
importFrom(aricode,NMI)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,mutate)
importFrom(dplyr,n_distinct)
importFrom(gganimate,transition_states)
importFrom(magrittr,"%>%")
importFrom(purrr,map_df)
importFrom(tidyr,gather)

Dune - A composable build system
================================

Dune is a build system designed for OCaml/Reason projects only. It
focuses on providing the user with a consistent experience and takes
care of most of the low-level details of OCaml compilation. All you
have to do is provide a description of your project and dune will
do the rest.

The scheme it implements is inspired from the one used inside Jane
Street and adapted to the open source world. It has matured over a
long time and is used daily by hundreds of developers, which means
that it is highly tested and productive.

Dune comes with a [manual][manual]. If you want to get started
without reading too much, you can look at the [quick start
guide][quick-start] or watch [this introduction video][video].

The [example][example] directory contains examples of projects using
dune.

[![Travis status][travis-img]][travis] [![AppVeyor status][appveyor-img]][appveyor]

[manual]:         https://dune.readthedocs.io/en/latest/
[quick-start]:    https://dune.readthedocs.io/en/latest/quick-start.html
[example]:        https://github.com/ocaml/dune/tree/master/example
[travis]:         https://travis-ci.org/ocaml/dune
[travis-img]:     https://travis-ci.org/ocaml/dune.svg?branch=master
[appveyor]:       https://ci.appveyor.com/project/diml/dune/branch/master
[appveyor-img]:   https://ci.appveyor.com/api/projects/status/rsxayce22e8f2jkp?svg=true
[merlin]:         https://github.com/ocaml/merlin
[opam]:           https://opam.ocaml.org
[issues]:         https://github.com/ocaml/dune/issues
[topkg-jbuilder]: https://github.com/diml/topkg-jbuilder
[video]:          https://youtu.be/BNZhmMAJarw

Overview
--------

Dune reads project metadata from `dune` files, which are either
static files in a simple S-expression syntax or OCaml scripts. It uses
this information to setup build rules, generate configuration files
for development tools such as [merlin][merlin], handle installation,
etc...

Dune itself is fast, has very low overhead and supports parallel
builds on all platforms. It has no system dependencies: all you need
to build dune and packages using dune is OCaml. You don't need
`make` or `bash` as long as the packages themselves don't use `bash`
explicitly.

Especially, one can install OCaml on Windows with a binary installer
and then use only the Windows Console to build dune and packages
using dune.

Strengths
---------

### Composable

Take n repositories that use dune, arrange them in any way on the
file system and the result is still a single repository that dune
knows how to build at once.

This make simultaneous development on multiple packages trivial.

### Gracefully handles multi-package repositories

Dune knows how to handle repositories containing several
packages. When building via [opam][opam], it is able to correctly use
libraries that were previously installed even if they are already
present in the source tree.

The magic invocation is:

```sh
$ dune build --only-packages <package-name> @install
```

### Building against several configurations at once

Dune is able to build a given source code repository against
several configurations simultaneously. This helps maintaining packages
across several versions of OCaml as you can test them all at once
without hassle.

This feature should make cross-compilation easy, see details in the
[roadmap](ROADMAP.md).

This feature requires [opam][opam].

Requirements
------------

Dune requires OCaml version 4.02.3 or greater.

Installation
------------

The recommended way to install dune is via the [opam package manager][opam]:

```sh
$ opam install dune
```

You can also build it manually with:

```sh
$ make release
$ make install
```

Running simply `make` will build dune using the development
settings.

If you do not have `make`, you can do the following:

```sh
$ ocaml bootstrap.ml
$ ./boot.exe
$ ./_build/default/bin/main_dune.exe install dune
```

Support
-------

If you have questions about dune, you can send an email to
ocaml-core@googlegroups.com or [open a ticket on github][issues].


Migration from jbuilder
-----------------------

Dune was formerly known as jbuilder. Migration from jbuilder to dune is
described in the [manual](http://dune.readthedocs.io/en/latest/migration.html).

Status
------

Dune is now fairly stable and is used by the majority of packages on
opam. The package is still in beta version as we are waiting for the
renaming from Jbuilder to dune before releasing version 1.0.0. Note
that dune will have backward compatibility with Jbuilder, in particular
existing Jbuilder projects will continue to be buildable with
dune.

Dune (Jbuilder) - A composable build system
===========================================

__Jbuilder has been renamed to Dune__. A full renaming of the documentation and
the tool will be done as part of the 1.0 release.

Jbuilder is a build system designed for OCaml/Reason projects only. It
focuses on providing the user with a consistent experience and takes
care of most of the low-level details of OCaml compilation. All you
have to do is provide a description of your project and Jbuilder will
do the rest.

The scheme it implements is inspired from the one used inside Jane
Street and adapted to the open source world. It has matured over a
long time and is used daily by hundreds of developers, which means
that it is highly tested and productive.

Jbuilder comes with a [manual][manual]. If you want to get started
without reading too much, you can look at the [quick start
guide][quick-start] or watch [this introduction video][video].

The [example][example] directory contains examples of projects using
jbuilder.

[![Travis status][travis-img]][travis] [![AppVeyor status][appveyor-img]][appveyor]

[manual]:         https://jbuilder.readthedocs.io/en/latest/
[quick-start]:    https://jbuilder.readthedocs.io/en/latest/quick-start.html
[example]:        https://github.com/ocaml/dune/tree/master/example
[travis]:         https://travis-ci.org/ocaml/dune
[travis-img]:     https://travis-ci.org/ocaml/dune.svg?branch=master
[appveyor]:       https://ci.appveyor.com/project/diml/dune/branch/master
[appveyor-img]:   https://ci.appveyor.com/api/projects/status/rsxayce22e8f2jkp?svg=true
[merlin]:         https://github.com/ocaml/merlin
[opam]:           https://opam.ocaml.org
[jenga]:          https://github.com/janestreet/jenga
[issues]:         https://github.com/ocaml/dune/issues
[topkg-jbuilder]: https://github.com/diml/topkg-jbuilder
[video]:          https://youtu.be/BNZhmMAJarw

Overview
--------

Jbuilder reads project metadata from `jbuild` files, which are either
static files in a simple S-expression syntax or OCaml scripts. It uses
this information to setup build rules, generate configuration files
for development tools such as [merlin][merlin], handle installation,
etc...

Jbuilder itself is fast, has very low overhead and supports parallel
builds on all platforms. It has no system dependencies: all you need
to build jbuilder and packages using jbuilder is OCaml. You don't need
`make` or `bash` as long as the packages themselves don't use `bash`
explicitly.

Especially, one can install OCaml on Windows with a binary installer
and then use only the Windows Console to build Jbuilder and packages
using Jbuilder.

Strengths
---------

### Composable

Take n repositories that use Jbuilder, arrange them in any way on the
file system and the result is still a single repository that Jbuilder
knows how to build at once.

This make simultaneous development on multiple packages trivial.

### Gracefully handles multi-package repositories

Jbuilder knows how to handle repositories containing several
packages. When building via [opam][opam], it is able to correctly use
libraries that were previously installed even if they are already
present in the source tree.

The magic invocation is:

```sh
$ jbuilder build --only-packages <package-name> @install
```

### Building against several configurations at once

Jbuilder is able to build a given source code repository against
several configurations simultaneously. This helps maintaining packages
across several versions of OCaml as you can tests them all at once
without hassle.

This feature should make cross-compilation easy, see details in the
[roadmap](ROADMAP.md).

This feature requires [opam][opam].

### Jenga bridge

[Jenga][jenga] is another build system for OCaml that has more
advanced features such as polling or much better editor
integration. Jenga is more powerful and more complex and as a result
has many more dependencies.  It is planned to implement a small bridge
between the two so that a Jbuilder project can build with Jenga using
this bridge.

Requirements
------------

Jbuilder requires OCaml version 4.02.3 or greater.

installation
------------

The recommended way to install jbuilder is via the
[opam package manager][opam]:

```sh
$ opam install jbuilder
```

You can also build it manually with:

```sh
$ make release
$ make install
```

Note however that `make install` requires the `opam-installer`
tool. Running simply `make` will build jbuilder using the development
settings.

If you do not have `make`, you can do the following:

```sh
$ ocaml bootstrap.ml
$ ./boot.exe
$ ./_build/default/bin/main.exe install
```

Support
-------

If you have questions about jbuilder, you can send an email to
ocaml-core@googlegroups.com or [open a ticket on github][issues].

Status
------

Jbuilder is now in beta testing stage. Once a bit more testing has
been done, it will be released in 1.0.

Roadmap
-------

See [the roadmap](ROADMAP.md) for the current plan. Help on any of
these points is welcome!

FAQ
---

### Why do many Jbuilder projects contain a Makefile?

Many Jbuilder project contain a toplevel `Makefile`. It is often only
there for convenience, for the following reasons:

1. there are many different build systems out there, all with a
   different CLI. If you have been hacking for a long time, the one
   true invocation you know is `make && make install`, possibly
   preceded by `./configure`

2. you often have a few common operations that are not part of the
   build and `make <blah>` is a good way to provide them

3. `make` is shorter to type than `jbuilder build @install`

### How to add a configure step to a jbuilder project?

[example/sample-projects/with-configure-step](example/sample-projects/with-configure-step) shows
one way to do it which preserves composability; i.e. it doesn't require manually
running `./configure` script when working on multiple projects at the same time.

### Can I use topkg with jbuilder?

Yes, have a look at the [topkg-jbuilder][topkg-jbuilder] project for
more details.

### Where can I find some examples of projects using Jbuilder?

The [dune-universe](https://github.com/dune-universe/dune-universe)
repository contains a snapshot of the latest versions of all opam packages
depending on jbuilder.  It is therefore a useful reference to search through
to find different approaches to constructing build rules.

Known issues
------------

### mli only modules

These are supported, however using them might cause make it impossible
for non-jbuilder users to use your library. We tried to use them for
some internal module generated by Jbuilder and it broke the build of
projects not using Jbuilder:

https://github.com/ocaml/dune/issues/567

So, while they are supported, you should be careful where you use
them. Using a `.ml` only module is still preferable.

Implementation details
----------------------

This section is for people who want to work on Jbuilder itself.

### Bootstrap

In order to build itself, Jbuilder uses an OCaml script
([bootstrap.ml](bootstrap.ml)) that dumps most of the sources of Jbuilder into a
single `boot.ml` file. This file is built using `ocamlopt` or `ocamlc`
and used to build everything else.

Note that we don't include all of the sources in boot.ml. We skip a
few parts to speed up the build. In particular:
- vendored libraries are replaced by simpler implementations taken
  from `vendor/boot`
- a few files in `src` have an alternative version. These alternatives
  versions are named `XXX.boot.EXT`. For instance: `glob_lexer.boot.ml`

### OCaml compatibility test

Install opam switches for all the entries in the
[jbuild-workspace.dev](jbuild-workspace.dev) file and run:

```sh
$ make all-supported-ocaml-versions
```

### Repository organization

- `vendor/` contains dependencies of Jbuilder, that have been vendored
- `plugin/` contains the API given to `jbuild` files that are OCaml
  scripts
- `src/` contains the core of `Jbuilder`, as a library so that it can
  be used to implement the Jenga bridge later
- `bin/` contains the command line interface
- `doc/` contains the manual and rules to generate the manual pages

### Design

Jbuilder was initially designed to sort out the public release of Jane
Street packages which became incredibly complicated over time. It is
still successfully used for this purpose.

One necessary feature to achieve this is the ability to precisely
report the external dependencies necessary to build a given set of
targets without running any command, just by looking at the source
tree. This is used to automatically generate the `<package>.opam`
files for all Jane Street packages.

To implement this, the build rules are described using a build arrow,
which is defined in [src/build.mli](src/build.mli). In the end it makes the
development of the internal rules of Jbuilder very composable and
quite pleasant.

To deal with process multiplexing, Jbuilder uses a simplified
Lwt/Async-like monad, implemented in [src/future.mli](src/future.mli).

#### Code flow

- [src/jbuild.mli](src/jbuild.mli) contains the internal representation
  of `jbuild` files and the parsing code
- [src/jbuild_load.mli](src/jbuild_load.mli) contains the code to scan
  a source tree and build the internal database by reading
  the `jbuild` files
- [src/gen_rules.mli](src/gen_rules.mli) contains all the build rules
  of Jbuilder
- [src/build_system.mli](src/build_system.mli) contains a trivial
  implementation of a Build system. This is what Jenga will provide
  when implementing the bridge

Dune (Jbuilder) - A composable build system
===========================================

__Jbuilder has been renamed to Dune__. A full renaming of the documentation and
the tool will be done as part of the 1.0 release.

Jbuilder is a build system designed for OCaml/Reason projects only. It
focuses on providing the user with a consistent experience and takes
care of most of the low-level details of OCaml compilation. All you
have to do is provide a description of your project and Jbuilder will
do the rest.

The scheme it implements is inspired from the one used inside Jane
Street and adapted to the open source world. It has matured over a
long time and is used daily by hundreds of developers, which means
that it is highly tested and productive.

Jbuilder comes with a [manual][manual]. If you want to get started
without reading too much, you can look at the [quick start
guide][quick-start] or watch [this introduction video][video].

The [example][example] directory contains examples of projects using
jbuilder.

[![Travis status][travis-img]][travis] [![AppVeyor status][appveyor-img]][appveyor]

[manual]:         https://jbuilder.readthedocs.io/en/latest/
[quick-start]:    https://jbuilder.readthedocs.io/en/latest/quick-start.html
[example]:        https://github.com/ocaml/dune/tree/master/example
[travis]:         https://travis-ci.org/ocaml/dune
[travis-img]:     https://travis-ci.org/ocaml/dune.svg?branch=master
[appveyor]:       https://ci.appveyor.com/project/diml/dune/branch/master
[appveyor-img]:   https://ci.appveyor.com/api/projects/status/rsxayce22e8f2jkp?svg=true
[merlin]:         https://github.com/ocaml/merlin
[opam]:           https://opam.ocaml.org
[jenga]:          https://github.com/janestreet/jenga
[issues]:         https://github.com/ocaml/dune/issues
[topkg-jbuilder]: https://github.com/diml/topkg-jbuilder
[video]:          https://youtu.be/BNZhmMAJarw

Overview
--------

Jbuilder reads project metadata from `jbuild` files, which are either
static files in a simple S-expression syntax or OCaml scripts. It uses
this information to setup build rules, generate configuration files
for development tools such as [merlin][merlin], handle installation,
etc...

Jbuilder itself is fast, has very low overhead and supports parallel
builds on all platforms. It has no system dependencies: all you need
to build jbuilder and packages using jbuilder is OCaml. You don't need
`make` or `bash` as long as the packages themselves don't use `bash`
explicitly.

Especially, one can install OCaml on Windows with a binary installer
and then use only the Windows Console to build Jbuilder and packages
using Jbuilder.

Strengths
---------

### Composable

Take n repositories that use Jbuilder, arrange them in any way on the
file system and the result is still a single repository that Jbuilder
knows how to build at once.

This make simultaneous development on multiple packages trivial.

### Gracefully handles multi-package repositories

Jbuilder knows how to handle repositories containing several
packages. When building via [opam][opam], it is able to correctly use
libraries that were previously installed even if they are already
present in the source tree.

The magic invocation is:

```sh
$ jbuilder build --only-packages <package-name> @install
```

### Building against several configurations at once

Jbuilder is able to build a given source code repository against
several configurations simultaneously. This helps maintaining packages
across several versions of OCaml as you can tests them all at once
without hassle.

This feature should make cross-compilation easy, see details in the
[roadmap](ROADMAP.md).

This feature requires [opam][opam].

### Jenga bridge

[Jenga][jenga] is another build system for OCaml that has more
advanced features such as polling or much better editor
integration. Jenga is more powerful and more complex and as a result
has many more dependencies.  It is planned to implement a small bridge
between the two so that a Jbuilder project can build with Jenga using
this bridge.

Requirements
------------

Jbuilder requires OCaml version 4.02.3 or greater.

installation
------------

The recommended way to install jbuilder is via the
[opam package manager][opam]:

```sh
$ opam install jbuilder
```

You can also build it manually with:

```sh
$ make release
$ make install
```

Note however that `make install` requires the `opam-installer`
tool. Running simply `make` will build jbuilder using the development
settings.

If you do not have `make`, you can do the following:

```sh
$ ocaml bootstrap.ml
$ ./boot.exe
$ ./_build/default/bin/main.exe install
```

Support
-------

If you have questions about jbuilder, you can send an email to
ocaml-core@googlegroups.com or [open a ticket on github][issues].

Status
------

Jbuilder is now in beta testing stage. Once a bit more testing has
been done, it will be released in 1.0.

Roadmap
-------

See [the roadmap](ROADMAP.md) for the current plan. Help on any of
these points is welcome!

FAQ
---

### Why do many Jbuilder projects contain a Makefile?

Many Jbuilder project contain a toplevel `Makefile`. It is often only
there for convenience, for the following reasons:

1. there are many different build systems out there, all with a
   different CLI. If you have been hacking for a long time, the one
   true invocation you know is `make && make install`, possibly
   preceded by `./configure`

2. you often have a few common operations that are not part of the
   build and `make <blah>` is a good way to provide them

3. `make` is shorter to type than `jbuilder build @install`

### How to add a configure step to a jbuilder project?

[example/sample-projects/with-configure-step](example/sample-projects/with-configure-step) shows
one way to do it which preserves composability; i.e. it doesn't require manually
running `./configure` script when working on multiple projects at the same time.

### Can I use topkg with jbuilder?

Yes, have a look at the [topkg-jbuilder][topkg-jbuilder] project for
more details.

### Where can I find some examples of projects using Jbuilder?

The [jbuilder-universe](https://github.com/janestreet/jbuilder-universe)
repository contains a snapshot of the latest versions of all opam packages
depending on jbuilder.  It is therefore a useful reference to search through
to find different approaches to constructing build rules.

Known issues
------------

### Optional libraries inside a multilib directory

https://github.com/ocaml/dune/issues/51

If a directory contains several libraries and some are marked as
optional (by adding `(optional)` in the `(library ...)` stanza), then
the dependencies will still be required to perform the build.

This could be sorted out with some refactoring, but there is a simple
workaround, so it is low-priority.

#### Workaround

Put each optional library in a separate directory.

### mli only modules

https://github.com/ocaml/dune/issues/9

Due to the low-level details of OCaml compilation, it is currently
possible to write a module that has only a `.mli` and no `.ml`
file. This works as long as the mli contains only type declarations.

This is not a properly supported feature of the compiler, and in
particular it is not possible to alias such modules or use them as the
argument of a functor. Moreover, if you do write a value declaration,
or even just define an exception in the `.mli`, then you won't get an
error until the point where you link an executable using this module.

For these reason, mli only modules are not recommended by Jbuilder
until the compiler support them properly.

#### Workaround

As long as a module type contains no value declaration, it is possible
to turn in to an implementation by using a recursive module:

```ocaml
module rec M : sig
  type t = A | B
end = M
include M
```

So if you have a module without a `.ml` file, simply generate a `.ml`
from the `.mli` using this trick. For instance you can add the
following rule into your jbuild file:

```scheme
(rule (with-output-to foo.ml
       (progn
        (echo "module rec HACK : sig\n")
        (cat foo.mli)
        (echo "\nend = HACK\ninclue HACK\n"))))
```

In fact, jbuilder will automatically add this rule if you have a
module without imlpementation. However it will print a warning.

Implementation details
----------------------

This section is for people who want to work on Jbuilder itself.

### Bootstrap

In order to build itself, Jbuilder uses an OCaml script
([bootstrap.ml](bootstrap.ml)) that dumps most of the sources of Jbuilder into a
single `boot.ml` file. This file is built using `ocamlopt` or `ocamlc`
and used to build everything else.

Note that we don't include all of the sources in boot.ml. We skip a
few parts to speed up the build. In particular:
- vendored libraries are replaced by simpler implementations taken
  from `vendor/boot`
- a few files in `src` have an alternative version. These alternatives
  versions are named `XXX.boot.EXT`. For instance: `glob_lexer.boot.ml`

### OCaml compatibility test

Install opam switches for all the entries in the
[jbuild-workspace.dev](jbuild-workspace.dev) file and run:

```sh
$ make all-supported-ocaml-versions
```

### Repository organization

- `vendor/` contains dependencies of Jbuilder, that have been vendored
- `plugin/` contains the API given to `jbuild` files that are OCaml
  scripts
- `src/` contains the core of `Jbuilder`, as a library so that it can
  be used to implement the Jenga bridge later
- `bin/` contains the command line interface
- `doc/` contains the manual and rules to generate the manual pages

### Design

Jbuilder was initially designed to sort out the public release of Jane
Street packages which became incredibly complicated over time. It is
still successfully used for this purpose.

One necessary feature to achieve this is the ability to precisely
report the external dependencies necessary to build a given set of
targets without running any command, just by looking at the source
tree. This is used to automatically generate the `<package>.opam`
files for all Jane Street packages.

To implement this, the build rules are described using a build arrow,
which is defined in [src/build.mli](src/build.mli). In the end it makes the
development of the internal rules of Jbuilder very composable and
quite pleasant.

To deal with process multiplexing, Jbuilder uses a simplified
Lwt/Async-like monad, implemented in [src/future.mli](src/future.mli).

#### Code flow

- [src/jbuild.mli](src/jbuild.mli) contains the internal representation
  of `jbuild` files and the parsing code
- [src/jbuild_load.mli](src/jbuild_load.mli) contains the code to scan
  a source tree and build the internal database by reading
  the `jbuild` files
- [src/gen_rules.mli](src/gen_rules.mli) contains all the build rules
  of Jbuilder
- [src/build_system.mli](src/build_system.mli) contains a trivial
  implementation of a Build system. This is what Jenga will provide
  when implementing the bridge

# Dune

<!-- badges: start -->
  [![Build Status](https://travis-ci.com/HectorRDB/Dune.svg?token=zVqScmB6wvrS9ZJSK57p&branch=before_bioc_release)](https://travis-ci.com/HectorRDB/Dune) [![codecov](https://codecov.io/gh/HectorRDB/Dune/branch/master/graph/badge.svg?token=snxfXtj87B)](https://codecov.io/gh/HectorRDB/Dune)
<!-- badges: end -->

<p align="center">
  <img src="vignettes/logo.png" width="30%"/>
</p>

## Contents

- [Overview](#overview)
- [Installation](#installation)
- [Demo](#demo)
- [Issues](#Issues-and-bug-reports)
- [License](./LICENSE.md)
- [Citation](./inst/CITATION)


## Overview

Dune is an R Package that provides a parameter-free method for optimizing the trade-off between the resolutionof the clusters and their replicability across datasets. Dune  method takes as input a set of clustering results on a dataset, and iteratively merges clusters within those clusterings in order to maximize their concordance.  


## Installation

To install the current version of *Dune* , you will need Bioconductor 3.11 (i.e devel version):

```
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# The following initializes usage of Bioc devel
BiocManager::install(version='devel')

BiocManager::install("Dune")
```
A version compatible with R 3.6 is also available with the following commands:

```
if(!requireNamespace("devtools", quietly = TRUE)) {
 install.packages("devtools") 
}
devtools::install_github("HectorRDB/Dune", ref = "before_bioc_release")
```

The installation should only take a few seconds.
The dependencies of the package are listed in the DESCRIPTION file of the package. 
To install R and Bioconductor, please refer to the [Bioconductor install page](https://www.bioconductor.org/install/).

## Issues and bug reports

Please use the [github issues](https://github.com/HectorRDB/Dune/issues) to submit issues, bug reports, and comments.

## Demo 

Start with the vignette [online](https://hectorRDB.github.io/Dune/articles/Dune.html).


If you to look at the vignette source code, you can either run `browseVignettes(package = "Dune")` if you install the package from Bioconductor find the code for the vignette [here](./vignettes/Dune.Rmd)

# Developer documentation

We use `doc/dev` for storing developer documentation, including specification,
design and implementation notes. Anything that should be user-facing should go
into the `doc` folder instead.

# setup-dkml

Everything in this folder except this README.md is auto-generated by `setup-dkml`.

The documentation for `setup-dkml` is at https://github.com/diskuv/dkml-workflows#readme.

Bug reports go to https://github.com/diskuv/dkml-workflows/issues

## FAQ

### What is the workflow to update the generated files?

Use the following to update the generated code:

```bash
opam update
make update-dkml
```

Of course that only works if there are updates available.
And it will only do updates for the latest OCaml compiler.

### Can we test more than one OCaml compiler?

Technically no; `setup-dkml` follows the supported OCaml compiler of DKML. What
we've done in this project is rename the directories that supported the older
DKML versions (`4.12.1`). Essentially, the older DKML versions are frozen and
will no longer get updates. You may want to eventually drop the older DKML
versions by removing those directories and removing those versions from
`workflow.yml`.

### What is the version of MSVC targeted? What is the required environment to execute the `desktop-ci-*` targets in the main `Makefile`?

There is no "target" per se. The CI machine (or your developer machine) is queried for all MSVC versions, and any versions that are known to be compatible with OCaml are used.
The precise definition is `Get-CompatibleVisualStudios` at https://github.com/diskuv/dkml-runtime-distribution/blob/main/src/windows/Machine/Machine.psm1; that package is available for common use in the central Opam repository.

The compatible versions as of 2022-11-23 are Visual Studio 2019 with
a) Windows 10 SDK (10.0.18362.0) and b) MSVC v142 - VS 2019 C++ x64/x86 build tools (v14.25) or (v14.26).

*Advanced: Machine.psm1 has notes why other MSVC 2019 versions are not compatible with OCaml*

More importantly, just running it will show what is on the system and what Visual Studio components are missing (if any).

# Monorepo Bench

Files for building a docker image for benchmarking dune building a large
monorepo composed of packages from the opam repo. Running `make bench` will
build the monorepo and print out a json object with benchmark results expected
to be consumed by current-bench.

During `docker build`, an executable dune project is created called
`monorepo_bench` with
[this](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/benchmark/dune)
dune file listing all the libraries to include in the build.
This project is assembled with opam-monorepo, but we avoid running `opam
monorepo pull` while building the docker image as that command can fail if any
one of the thousands of dependencies are unavailable (which can happen for a
number of reasons). Instead, the duniverse directory must be provided as a
docker volume when running the container. For compatibility with the
current-bench environment, the duniverse directory must be located at:
```
/home/opam/bench-dir/current-bench-data/duniverse
```
The duniverse directory can be created with [this
script](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/generate-duniverse.sh).

The benchmark runner is a program which invokes dune in a number of
benchmarking scenarios and prints out the results in a format understood by
[current-bench](https://github.com/ocurrent/current-bench). The source code for
the benchmark runner is
[here](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/tree/main/dune-benchmark-runner).

## Running the benchmark locally

Build the benchmark docker image. Run this from the top level of the dune repo:
```
$ docker build . -f bench/monorepo/bench.Dockerfile --tag dune-monorepo-benchmark
```

Get the script that generates the duniverse dir:
```
$ git clone https://github.com/ocaml-dune/ocaml-monorepo-benchmark.git
$ cd ocaml-monorepo-benchmark
```

Run the script to generate the duniverse dir. This will generate it at
/tmp/duniverse but you can put it anywhere.
```
$ ./generate-duniverse.sh /tmp
```

Run the benchmark. The `/tmp/duniverse` in the argument to `--volume`
corresponds to the duniverse directory generated above.
```
$ docker run --rm -it --volume=/tmp/duniverse:/home/opam/bench-dir/current-bench-data/duniverse dune-monorepo-benchmark bash --login -c 'make bench'
```

Dune - A Composable Build System
================================

Dune is a build system designed for OCaml/Reason projects only. It
focuses on providing the user with a consistent experience and takes
care of most low-level details of OCaml compilations. It's merely necessary 
to provide a description of your project, and Dune will
do the rest.

It implements a scheme that's inspired from the one used inside Jane
Street and adapted to the open source world. It has matured over a
long time and is used daily by hundreds of developers, meaning 
it's highly tested and productive.

Dune comes with a [manual][manual]. If you want to get started
without reading too much, look at the [quick start
guide][quick-start] or watch [this introduction video][video].

The [example][example] directory contains examples of projects using
dune.

[![Main workflow](https://github.com/ocaml/dune/actions/workflows/workflow.yml/badge.svg)](https://github.com/ocaml/dune/actions/workflows/workflow.yml)
[![AppVeyor status][appveyor-img]][appveyor]
[![Release][release-img]][release]

[manual]:         https://dune.readthedocs.io/en/latest/
[quick-start]:    https://dune.readthedocs.io/en/latest/quick-start.html
[example]:        https://github.com/ocaml/dune/tree/master/example
[appveyor]:       https://ci.appveyor.com/project/diml/dune/branch/master
[appveyor-img]:   https://ci.appveyor.com/api/projects/status/rsxayce22e8f2jkp?svg=true
[release]:        https://github.com/ocaml/dune/releases
[release-img]:    https://img.shields.io/github/release/ocaml/dune.svg
[merlin]:         https://github.com/ocaml/merlin
[opam]:           https://opam.ocaml.org
[issues]:         https://github.com/ocaml/dune/issues
[dune-release]:   https://github.com/ocamllabs/dune-release
[video]:          https://youtu.be/BNZhmMAJarw

Overview
--------

Dune reads project metadata from `dune` files, which are either
static files in a simple S-expression syntax or OCaml scripts. It uses
this information to setup build rules, generate configuration files
for development tools such as [Merlin][merlin], handle installation,
etc.

Dune itself is fast, has very low overhead, and supports parallel
builds on all platforms. It has no system dependencies. OCaml is all you need
to build Dune and packages using Dune. You don't need
`make` or `bash`, as long as the packages themselves don't use `bash`
explicitly.

In particular, one can install OCaml on Windows with a binary installer
and then use only the Windows Console to build Dune and packages
using Dune.

Strengths
---------

### Composable

Take *n* repositories that use Dune and arrange them in any way on the
file system. The result is still a single repository that Dune
knows how to build at once.

This make simultaneous development on multiple packages trivial.

### Gracefully Handles Multi-Package Repositories

Dune knows how to handle repositories containing several
packages. When building via [Opam][opam], it is able to correctly use
libraries that were previously installed, even if they are already
present in the source tree.

The magic invocation is:

```sh
$ dune build --only-packages <package-name> @install
```

### Building Against Several Configurations at Once

Dune can build a given source code repository against
several configurations simultaneously. This helps maintaining packages
across several versions of OCaml, as you can test them all at once
without hassle.

In particular, this makes it easy to handle [cross-compilation](https://dune.readthedocs.io/en/latest/cross-compilation.html).

This feature requires [Opam][opam].

Requirements
------------

Dune requires OCaml version 4.08.0 to build itself and can build OCaml
projects using OCaml 4.02.3 or greater.

Installation
------------

We recommended installing Dune via the [Opam package manager][opam]:

```sh
$ opam install dune
```

If you are new to Opam, make sure to run `eval $(opam config env)` to
make `dune` available in your `PATH`. The `dune` binary is self-contained 
and relocatable, so you can safely copy it somewhere else to
make it permanently available.

You can also build it manually with:

```sh
$ make release
$ make install
```

If you do not have `make`, you can do the following:

```sh
$ ocaml boot/bootstrap.ml
$ ./dune.exe build -p dune --profile dune-bootstrap
$ ./dune.exe install dune
```

The first command builds the `dune.exe` binary. The second builds the
additional files installed by Dune, such as the *man* pages, and
the last simply installs all of that on the system.

**Please note**: unless you ran the optional `./configure` script, you can
simply copy `dune.exe` anywhere and it will just work. `dune` is
fully relocatable and discovers its environment at runtime rather than
hard-coding it at compilation time.

Support
-------

If you have questions about Dune, you can send an email to
ocaml-core@googlegroups.com or [open a ticket on GitHub][issues].


Migration from Jbuilder
-----------------------

Dune was formerly known as Jbuilder. Migration from Jbuilder to Dune is
described in the [manual](http://dune.readthedocs.io/en/latest/migration.html).

Status
------

Dune is fairly stable and used by the majority of packages on
Opam. Note that Dune retains backward compatibility with Jbuilder, and
in particular, existing Jbuilder projects will continue to be buildable
with Dune.

This project is called `hello_world`. It defines one library called
`hello_world` and one executable called `hello_world`.

The library is defined in `lib` and the executable in `bin`. It also defines a
test in `test`.

At the toplevel of the project, there is a `hello_world.opam` file. This file
is required so that Dune knows that this is the `hello_world` project.

To build everything that is meant to be installed in this project, type:

    $ dune build @install

To run the tests, type:

    $ dune runtest

This directory has various tests for the digest cache of Dune.

In most of the tests, we use --wait-for-filesystem-clock to make the
tests more reproducible. Without it, Dune might drop some entries or
not depending on whether the time lap between targets being produced
and dune saving the digest cache is smaller than the file system
granularity.

This file describes how the dune test suite is organized.

# Unit testing

The `unit-tests` directory contains unit tests, written in expectation
style. The test logic is implemented in `unit-tests/expect_test.mll`
and the various tests are in `.mlt` files.

The way they work is quite simple; each `.mlt` file is a succession of
toplevel phrases followed by the output reported by the OCaml toplevel
enclosed in an `[%%expect]` extension point. For instance:

```ocaml
6 * 7;;
[%%expect{|
- : int = 42
|}]
```

# Blackbox testing

The `blackbox-tests` contains blackbox tests. I.e. we are testing the
fully built `dune` executable on various example projects.

The tests are written in [cram](https://bitheap.org/cram/) style. The
logic is implemented in `blackbox-tests/cram.mll`. It only implements
a minimal subset of cram testing. In particular the shell environment
is currently not preserved between commands, so you cannot define a
variable and use it.

Test cases are in `blackbox-tests/test-cases`. Each sub-directory is a
full blown dune project. Each sub-directory contains a `run.t`
file, which represents a few invocations of dune along with the
expected output.

Here is a sample `.t` file:

```
This is a comment

  $ echo 'Hello, world!'
  Hello, world!

  $ cat plop
  cat: plop: No such file or directory
  [1]

The [1] represent the exit code of the command. It is printed when it
is non-zero
```

This project shows how to add a configure step to a project using Dune.

In order to keep things composable, it offers several way to configure
the project:

1.  with the classic `./configure <args...>`. When doing this, the
    configuration script are run immediately, and the resulting configuration
    is frozen for the rest of the build.

2.  by copying `config.defaults` to `config` and editing it. The configuration
    scripts will be run as part of the build using what is written in `config`.
    When `config` is edited, the configuration scripts will be automatically
    rerun.

3.  by doing nothing. The configuration scripts will be run as part of the
    build using the default values provided in `config.defaults`.

Technically this is how it works:

`configure` is a simple OCaml script that:

- parses the command line
- generates a `config` file using what was given on the command line
- calls `ocaml real_configure.ml`
- `real_configure.ml` does some stuff and generates `config.full`

`config.full` is what is used by the rest of the build.

Now in order to support (2) and (3), we setup the following rules in the
toplevel `dune` file:

- a rule to produce `config` by copying `config.default`
- a rule to produce `config.full` by running `real_configure.ml`

This all works as described because if Dune knows how to build a file and this
file is already present in the source tree, it will always prefer the file
that's already there.

# RPC Client Example

This project contains an executable `rpc_client` which connects to the RPC
server started when Dune is run in watch mode. To use this program, start Dune
in watch mode:

```
$ dune build --watch
```

Then run `rpc_client` from the same directory as `dune` was run from. The
`rpc_client` program will connect to the server (`dune build --watch` starts an
RPC server) and perform several RPC calls to it before instructing the server
(over RPC) to shutdown.

Cannibalized bits of Uucp:

- `Notty_uucp_data` is generated from an actual Uucp installation.
- `Notty_uucp` uses it to provide the few Unicode properties that Notty needs.
- `Notty_grapheme_cluster` is `Grapheme_cluster` from Uuseg, adapted to use the
  above.

Compiled size of these is on the order of 70K. Uucp is presently a monolithic 10M.

The idea is to remove these in favor of the actual Uucp/Uuseg, as soon as it
becomes possible to depend only on the necessary parts of Uucp.

Uucp and Uuseg are Copyright (c) 2014 Daniel C. Bünzli.

Using the mdx stanza, you can set mdx prelude which are used to evaluate
code in the toplevel environments before checking a document. To do that
you can set the prelude field of the stanza: `(preludes <prelude_file> ...)`.

```ocaml
# x;;
- : int = 1
```

Different environment can have their own prelude. This is set using the
`(env <env_name> <prelude_file>)` syntax for the prelude field.
Here x and why are set in `alt.ml`.

```ocaml env=a
# x + y;;
- : int = 11
```

```ocaml env=b
# x + y;;
- : int = 21
```

We can use our local packages, in our documentation thanks to dune's mdx stanza:

```ocaml
# #require "pkg.public-lib";;
# Public_lib.x;;
- : int = 0
```

We can also use our local public executables:

```sh
$ public-bin
Hey!
```

# README for incremental_cycles library

This library is vendored from
https://gitlab.inria.fr/agueneau/incremental-cycles

## Details on the vendoring process

The vendoring process is a bit involved due to the way the library is
specified upstream. In particular, it assumes a graph interface
`Raw_graph` that we have to copy by hand in Dune [see
`src/dag/dag.ml`], and in particular we have to be careful about not
altering the complexity guarantees.


## Complexity guarantees

The complexity and correctness of the implementation of
`incremental_cycles` has been mechanically-verified using the Coq
theorem prover. Note however, that for the main theorem to hold there
are a few requirements that cannot be captured by ML-level interfaces;
more concretely:

- the current specification for the algorithm requires the
  `get_outgoing` function provided by the client to return a list of
  all successors, and do so in constant time. This is quite demanding,
  as basically requires the client to already have the list at hand.

- the main theorem for `Dag.add` does require that the vertex is not
  already in the graph; otherwise the theorem doesn't apply. Thus,
  clients must ensure that no duplicate edge is added to the graph.

## Dune-specific modifications

Dune uses incremental_cycles in a way that the no-duplicate-egdes
requirement is not satisfied by construction; thus, before a call to
`Dag.add` edge membership on the graph must be checked.

This is a common operation and thus should be done efficiently, thus
Dune performs the following modifications to `dag.ml`:

- we add a set of children nodes in addition to the current list
- we modify `raw_add_edge` so it updates this set, and `is_child` so
  it uses the efficient membership set

The rationale for adding a duplicate children field is to actually
preserve the order the edges were added, this could be important in
other parts of the algo, see comment on `is_child` use at `memo.ml`.

For more details see discussion at https://github.com/ocaml/dune/pull/2959

Dependencies for file-include mdx blocks should be computed by dune, using
`ocaml-mdx deps`:

```ocaml file=example.ml
let x = 1
```

Same goes for mdx's `dir` labels:

```sh dir=stuff
$ ls
blah
bleh
blih
```

# Sample dune-coq project

This is a simple dune build which has 2 projects:

1. `example-ocaml` - simple pure OCaml library
2. `example-coq` - simple OCaml library extracted from Coq

In addition to Coq to OCaml extraction, it also demonstrates a
workaround for long-standing Coq [issue](https://github.com/coq/coq/issues/6614) and until it is fixed
could be used as a template for dune projects using Coq extraction.

We can use our local packages, in our documentation thanks to dune's mdx stanza:

```ocaml
# #require "pkg.public-lib";;
# Public_lib.x;;
- : int = 1
```

```ocaml
# #require "pkg2.public-lib2";;
# Public_lib2.x;;
- : int = 2
```

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(paste0(
    "Dune now uses the Normalized Mutual Information",
    "instead of the adjusted Rand Index. You can restore the old option by",
    "setting metric'ARI' in the Dune function."))
}

utils::globalVariables(c("new"))

#' @title ARI Matrix
#' @param clusMat The clustering matrix with a row per cell and a column per
#' clustering label type
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @return The ARI matrix
#' @details In the ARI matrix where each cell **i,j** is the adjusted Rand Index
#' between columns **i** and **j** of the original \code{clusMat}.
#' If \code{unclustered} is not NULL, the cells which have been assigned to the
#' \code{unclustered} cluster will not be counted towards computing the ARI.
#' @importFrom aricode ARI
#' @examples
#' data("clusMat", package = "Dune")
#' ARIs(clusMat)
#' @export
ARIs <- function(clusMat, unclustered = NULL) {
  ARI <- apply(clusMat, 2, function(x) {
    apply(clusMat, 2, function(y) {
      if (!is.null(unclustered)) {
        x_unc <- x[x != unclustered & y != unclustered]
        y_unc <- y[x != unclustered & y != unclustered]
      } else {
        x_unc <- x
        y_unc <- y
      }
      aricode::ARI(x_unc, y_unc)
    })
  })
  if (is.null(colnames(clusMat))) {
    rownames(ARI) <- colnames(ARI) <- seq_len(ncol(clusMat))
  } else {
    rownames(ARI) <- colnames(ARI) <- colnames(clusMat)
  }

  return(ARI)
}

#' @title NMI Matrix
#' @param clusMat The clustering matrix with a row per cell and a column per
#' clustering label type
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @return The NMI matrix
#' @details In the NMI matrix where each cell **i,j** is the normalized mutual 
#' information between columns **i** and **j** of the original \code{clusMat}.
#' If \code{unclustered} is not NULL, the cells which have been assigned to the
#' \code{unclustered} cluster will not be counted towards computing the NMI.
#' @importFrom aricode NMI
#' @examples
#' data("clusMat", package = "Dune")
#' NMIs(clusMat)
#' @export
NMIs <- function(clusMat, unclustered = NULL) {
  NMI <- apply(clusMat, 2, function(x) {
    apply(clusMat, 2, function(y) {
      if (!is.null(unclustered)) {
        x_unc <- x[x != unclustered & y != unclustered]
        y_unc <- y[x != unclustered & y != unclustered]
      } else {
        x_unc <- x
        y_unc <- y
      }
      aricode::NMI(x_unc, y_unc, variant = "sum")
    })
  })
  if (is.null(colnames(clusMat))) {
    rownames(NMI) <- colnames(NMI) <- seq_len(ncol(clusMat))
  } else {
    rownames(NMI) <- colnames(NMI) <- colnames(clusMat)
  }
  
  return(NMI)
}

#' @title When to Stop
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param p A value between 0 and 1. We stop when the metric used for merging has
#'  improved by p of the final total improvement. Default to 1 (i.e running the full merging).
#' @param average_n Alternatively, you can specify the average number of clusters you 
#' want to have. 
#' @return An integer giving the step where to stop.
#' @details The \code{\link{Dune}} process improves the metric. This return
#' the first merging step after which the metric has been improved by p of the
#' total. Setting p = 1 just return the number of merges.
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' whenToStop(merger, p = .5)
#' @importFrom dplyr n_distinct
#' @export
whenToStop <- function(merger, p = 1, average_n = NULL) {
  if (!is.null(average_n)) {
    init_average <- lapply(as.data.frame(merger$initialMat),
                           dplyr::n_distinct) %>%
      unlist() %>% mean()
    final_average <- lapply(as.data.frame(merger$currentMat),
                           dplyr::n_distinct) %>%
      unlist() %>% mean()
    if (average_n > init_average | average_n < final_average) {
      stop("the requested average is not possible given the initial and final clusters")
    }
    j <- min(which(
      init_average - 0:nrow(merger$merges) / ncol(merger$currentMat) < 
        average_n
    )) - 1
  } else {
    if (p < 0 | p > 1) stop("p must be between 0 and 1")
    if (merger$metric == "ARI") {
      ARI <- ARIImp(merger)
      j <- min(which(
        ARI[1:length(ARI)] >= min(ARI) + p * (max(ARI) - min(ARI))
        )) - 1
    } else if (merger$metric == "NMI") {
      NMI <- NMIImp(merger)
      j <- min(which(
        NMI[1:length(NMI)] >= min(NMI) + p * (max(NMI) - min(NMI))
      )) - 1
    }  
  }
  return(j)
}

#' @title adjustedRandIndex
#' @param tab The confusion matrix
#' @return The ARI
.adjustedRandIndex <- function(tab) {
  if (all(dim(tab) == c(1, 1))) return(1)
  a <- sum(choose(tab, 2))
  b <- sum(choose(rowSums(tab), 2)) - a
  c <- sum(choose(colSums(tab), 2)) - a
  d <- choose(sum(tab), 2) - a - b - c
  ARI <- (a - (a + b) * (a + c)/(a + b + c + d)) /
    ((a + b + a + c) / 2 - (a + b) * (a + c) / (a + b + c + d))
  return(ARI)
}

.NormalizedMutualInformation <- function(tab){
  if (all(dim(tab) == c(1, 1))) return(1)
  freqs_x <- rowSums(tab) 
  freqs_x <- freqs_x / sum(freqs_x)
  H_x <- -sum(ifelse(freqs_x > 0, freqs_x * log(freqs_x), 0))
  freqs_y <- colSums(tab)
  freqs_y <- freqs_y / sum(freqs_y)
  H_y <- -sum(ifelse(freqs_y > 0, freqs_y * log(freqs_y), 0))
  I_xy <- base::apply(tab, 1, function(freqs) {
    freqs <- freqs / sum(freqs)
    return(-sum(ifelse(freqs > 0, freqs * log(freqs), 0)))
  })
  I_xy <- H_y - sum(freqs_x * I_xy)
  return(2 * I_xy / (H_x + H_y))
}

.confusionMatrices <- function(pairPartitions, currentMat, metric) {
  confMats <- lapply(
    as.data.frame(pairPartitions, stringsAsFactors = FALSE),
    function(pair){
      C1 <- pair[1]
      C2 <- pair[2]
      confusionMatrix <- table(currentMat[, C1], currentMat[, C2])
      if (metric == "ARI") {
        Metric <- .adjustedRandIndex(confusionMatrix)
      } else if (metric == "NMI") {
        Metric <- .NormalizedMutualInformation(confusionMatrix)
      }
      
      return(list("C1" = C1, "C2" = C2, "confusionMatrix" = confusionMatrix,
                  "Metric" = Metric))
    }
  )
  return(confMats)
}

.localARI <- function(pair, confMats, C) {
  m1 <- as.character(min(pair))
  m2 <- as.character(max(pair))
  # We look at every confusion matrix
  localARIs <- lapply(confMats, function(confMat){
    # If the confusion matrix is one between C and another partition
    if (confMat$C1 == C) {
      confusionMatrix <- confMat$confusionMatrix
      # We merge the row
      confusionMatrix[m1, ] <- confusionMatrix[m1, ] +
        confusionMatrix[m2, ]
      confusionMatrix <- confusionMatrix[
        -which(rownames(confusionMatrix) == m2), ]
      # And we update the ARI
      return(.adjustedRandIndex(confusionMatrix) - confMat$Metric)
    } else{
      # If the confusion matrix is one between another partition and C
      if (confMat$C2 == C)  {
        confusionMatrix <- confMat$confusionMatrix
        # We merge the columns
        confusionMatrix[, m1] <- confusionMatrix[, m1] +
          confusionMatrix[, m2]
        confusionMatrix <- confusionMatrix[,
                                        -which(colnames(confusionMatrix) == m2)]
        # And we update the ARI
        return(.adjustedRandIndex(confusionMatrix) - confMat$Metric)
      } else {
        # Otherwise, nothing happens in that confusion matrix
        return(0)
      }
    }
  })
  localARIs <- unlist(localARIs)
  return(localARIs)
}

.localNMI <- function(pair, confMats, C) {
  m1 <- as.character(min(pair))
  m2 <- as.character(max(pair))
  # We look at every confusion matrix
  localNMIs <- lapply(confMats, function(confMat){
    # If the confusion matrix is one between C and another partition
    if (confMat$C1 == C) {
      confusionMatrix <- confMat$confusionMatrix
      # We merge the row
      confusionMatrix[m1, ] <- confusionMatrix[m1, ] +
        confusionMatrix[m2, ]
      confusionMatrix <- confusionMatrix[
        -which(rownames(confusionMatrix) == m2), ]
      # And we update the NMI
      return(.NormalizedMutualInformation(confusionMatrix) - confMat$Metric)
    } else{
      # If the confusion matrix is one between another partition and C
      if (confMat$C2 == C)  {
        confusionMatrix <- confMat$confusionMatrix
        # We merge the columns
        confusionMatrix[, m1] <- confusionMatrix[, m1] +
          confusionMatrix[, m2]
        confusionMatrix <- confusionMatrix[,
                                           -which(colnames(confusionMatrix) == m2)]
        # And we update the NMI
        return(.NormalizedMutualInformation(confusionMatrix) - confMat$Metric)
      } else {
        # Otherwise, nothing happens in that confusion matrix
        return(0)
      }
    }
  })
  localNMIs <- unlist(localNMIs)
  return(localNMIs)
}

.updatedConfMats <- function(pair, confMats, clusLabel, metric) {
  m1 <- as.character(min(pair))
  m2 <- as.character(max(pair))
  updatedConfMats <- lapply(confMats, function(confMat){
    updatedConfMat <- confMat
    if (confMat$C1 == clusLabel) {
      confusionMatrix <- confMat$confusionMatrix
      confusionMatrix[m1, ] <- confusionMatrix[m1, ] +
        confusionMatrix[m2, ]
      confusionMatrix <- confusionMatrix[
        -which(rownames(confusionMatrix) == m2), ]
      # If we have merged all, make sure it stays in matrix format
      if (!is.matrix(confusionMatrix)) {
        confusionMatrix <- matrix(confusionMatrix, nrow = 1)
        rownames(confusionMatrix) <- m1
        colnames(confusionMatrix) <- colnames(confMat$confusionMatrix)
      }
      updatedConfMat$confusionMatrix <- confusionMatrix
      if (metric == "ARI") {
        updatedConfMat$Metric <- .adjustedRandIndex(confusionMatrix)  
      } else if (metric == "NMI") {
        updatedConfMat$Metric <- .NormalizedMutualInformation(confusionMatrix)  
      }
      
    } else{
      if (confMat$C2 == clusLabel) {
        confusionMatrix <- confMat$confusionMatrix
        confusionMatrix[, m1] <- confusionMatrix[, m1] +
          confusionMatrix[, m2]
        confusionMatrix <- confusionMatrix[,
                                           -which(colnames(confusionMatrix) == m2)]
        # If we have merged all, make sure it stays in matrix format
        if (!is.matrix(confusionMatrix)) {
          confusionMatrix <- matrix(confusionMatrix, ncol = 1)
          colnames(confusionMatrix) <- m1
          rownames(confusionMatrix) <- rownames(confMat$confusionMatrix)
        }
        updatedConfMat$confusionMatrix <- confusionMatrix
        if (metric == "ARI") {
          updatedConfMat$Metric <- .adjustedRandIndex(confusionMatrix)  
        } else if (metric == "NMI") {
          updatedConfMat$Metric <- .NormalizedMutualInformation(confusionMatrix)  
        }
      }}
    return(updatedConfMat)
  })
  return(updatedConfMats)
}

context("Dune's main algorithm works as intended")

library(Dune)
library(dplyr)
set.seed(107070770)

test_that("Dune returns the right type of output", {
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat)
  expect_equal(length(merger), 5)
  expect_equal(names(merger),
               c("initialMat", "currentMat", "merges", "ImpMetric", "metric"))
  expect_is(merger$initialMat, "matrix")
  expect_is(merger$currentMat, "data.frame")
  expect_is(merger$merges, "data.frame")
  expect_is(merger$ImpMetric, "numeric")
  expect_output(Dune(clusMat, verbose = TRUE))
})


test_that("Dune correctly picks the best cluster for ARI", {
  for (i in 1:10) {
    clusMat <- matrix(sample(1:5, 500, replace = TRUE), ncol = 5)
    merger <- Dune(clusMat, metric = "ARI")
    if (nrow(merger$merges) <= 3) next()
    df <- intermediateMat(merger, n_steps = nrow(merger$merges) - 2)
    df <- as.matrix(df[, -1])
    init_ARI <- ARIs(df)
    for (j in 1:20) {
      col <- sample(colnames(df), 1)
      pair <- sample(t(unique(df[,col])), 2)
      m2 <- max(pair)
      m1 <- min(pair)
      df2 <- df
      df2[df2[, col] == m2, col] <- m1
      final_ARI <- ARIs(df2)
      expect_true(mean((final_ARI - init_ARI)[upper.tri(init_ARI)]) <=
                    merger$ImpMetric[nrow(merger$merges) - 1])
    }
  }
})

test_that("Dune correctly picks the best cluster for NMI", {
  for (i in 1:10) {
    clusMat <- matrix(sample(1:5, 500, replace = TRUE), ncol = 5)
    merger <- Dune(clusMat, metric = "ARI")
    if (nrow(merger$merges) <= 3) next()
    df <- intermediateMat(merger, n_steps = nrow(merger$merges) - 2)
    df <- as.matrix(df[, -1])
    init_ARI <- ARIs(df)
    for (j in 1:20) {
      col <- sample(colnames(df), 1)
      pair <- sample(t(unique(df[,col])), 2)
      m2 <- max(pair)
      m1 <- min(pair)
      df2 <- df
      df2[df2[, col] == m2, col] <- m1
      final_ARI <- ARIs(df2)
      expect_true(mean((final_ARI - init_ARI)[upper.tri(init_ARI)]) <=
                    merger$ImpMetric[nrow(merger$merges) - 1])
    }
  }
})

test_that('Dune works correctly for identical inputs',{
  data("clusMat", package = "Dune")
  expect_error(Dune(clusMat[,c(1,1)]))
  expect_output(Dune(cbind(rep(1:3, each = 20), rep(1:2, each = 30)),
                     verbose = TRUE),
                regexp = ".*We merge one of the partition entirely")
})


test_that('Dune works with unclustered values',{
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat, unclustered = -2)
  merger2 <- Dune(clusMat)
  expect_identical(merger, merger2)
})

#' A clustering matrix used to demonstrate the ari-merging process.
#'
#' @details This matrix has 100 samples with 5 cluster labels. Cluster labels 2
#'  trought 5 are modified versions of cluster label 1, where some clusters from
#'  label 1 where broken down into smaller clusters. 
#'  It is just a toy dataset that can be re-generated with the code in
#'  https://github.com/HectorRDB/Pipeline_Brain/blob/master/Sandbox/createToyDataset.R
#'
"clusMat"

#' Cluster labels for a subset of the allen Smart-Seq nuclei dataset
#'
#' @details This matrix of clusters was obtained by running 3 clustering algorithms
#' on a brain snRNA-Seq dataset from Tasic et .al (https://doi.org/10.1038/s41586-018-0654-5).
#' This dataset was then subsetted to the GABAergic neurons. 
#' Code to reproduce all this can be found in the github repository from the 
#' Dune paper (https://github.com/HectorRDB/Dune_Paper).
#'
"nuclei"

#' ARI improvement
#'
#' Compute the ARI improvement over the ARI merging procedure
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @return a vector with the mean ARI between methods at each merge
#' @seealso ARItrend
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' plot(0:nrow(merger$merges), ARIImp(merger))
#' @export
ARIImp <- function(merger, unclustered = NULL) {
  if (merger$metric != "ARI") {
    f <- function(clusMat) {
      ARI <- ARIs(clusMat %>% select(-cells)%>% as.matrix())
      return(mean(ARI[upper.tri(ARI)]))
    }
    ARI <- functionTracking(merger, f)
  } else if (merger$metric == "ARI") {
    baseARI <- ARIs(merger$initialMat, unclustered = unclustered)
    baseARI <- baseARI[upper.tri(baseARI)] %>% mean()
    ARI <- merger$ImpMetric  
    ARI <- c(baseARI, ARI)
    ARI <- cumsum(ARI)
  }
  return(ARI)
}

#' NMI improvement
#'
#' Compute the NMI improvement over the NMI merging procedure
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @return a vector with the mean NMI between methods at each merge
#' @seealso NMItrend
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' plot(0:nrow(merger$merges), NMIImp(merger))
#' @export
NMIImp <- function(merger, unclustered = NULL) {
  if (merger$metric != "NMI") {
    f <- function(clusMat) {
      NMI <- NMIs(clusMat %>% select(-cells) %>% as.matrix())
      return(mean(NMI[upper.tri(NMI)]))
    }
    NMI <- functionTracking(merger, f)
  } else if (merger$metric == "NMI") {
    baseNMI <- NMIs(merger$initialMat, unclustered = unclustered)
    baseNMI <- baseNMI[upper.tri(baseNMI)] %>% mean()
    NMI <- merger$ImpMetric  
    NMI <- c(baseNMI, NMI)
    NMI <- cumsum(NMI)
  }
  return(NMI)
}

#' clusterConversion
#'
#' Find the conversion between the old cluster and the final clusters
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param p A value between 0 and 1. We stop when the metric used for merging has
#'  improved by p of the final total improvement. Default to 1 (i.e running the full merging).
#' @param average_n Alternatively, you can specify the average number of clusters you 
#' want to have. 
#' @param n_steps Finally, you can specify the number of merging steps to
#' do before stopping.
#' @details If more than one of `p`,`average_n` and `n_steps` is specified, 
#' then the order of preference is `n_steps`, then `average_n` then `p`.
#' @return A list containing a matrix per clustering method, with a column for the old labels
#' and a column for the new labels.
#' @importFrom magrittr %>%
#' @importFrom dplyr n_distinct filter
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' clusterConversion(merger)[[2]]
#' @export
clusterConversion <- function(merger, p = 1, average_n = NULL, n_steps = NULL) {
  if (is.null(n_steps)) {
    j <- whenToStop(merger, p = p, average_n = average_n)
  } else {
    j <- n_steps
  }
  if (j == 0) {
    updates <- lapply(as.data.frame(merger$initialMat), function(clusters){
      return(data.frame(old = unique(clusters), new = unique(clusters)))
    })
  } else {
  # Compute ARI imp and find where to stop the merge
    merges <- merger$merges
    merges <- merges[seq_len(j), ]
    updates <- lapply(colnames(merger$initialMat), function(clusLab){
      clusters <- unique(merger$initialMat[, clusLab])
      update <- data.frame(old = clusters, new = clusters)
      return(update)
    })
    names(updates) <- colnames(merger$initialMat)
    for (i in seq_len(nrow(merges))) {
      clusLab <- merges[i, 1]
      pair <- as.numeric(merges[i, 2:3])
      clus <- max(pair)
      id <- updates[[clusLab]]$new == clus
      updates[[clusLab]][id, "new"] <- min(pair)
    }
  }
  return(updates)
}

#' Find the clustering matrix that we would get if we stopped the ARI merging
#' early
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param p A value between 0 and 1. We stop when the metric used for merging has
#'  improved by p of the final total improvement. Default to 1 (i.e running the full merging).
#' @param average_n Alternatively, you can specify the average number of clusters you 
#' want to have. 
#' @param n_steps Finally, you can specify the number of merging steps to
#' do before stopping.
#' @return A data.frame with the same dimensions as the currentMat of the merger
#' argument, plus one column with cell names, related to the rownames of the
#'  original input 
#' @details If more than one of `p`,`average_n` and `n_steps` is specified, 
#' then the order of preference is `n_steps`, then `average_n` then `p`.
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' head(intermediateMat(merger, n_steps = 1))
#' @importFrom magrittr %>%
#' @import dplyr
#' @import tidyr
#' @export
intermediateMat <- function(merger, p = 1, average_n = NULL, n_steps = NULL) {
  # Get the conversion
  oldToNew <- clusterConversion(merger = merger, p = p, average_n = NULL,
                                n_steps = n_steps)
  initialMat <- merger$initialMat %>% as.data.frame()
  names(oldToNew) <- colnames(initialMat)
  oldToNew <- bind_rows(oldToNew, .id = "cluster_label")
  initialMat <- initialMat %>%
    dplyr::mutate(cells = rownames(initialMat)) %>%
    tidyr::pivot_longer(cols = colnames(initialMat), values_to = "old",
                        names_to = "cluster_label")
  newMat <- full_join(oldToNew, initialMat,
                      by = c("old" = "old", "cluster_label" = "cluster_label")) %>%
    dplyr::select("cluster_label", "new", "cells") %>%
    dplyr::mutate("new" = as.integer(new)) %>%
    tidyr::pivot_wider(names_from = "cluster_label", values_from = "new")
  if (is.numeric(rownames(initialMat))) {
    newMat <- newMat %>%
      dplyr::mutate(cells = as.numeric(cells))
  }
  return(newMat)
}

#' Track the evolution of a function along merging
#'
#' For a given ARI merging, compute the evolution on the function f
#'
#' @param merger the result from having run \code{\link{Dune}}
#' on the dataset
#' @param f the function used. It must takes as input a clustering matrix and
#' return a value
#' @param p A value between 0 and 1. We stop when the metric used for merging has
#'  improved by p of the final total improvement. Default to 1 (i.e running the full merging).
#' @param n_steps Alternatively, you can specifiy the number of merging steps to
#' do before stopping.
#' @param ... additional arguments passed to f
#' @return a vector of length the number of merges
#' @importFrom magrittr %>%
#' @import dplyr
#' @examples
#' # Return the number of clusters for the fourth cluster label
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' f <- function(clusMat, i) dplyr::n_distinct(clusMat[, i])
#' functionTracking(merger, f, i = 4)
#' @export
functionTracking <- function(merger, f, p = 1, n_steps = NULL, ...){
  # Go over the merge and compute the function as we go
  if (is.null(n_steps)) {
    j <- whenToStop(merger, p = p)
  } else {
    j <- n_steps
  }
  values <- rep(0, j + 1)
  for (i in seq(0, j)) {
    values[i + 1] <- f(intermediateMat(merger, n_steps = i),
                       ...)
  }
  return(values)
}

context("Dune auxilliary functions")
library(Dune)
library(dplyr)
set.seed(76780915)
test_that("functionTracking and ARIImp are coherent", {
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat)
  f <- function(clusMat) {
    ARI <- ARIs(clusMat %>% select(-cells) %>% as.matrix())
    return(mean(ARI[upper.tri(ARI)]))
  }
  expect_equal(ARIImp(merger), functionTracking(merger, f))
  expect_equal(ARIImp(merger),
               functionTracking(merger, f, n_steps = length(merger$ImpMetric)))
})

test_that("functionTracking and NMIImp are coherent", {
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat, metric = "NMI")
  f <- function(clusMat) {
    NMI <- NMIs(clusMat %>% select(-cells) %>% as.matrix())
    return(mean(NMI[upper.tri(NMI)]))
  }
  expect_equal(NMIImp(merger), functionTracking(merger, f))
  expect_equal(NMIImp(merger),
               functionTracking(merger, f, n_steps = length(merger$ImpMetric)))
})

test_that("clusterConversion and intermediateMat are coherent with Dune", {
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat)
  # Same at the end 
  df <- intermediateMat(merger, p = 1) %>%
    mutate(cells = as.numeric(cells)) %>%
    arrange(cells) %>%
    select(-cells) %>%
    as.data.frame()
  expect_equal(df, merger$currentMat)
  # Same at the beginning
  df <- intermediateMat(merger, p = 0) %>%
    mutate(cells = as.numeric(cells)) %>%
    arrange(cells) %>%
    select(-cells) %>%
    as.matrix()
  expect_equal(df, merger$initialMat)
  # Same at the beginning
  df <- intermediateMat(merger, n_steps = 0) %>%
    mutate(cells = as.numeric(cells)) %>%
    arrange(cells) %>%
    select(-cells) %>%
    as.matrix()
  expect_equal(df, merger$initialMat)
  # Same if rownames
  rownames(merger$initialMat) <- seq_len(nrow(clusMat))
  df <- intermediateMat(merger, p = 1) %>%
    mutate(cells = as.numeric(cells)) %>%
    arrange(cells) %>%
    select(-cells) %>%
    as.data.frame()
  expect_equal(df, merger$currentMat)
})

test_that("Dune output", {
  data("clusMat", package = "Dune")
  expect_equal(sum(ARIs(Dune(clusMat)$currentMat) == 0), 0)
  expect_equal(nrow(Dune(clusMat)$merges), 7)
  expect_equal(unique(
    unlist(lapply(Dune(clusMat)$currentMat, n_distinct))),
    10)
})

test_that("You can only stop between zero and one", {
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat)
  expect_error(whenToStop(merger, -1))
  expect_error(whenToStop(merger, 2))
  expect_error(whenToStop(merger, "A"))
})

.findMergeResults <- function(C, clusters, unclustered, confMats, metric = "NMI") {
  # Get all pairs of clusters
  clusterNames <- clusters[[C]]
  if (!is.null(unclustered)) {
    if (length(clusterNames[clusterNames != unclustered]) == 1) {
      return(rep(0, length(confMats)))
    }
    clusPairs <- utils::combn(clusterNames[clusterNames != unclustered], 2)
  } else {
    if (length(clusterNames) == 1) {
      return(rep(0, length(confMats)))
    }
    clusPairs <- utils::combn(clusterNames, 2)
  }

  # For every pair of cluster in that list, compute how the Metric change if
  # we merge
  if (metric == "ARI") {
    deltaMetric <- apply(clusPairs, 2, .localARI, confMats = confMats, C = C) 
  } else if (metric == "NMI"){
    deltaMetric <- apply(clusPairs, 2, .localNMI, confMats = confMats, C = C) 
  }

  # Needed for the cases with only two partitions
  if (is.null(dim(deltaMetric))) {
    return(deltaMetric)
  } else {
    return(colMeans(deltaMetric))
  }
}

.Dune <- function(clusMat,
                  unclustered = NULL,
                  verbose = FALSE,
                  parallel = FALSE,
                  BPPARAM = BiocParallel::bpparam(),
                  metric = "NMI"){
  
  if (!metric %in% c("ARI", "NMI")) {
    stop("For now, only the ARI and NMI are accepted as metrics")
  }
  
  # Initialize the values
  ## Unique cluster labels for all partitions
  currentMat <- clusMat
  if (is.null(colnames(clusMat))) colnames(currentMat) <- seq_len(ncol(clusMat))
  clusters <- lapply(as.data.frame(currentMat, stringsAsFactors = FALSE),
                     unique)
  pairPartitions <- utils::combn(colnames(currentMat), 2)
  # All confusion matrices with associated partitions and Metrics
  confMats <- .confusionMatrices(pairPartitions, currentMat, metric = metric)

  # This is used to keep track of all the info
  working <- TRUE
  merges <- NULL
  ImpMetric <- NULL

  # Try to see if any merge would increse
  while (working) {
    # For every partition
    if (parallel) {
      mergeResults <- BiocParallel::bplapply(
        colnames(currentMat), .findMergeResults, clusters = clusters,
        unclustered = unclustered, confMats = confMats, BPPARAM = BPPARAM,
        metric = metric
      )
    } else {
      mergeResults <- lapply(colnames(currentMat), .findMergeResults,
                             clusters = clusters, unclustered = unclustered,
                             confMats = confMats, metric = metric)
    }

    # Find best pair to merge
    maxs <- sapply(mergeResults, max)

    # Only merge if it improves Metric
    if (max(maxs) > 0) {
      # Find the partition where we merge
      clusLabel <- colnames(currentMat)[which.max(maxs)]
      # Find the two clusters to merge
      clusterNames <- clusters[[clusLabel]]
      if (!is.null(unclustered)) {
        clusPairs <- utils::combn(clusterNames[clusterNames != unclustered], 2)
      } else {
        clusPairs <- utils::combn(clusterNames, 2)
      }
      pair <- clusPairs[, which.max(mergeResults[[which.max(maxs)]])]
      # Find which cells belonged to that cluster
      indsPair <- which(currentMat[, clusLabel] %in% pair)
      # Replace their cluster label with a new label
      currentMat[indsPair, clusLabel] <- min(pair)
      # Remove the old cluster from the list
      clusters[[clusLabel]] <-
        clusters[[clusLabel]][clusters[[clusLabel]] != max(pair)]

      # Update the confusion matrices
      confMats <- .updatedConfMats(pair, confMats, clusLabel, metric)
      # Tracking
      merges <- rbind(merges, c(clusLabel, pair))
      ImpMetric <- c(ImpMetric, max(maxs))
      if (verbose) {
        print(c(clusLabel, pair))
      }
    } else {
      working <- FALSE
    }

    # If no more to merge in any of them, stop
    if (any(sapply(clusters, length) == 1)) {
      if (verbose) {
        print("We merge one of the partition entirely")
      }
      break()
    }
  }

  if (is.null(merges)) {
    stop("This resulted in no merges. Check input cluster labels")
  }
  colnames(merges) <- c("clusteringLabel", "cluster1", "cluster2")
  merger <- list(
    "initialMat" = clusMat,
    "currentMat" = as.data.frame(currentMat, stringsAsFactors = FALSE),
    "merges" = as.data.frame(merges, stringsAsFactors = FALSE),
    "ImpMetric" = ImpMetric,
    metric = metric
  )
  if (is.null(colnames(clusMat))) {
    colnames(merger$initialMat) <- colnames(currentMat)
  }
  return(merger)
}

#' Perform the Metric merging
#'
#' Compute the Metric between every pair of clustering labels after merging every possible pair of clusters. Find the one that improves the Metric merging the most, merge the pair. Repeat until there is no improvement.
#' @param clusMat the matrix of samples by clustering labels.
#' @param cluster_columns if \code{clusMat} is a \code{\link{SummarizedExperiment}},
#'  then this defines the columns of \code{colData} that are outputs from a clustering algorithm.
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @param parallel Logical, defaults to FALSE.
#' Set to TRUE if you want to parallellize the fitting.
#' @param BPPARAM object of class \code{bpparamClass} that specifies the
#'   back-end to be used for computations.
#'   See \code{bpparam} in \code{BiocParallel} package for details.
#'   Won't be used if \code{parallel} is FALSE.
#' @param verbose Whether or not the print cluster merging as it happens.
#' @param metric The metric that is tracked to decide which clusters to merge. For now,
#' either ARI and NMI are accepted. Default to NMI. See details. 
#' @return A list with four components: the initial matrix of clustering labels,
#'  the final matrix of clustering labels, the merge info matrix and the Metric
#'  improvement vector.
#' @details The Dune algorithm merges pairs of clusters in order to improve the
#' mean adjusted Rand Index or the mean normalized mutual information with other
#' clustering labels. It returns a list with five components.:
#'  #' \itemize{
#'   \item \code{initialMat}: The initial matrix of cluster labels
#'   \item \code{currentMat}: The final matrix of cluster labels
#'   \item \code{merges}: The step-by-step detail of the merges, recapitulating
#'   which clusters where merged in which cluster label
#'   \item \code{impMetric}: How much each merge improved the mean Metric between the
#'   cluster label that has been merged and the other cluster labels.
#'   \item \code{metric}: The metric that was used to find the merges.
#' }
#' @seealso clusterConversion ARIImp
#' @importFrom BiocParallel bplapply bpparam
#' @import utils
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' # clusters 11 to 14 from cluster label 5 and 3 are subset of cluster 2 from
#' # other cluster labels. Designing cluster 2 as unclustered therefore means we
#' # do fewer merges.
#' merger2 <- Dune(clusMat = clusMat, unclustered = 2)
#' merger$merges
#' merger2$merges
#' @export
#' @rdname Dune
setMethod(f = "Dune",
          signature = c(clusMat = "matrix"),
          definition = function(clusMat,
                                unclustered = NULL,
                                verbose = FALSE,
                                parallel = FALSE,
                                BPPARAM = BiocParallel::bpparam(),
                                metric = "NMI") {
            merger <- .Dune(clusMat = clusMat, unclustered = unclustered,
                            verbose = verbose, BPPARAM = BPPARAM, metric = metric)
            return(merger)
})

#' @rdname Dune
#' @importFrom BiocParallel bplapply bpparam
setMethod(f = "Dune",
          signature = c(clusMat = "data.frame"),
          definition = function(clusMat,
                                unclustered = NULL,
                                verbose = FALSE,
                                parallel = FALSE,
                                BPPARAM = BiocParallel::bpparam(),
                                metric = "NMI") {
            merger <- .Dune(clusMat = clusMat, unclustered = unclustered,
                            verbose = verbose, BPPARAM = BPPARAM, metric = metric)
            return(merger)
          }
)

#' @rdname Dune
#' @importFrom SummarizedExperiment colData
#' @importFrom BiocParallel bplapply bpparam
setMethod(f = "Dune",
          signature = c(clusMat = "SummarizedExperiment"),
          definition = function(clusMat,
                                cluster_columns,
                                unclustered = NULL,
                                verbose = FALSE,
                                parallel = FALSE,
                                BPPARAM = BiocParallel::bpparam(),
                                metric = "NMI") {
            df <- colData(clusMat)
            if (any(!cluster_columns %in% colnames(df))) {
              stop("All elements of cluster_columns should be in colData(clusMat)")
            }
            df <- as.data.frame(df)
            df <- df[,cluster_columns]
            merger <- Dune(clusMat = df, unclustered = unclustered,
                           verbose = verbose, BPPARAM = BPPARAM, metric = metric)
            return(merger)
          }
)

utils::globalVariables(c("methods", "time"))

# Heatmaps ----
.plotMetric <- function(Mat, values = TRUE,  numericalLabels = FALSE) {
  df <- Mat %>% as.data.frame() %>%
    dplyr::mutate(label1 = rownames(Mat)) %>%
    tidyr::gather(key = label2, value = metric, -(ncol(Mat) + 1))
  if (numericalLabels) {
    df <- df %>%
      dplyr::mutate(label1 = as.numeric(label1), label2 = as.numeric(label2))
  }
  p <- ggplot(df, aes(x = label1, y = label2, fill = metric)) +
    geom_tile() +
    scale_fill_gradientn(colours = brewer.pal(9, "Spectral"),
                         limits = c(0, 1)) +
    theme_classic() +
    theme(axis.line = element_blank())
  if (values) {
    p <- p  +
      geom_text(aes(label = round(metric, 2)), size = 4) +
      guides(fill = FALSE)
  }
  return(p)
}

#' Plot an heatmap of the ARI matrix
#'
#' We can compute the ARI between pairs of cluster labels. This function plots
#' a matrix where a cell is the adjusted Rand Index between cluster label of
#' row i and cluster label of column j.
#' @param clusMat The clustering matrix with a row per cell and a column per
#' clustering label type
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @param values Whether to also display the ARI values. Default to TRUE.
#' @param numericalLabels Whether labels are numerical values. Default to FALSE.
#' @return a \code{\link{ggplot}} object
#' @importFrom dplyr mutate
#' @importFrom tidyr gather
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' plotARIs(merger$initialMat)
#' plotARIs(merger$currentMat)
#' @import ggplot2
#' @import RColorBrewer
#' @export
plotARIs <- function(clusMat, unclustered = NULL, values = TRUE,
                     numericalLabels = FALSE) {
  ARI <- ARIs(clusMat, unclustered = unclustered)
  return(.plotMetric(ARI))
}

#' Plot an heatmap of the NMI matrix
#'
#' We can compute the NMI between pairs of cluster labels. This function plots
#' a matrix where a cell is the Normalized Mutual Information between cluster label of
#' row i and cluster label of column j.
#' @param clusMat The clustering matrix with a row per cell and a column per
#' clustering label type
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @param values Whether to also display the ARI values. Default to TRUE.
#' @param numericalLabels Whether labels are numerical values. Default to FALSE.
#' @return a \code{\link{ggplot}} object
#' @importFrom dplyr mutate
#' @importFrom tidyr gather
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat, metric = "NMI")
#' plotNMIs(merger$initialMat)
#' plotNMIs(merger$currentMat)
#' @import ggplot2
#' @import RColorBrewer
#' @export
plotNMIs <- function(clusMat, unclustered = NULL, values = TRUE,
                     numericalLabels = FALSE) {
  NMI <- NMIs(clusMat, unclustered = unclustered)
  return(.plotMetric(NMI))
}

# Evolutions ----

#' Plot the reduction in cluster size for an ARI merging with \code{Dune}
#' @param merger The output from an ARI merging, by calling \code{\link{Dune}}
#' @return a \code{\link{ggplot}} object
#' #' @importFrom dplyr mutate
#' @importFrom tidyr gather
#' @import ggplot2
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' plotPrePost(merger)
#' @export
plotPrePost <- function(merger) {
  pre <- apply(merger$initialMat, 2, function(x) length(unique(x)))
  post <- apply(merger$currentMat, 2, function(x) length(unique(x)))
  df <- data.frame("methods" = names(pre),
                   "before" = pre,
                   "after" = post) %>%
    tidyr::gather(key = "time", value = "Nb", -"methods") %>%
    dplyr::mutate("time" = factor(time, levels = c("before", "after")))
  p <- ggplot(df, aes(x = methods, y = Nb, fill = time)) +
    geom_bar(stat = "identity", position = "dodge") +
    theme_classic() +
    scale_fill_viridis_d(option = "E") +
    labs(x = "Clustering Methods", y = "Number of clusters", fill = "") +
    ggtitle("Reduction in number of clusters with ARI merging")
  return(p)
}

.metricTrend <- function(merger, Imp, metric) {
  baseMat <- merger$initialMat
  n_clus <- lapply(seq_len(nrow(merger$merges)), function(m){
    diff <- rep(0, ncol(baseMat))
    names(diff) <- colnames(merger$initialMat)
    diff[merger$merges[m, 1]] <- -1
    matrix(diff, nrow = 1)
  }) %>%
    do.call('rbind', args = .)
  n_clus <- rbind(apply(baseMat, 2, n_distinct) %>% matrix(data = ., nrow = 1),
                  n_clus)
  n_clus <- apply(n_clus, 2, cumsum)
  colnames(n_clus) <- colnames(baseMat)
  df <- data.frame(step = 0:length(merger$ImpMetric),
                   Metric_Imp = Imp,
                   n_clus) %>%
    tidyr::gather(key = "change", value = "value", -step) %>%
    dplyr::mutate(type = ifelse(change == "Metric_Imp",
                                paste0(metric, " Improvement"),
                                "Nb of clusters"))
  p <- ggplot(df, aes(x = step, y = value)) +
    geom_path(size = 2, aes(group = change, col = change)) +
    facet_wrap(~type, scales = "free") +
    theme_classic() +
    scale_color_brewer(type = "qual") +
    scale_x_continuous(breaks = c(0, length(merger$ImpMetric)),
                       labels = c("Initial", "Final")) +
    labs(y = "Change over merging",
         col = "Type")
  return(p)
}

#' ARI improvement plot
#'
#' A plot to see how ARI improves over merging
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param unclustered The value assigned to unclustered cells. Default to
#' \code{NULL}
#' @return a \code{\link{ggplot}} object
#' @importFrom dplyr mutate
#' @importFrom tidyr gather
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' ARItrend(merger)
#' @export
#' @import ggplot2
ARItrend <- function(merger, unclustered = NULL) {
  ARI <- ARIImp(merger, unclustered = unclustered)
  return(.metricTrend(merger, ARI, metric = "ARI"))
}

#' NMI improvement plot
#'
#' A plot to see how NMI improves over merging
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param unclustered The value assigned to unclustered cells. Default to
#' \code{NULL}
#' @return a \code{\link{ggplot}} object
#' @importFrom dplyr mutate
#' @importFrom tidyr gather
#' @importFrom magrittr %>%
#' @examples
#' data("clusMat", package = "Dune")
#' merger <- Dune(clusMat = clusMat)
#' NMItrend(merger)
#' @export
#' @import ggplot2
NMItrend <- function(merger, unclustered = NULL) {
  NMI <- NMIImp(merger, unclustered = unclustered)
  return(.metricTrend(merger, NMI, metric = "NMI"))
}
#' Plot confusion matrix
#'
#' A plot to visualize how alike two clustering labels are
#' @param x A vector of clustering labels or a matrix of clustering labels. See details.
#' @param y Optional. Another vector of clustering labels
#' @return a \code{\link{ggplot}} object
#' @importFrom dplyr mutate group_by
#' @importFrom tidyr gather
#' @importFrom magrittr %>%
#' @importFrom RColorBrewer brewer.pal
#' @examples
#' data("nuclei", package = "Dune")
#' ConfusionPlot(nuclei[, c("SC3", "Monocle")])
#' @export
#' @import ggplot2

ConfusionPlot <- function(x, y = NULL) {
  if (is.null(y)) {
    y <- x[, 2]
    x <- x[, 1]
  } else {
    if (length(x) != length(y)) {
      stop("x and y must have the same length")
    }
  }
  df <- table(x, y) %>%
    as.data.frame() %>%
    group_by(x) %>%
    mutate(total_x = sum(Freq)) %>%
    group_by(y) %>%
    mutate(total_y = sum(Freq),
           union = total_x + total_y - Freq,
           overlap = Freq / union) %>%
    ungroup() %>%
    arrange(desc(Freq)) %>%
    filter(Freq > 0)
  df$x <- factor(df$x, levels = unique(df$x))
  df$y <- factor(df$y, levels = unique(df$y))
  p <- ggplot(df, aes(x = x, y = y, col = overlap, size = Freq)) +
    geom_point() +
    theme_bw() +
    theme(legend.position = "top",
          rect = element_blank(),
          panel.border = element_blank(),
          legend.box.spacing = unit(0, units = "npc"),
          legend.margin	= margin(r = .1, l = .1, unit = "npc")) +
    NULL +
    labs(col = "Jaccard Index", size = "# of Cells") +
    scale_color_gradientn(colours = brewer.pal(11, "Spectral")) +
    guides(size = guide_legend(title.position = "top", fill = "grey"),
           col = guide_colourbar(title.position = "top",
                                 barwidth = unit(.2, "npc")))
  return(p)
  }

#' Plot the evolution of the ConfusionPlot as merging
#' happens
#'
#' Animated version of \code{\link{ConfusionPlot}}
#' @param merger the result from having run \code{\link{Dune}}
#'  on the dataset
#' @param unclustered The value assigned to unclustered cells. Default to \code{NULL}
#' @param x The name of the first cluster label to plot
#' @param y The name of the second cluster label to plot
#' @param state_length Time between steps. Default to 1. See \code{\link{transition_states}}
#' for details.
#' @return a \code{gganim} object
#' @details See \code{\link{ConfusionPlot}} and \code{\link{animate}}.
#' @importFrom purrr map_df
#' @import tidyr dplyr
#' @importFrom magrittr %>%
#' @importFrom RColorBrewer brewer.pal
#' @import ggplot2
#' @importFrom gganimate transition_states
#' @examples
#' \dontrun{
#'   data("clusMat", package = "Dune")
#'   merger <- Dune(clusMat = clusMat)
#'   ConfusionEvolution(merger, x = "A", y = "B")}
#' @export
ConfusionEvolution <- function(merger, unclustered = NULL, x, y, state_length = 1) {
  Freqs <- purrr::map_df(0:length(merger$ImpMetric), function(step){
    clusMat <- intermediateMat(merger, n_steps = step) %>%
      as.matrix()
    df <- table(x = clusMat[, x], y = clusMat[, y]) %>%
      as.data.frame() %>%
      mutate(x = as.numeric(x),
             y = as.numeric(y)) %>%
      group_by(x) %>%
      mutate(total_x = sum(Freq),
             step = step) %>%
      group_by(y) %>%
      mutate(total_y = sum(Freq),
             union = total_x + total_y - Freq,
             overlap = Freq / union) %>%
      ungroup() %>%
      arrange(desc(Freq)) %>%
      filter(Freq > 0)
  })
  Freqs$x <- factor(Freqs$x, levels = sort(unique(Freqs$x)))
  Freqs$y <- factor(Freqs$y, levels = sort(unique(Freqs$y)))

  p <- ggplot(Freqs, aes(x = x, y = y, col = overlap, size = Freq)) +
    geom_point() +
    theme_bw() +
    theme(legend.position = "top",
          rect = element_blank(),
          panel.border = element_blank(),
          legend.box.spacing = unit(0, units = "npc"),
          legend.margin	= margin(r = .1, l = .1, unit = "npc")) +
    NULL +
    labs(col = "Jaccard Index", size = "# of Cells") +
    scale_color_gradientn(colours = RColorBrewer::brewer.pal(11, "Spectral")) +
    guides(size = guide_legend(title.position = "top", fill = "grey"),
           col = guide_colourbar(title.position = "top",
                                 barwidth = unit(.2, "npc")))
  p <- p +
    gganimate::transition_states(step,
                                 transition_length = 0,
                                 state_length =  state_length / table(Freqs$step)[1]) +
    ggtitle(paste0('Step {closest_state} of ', max(Freqs$step)))
  return(p)
}

context("Dune plotting functions do return plots")
library(Dune)
library(dplyr)

test_that("All static plotting functions return a ggplot object", {
  data("clusMat", package = "Dune")
  data("nuclei", package = "Dune")
  merger <- Dune(clusMat)
  # plotARIs 1
  p <- plotARIs(merger$initialMat)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # plotARIs 2
  p <- plotARIs(merger$initialMat, unclustered = 1)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # plotARIs 3
  mat <- merger$initialMat
  colnames(mat) <- seq_len(ncol(mat))
  p <- plotARIs(mat, numericalLabels = TRUE)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # plotNMIs 1
  p <- plotNMIs(merger$initialMat)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # plotNMIs 2
  p <- plotNMIs(merger$initialMat, unclustered = 1)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # plotNMIs 3
  mat <- merger$initialMat
  colnames(mat) <- seq_len(ncol(mat))
  p <- plotNMIs(mat, numericalLabels = TRUE)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # plotPrePost
  p <- plotPrePost(merger)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # ARItrend
  p <- ARItrend(merger)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # NMItrend
  p <- NMItrend(merger)
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # ConfusionPlot
  p <- ConfusionPlot(nuclei[, c("SC3", "Monocle")])
  expect_is(p, "gg")
  expect_error(print(p), NA)
  # Error
  expect_error(ConfusionPlot(nuclei$SC3[1:20], nuclei$Monocle[1:10]))
})

test_that("All dynamic plotting functions return a gganim object", {
  data("clusMat", package = "Dune")
  data("nuclei", package = "Dune")
  merger <- Dune(clusMat)
  # Confusion Evolution
  p <- ConfusionEvolution(merger, x = "A", y = "B")
  expect_is(p, "gganim")
})

context("Dune works as expected for all inputs")
library(Dune)
library(SummarizedExperiment)

test_that("Dune reacts similarly to all kind of inputs", {
  data("clusMat", package = "Dune")
  merger_matrix <- Dune(clusMat)
  merger_df <- Dune(as.data.frame(clusMat))
  sce <- SummarizedExperiment::SummarizedExperiment(assays = list(counts = t(clusMat)),
                                                    colData = clusMat)
  merger_sExp <- Dune(clusMat = sce, cluster_columns = colnames(clusMat))
  expect_identical(merger_matrix$initialMat, as.matrix(merger_df$initialMat))
  expect_identical(merger_matrix$initialMat, as.matrix(merger_sExp$initialMat))
  expect_identical(merger_matrix[2:4], merger_df[2:4])
  expect_identical(merger_matrix[2:4], merger_sExp[2:4])
})

test_that("Dune fails when wrong column names are specified", {
  data("clusMat", package = "Dune")
  sce <- SummarizedExperiment::SummarizedExperiment(assays = list(counts = t(clusMat)),
                                                    colData = clusMat)
  expect_error(Dune(clusMat = sce, cluster_columns = c("A", "F")))
})

test_that("Dune fails when the wrong metric is specified", {
  data("clusMat", package = "Dune")
  expect_error(Dune(clusMat = clusMat, metric = "AEIHn"))
})

#!/bin/bash

version=1.7.1

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf re
mkdir -p re/src

(cd $TMP && opam source re.$version)

SRC=$TMP/re.$version

cp -v $SRC/LICENSE re

for m in re re_automata re_cset re_fmt; do
    for ext in ml mli; do
        fn=$SRC/lib/$m.$ext
        [[ -f $fn ]] && cp -v $fn re/src
    done
done

git checkout re/src/jbuild
git add -A .

REM Download and install OCaml and flexlink (unless it was already done).
REM Prepare the environment variables,... to use it.  OCaml is installed
REM at %OCAMLROOT%
REM
REM If you are using Cygwin, install it in C:\cygwin first and then
REM execute this script.  Execute bash with the option "-l".

REM set OCAMLROOT=%PROGRAMFILES%/OCaml
set OCAMLROOT=C:/PROGRA~1/OCaml

set OCAMLURL=https://github.com/Chris00/ocaml-appveyor/releases/download/0.2/ocaml-4.06.zip

REM Cygwin is always installed on AppVeyor.  Its path must come
REM before the one of Git but after those of MSCV and OCaml.
set Path=C:\cygwin\bin;%Path%

call "C:\Program Files\Microsoft SDKs\Windows\v7.1\Bin\SetEnv.cmd" /x64

set Path=%OCAMLROOT%\bin;%OCAMLROOT%\bin\flexdll;%Path%
set CAML_LD_LIBRARY_PATH=%OCAMLROOT%/lib/stublibs

set CYGWINBASH=C:\cygwin\bin\bash.exe

if not exist "%OCAMLROOT%/bin/ocaml.exe" (
  echo Downloading OCaml...
  appveyor DownloadFile "%OCAMLURL%" -FileName "C:\PROGRA~1\ocaml.zip"
  %CYGWINBASH% -lc "cd /cygdrive/c/Program\ Files && unzip -q ocaml.zip"
  del C:\PROGRA~1\ocaml.zip
)

if exist %CYGWINBASH% (
  REM Make sure that "link" is the MSVC one and not the Cynwin one.
  %CYGWINBASH% -lc "eval $(/cygdrive/c/Program\ Files/OCaml/tools/msvs-promote-path)"> ~/.bash_profile"
)

set <NUL /p=Ready to use OCaml & ocamlc -version

export DUNE_PKG_OVERRIDE_OCAML=1

dune="dune"

pkg_root="_build/_private/default/.pkg"

build_pkg() {
  $dune build _build/_private/default/.pkg/$1/target/
}

show_pkg() {
  find $pkg_root/$1 | sort | sed "s#$pkg_root/$1##"
}

show_pkg_targets() {
  find $pkg_root/$1/target | sort | sed "s#$pkg_root/$1/target##"
}

show_pkg_cookie() {
  $dune internal dump $pkg_root/$1/target/cookie
}

mkrepo() {
  mkdir -p mock-opam-repository
  cat >mock-opam-repository/repo <<EOF
opam-version: "2.0"
EOF
}

mock_packages="mock-opam-repository/packages"

mkpkg() {
  name=$1
  if [ "$#" -eq "1" ]
  then
    version="0.0.1"
  else
    version="$2"
  fi
  mkdir -p $mock_packages/$name/$name.$version
  cat >$mock_packages/$name/$name.$version/opam
}

solve_project() {
  cat >dune-project
  dune pkg lock --opam-repository-path=mock-opam-repository
}

make_lockdir() {
mkdir dune.lock
cat >dune.lock/lock.dune <<EOF
(lang package 0.1)
EOF
}

#!/bin/bash

version=1.0.0

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf cmdliner
mkdir -p cmdliner/src

(cd $TMP && opam source cmdliner.$version)

SRC=$TMP/cmdliner.$version

cp -v $SRC/LICENSE.md cmdliner
cp -v $SRC/src/*.{ml,mli} cmdliner/src

git checkout cmdliner/src/jbuild
git add -A .

export XDG_CACHE_HOME="$PWD/.cache"

dune="dune"

pkg_root="_build/_private/default/.pkg"

build_pkg() {
  $dune build $pkg_root/$1/target/
}

show_pkg() {
  find $pkg_root/$1 | sort | sed "s#$pkg_root/$1##"
}

strip_sandbox() {
  sed -E 's#[^ ]*.sandbox/[^/]+#$SANDBOX#g'
}

show_pkg_targets() {
  find $pkg_root/$1/target | sort | sed "s#$pkg_root/$1/target##"
}

show_pkg_cookie() {
  $dune internal dump $pkg_root/$1/target/cookie
}

mock_packages="mock-opam-repository/packages"

mkrepo() {
  mkdir -p $mock_packages
}

mkpkg() {
  name=$1
  if [ "$#" -eq "1" ]
  then
    version="0.0.1"
  else
    version="$2"
  fi
  mkdir -p $mock_packages/$name/$name.$version
  echo 'opam-version: "2.0"' > $mock_packages/$name/$name.$version/opam
  cat >>$mock_packages/$name/$name.$version/opam
}

add_mock_repo_if_needed() {
  if [ ! -e dune-workspace ]
  then
      cat >dune-workspace <<EOF
(lang dune 3.10)
(lock_dir
 (repositories mock))
(repository
 (name mock)
 (source "file://$(pwd)/mock-opam-repository"))
EOF
  else
    if ! grep '(name mock)' > /dev/null dune-workspace
    then
      # add the repo definition
      cat >>dune-workspace <<EOF
(repository
 (name mock)
 (source "file://$(pwd)/mock-opam-repository"))
EOF
 
      # reference the repo
      if grep -s '(repositories'
      then
        sed -i '' -e 's/(repositories \(.*\))/(repositories mock \1)/' dune-workspace
      else
        cat >>dune-workspace <<EOF
(lock_dir
 (repositories mock))
EOF
      fi
  
    fi
  fi
}

make_lockpkg() {
  local dir="dune.lock"
  mkdir -p $dir
  local f="$dir/$1.pkg"
  cat >$f
}

solve_project() {
  cat >dune-project
  add_mock_repo_if_needed
  dune pkg lock $@
}

make_lockdir() {
  mkdir -p dune.lock
  cat >dune.lock/lock.dune <<EOF
(lang package 0.1)
(repositories (complete true))
EOF
}

make_project() {
  cat <<EOF
(lang dune 3.11)
 (package
  (name x)
  (allow_empty)
  (depends $@))
EOF
}

print_source() {
  cat dune.lock/$1.pkg | sed -n "/source/,//p" | sed "s#$PWD#PWD#g" | tr '\n' ' '| tr -s " "
}

solve() {
  make_project $@ | solve_project
}

#!/bin/sh

# This script is used for scrubbing the output of Dune actions calling Coq
# binaries. It contains some common patterns which are scurbbed for any
# non-reproducible paths and the arguments are displayed line by line.

# \f (form feed) is used as a delayed new line for sed

sed 's/$ (cd .*coqc/coqc/' $1 | # cleanup coqc
sed 's/$ (cd .*coqdep/coqdep/' | # cleanup coqdep
sed 's/$ (cd .*coqdoc/coqdoc/' | # cleanup coqdoc
sed 's/)//' | # remove trailing )
sed 's/ -I/\f-I/g' | # new line for each -I
sed 's/ -nI/\f-nI/g' | # new line for each -nI
sed 's/ -R/\f-R/g' | # new line for each -R
sed 's/ -Q/\f-Q/g' | # new line for each -Q
sed 's/ -w/\f-w/g' | # new line for each -w
sed 's/ -boot/\f-boot/g' | # new line for each -boot
sed 's/ -native-compiler /\f-native-compiler /g' | # new line for each -native-compiler
sed 's/-R [^[:space:]]*coq /-R coq /g' | # scrub -R with coq
sed 's#-R [^[:space:]]*coq/theories #-R coq/theories #g' | # scrub -R with coq/theories
sed 's/-I [-A-Za-z0-9\/_\.]*lib\//-I lib\//g' | # scrub -I with lib/
sed 's/-nI [-A-Za-z0-9\/_\.]*lib\//-nI lib\//g' | # scrub -nI with lib/
sed 's/\(-R [^\f]* [^\f]*\) \(.*\)/\1\f\2/g' | # new line after each -R
sed 's/\(-I lib\/findlib\)\f\(.*\)/\2\f\1/g' | # move lib/findlib to the end
sed 's/\(-I lib\/zarith\)\f\(.*\)/\2\f\1/g' | # move lib/zarith to the end
tr '\f' '\n' # replace form feed with new line

#!/bin/bash

version=2.0.0~beta

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf opam-file-format
mkdir -p opam-file-format/src

(cd $TMP && opam source opam-file-format.$version)

SRC=$TMP/opam-file-format.$version

cp -v $SRC/src/*.{ml,mli,mll,mly} opam-file-format/src

git checkout opam-file-format/src/jbuild
git add -A .

#!/bin/bash

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf usexp
mkdir -p usexp/src

(cd $TMP && git clone https://github.com/janestreet/usexp.git)

SRC=$TMP/usexp

cp -v $SRC/LICENSE.md usexp

cp -v $SRC/src/*.{ml,mli} usexp/src

git checkout usexp/src/jbuild
git add -A .

#!/bin/bash

version=3c446b45b2d9e81bc72b57ada168fe7923f9b02c

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf lwd
mkdir -p lwd/lwd lwd/nottui

(
    cd $TMP
    git clone https://github.com/let-def/lwd.git
    cd lwd
    git checkout $version
)

SRC=$TMP/lwd/

cp -v $SRC/LICENSE lwd/
cp -v -R $SRC/lib/lwd/lwd.{ml,mli} lwd/lwd
cp -v -R $SRC/lib/lwd/lwd_utils.{ml,mli} lwd/lwd
cp -v -R $SRC/lib/nottui/nottui.{ml,mli} lwd/nottui


git checkout lwd/{lwd,nottui}/dune
git add -A .

DUNE_RUNNING=0

start_dune () {
    ((dune build "$@" --passive-watch-mode > .#dune-output 2>&1) || (echo exit $? >> .#dune-output)) &
    DUNE_PID=$!;
    DUNE_RUNNING=1;
}

timeout="$(command -v timeout || echo gtimeout)"

with_timeout () {
    $timeout 2 "$@"
    exit_code=$?
    if [ "$exit_code" = 124 ]
    then
        echo Timed out
        cat .#dune-output
    else
        return "$exit_code"
    fi
}

stop_dune () {
    with_timeout dune shutdown;
    # On Linux, we may run into a bash pid aliasing bug that causes wait to
    # reject the pid. Therefore we use tail to wait instead.
    if [ "$(uname -s)" = "Linux" ]
    then
        # wait for all child processes
        tail --pid=$DUNE_PID -f /dev/null;
    else
        # wait for dune to exit
        wait $DUNE_PID;
    fi
    cat .#dune-output;
}

build () {
    with_timeout dune rpc build --wait "$@"
}

#!/bin/bash

version=1.11.0

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf re
mkdir -p re/src

(
    cd $TMP
    git clone https://github.com/ocaml/ocaml-re.git
    cd ocaml-re
    git checkout $version
)

SRC=$TMP/ocaml-re

cp -v $SRC/LICENSE.md re
cp -v $SRC/lib/*.{ml,mli} re/src/

echo "include Re" > re/src/dune_re.ml
cat >re/src/dune <<EOF
(library
 (name dune_re)
 (public_name dune-private-libs.dune_re)
 (synopsis "Internal Dune library, do not use!"))
EOF

git add -A .

#!/bin/bash

version=1e2030a5d5183d84561cde142eecca40e03db2a3

set -e -o pipefail

TMP="$(mktemp -d)"

rm -rf incremental-cycles
mkdir incremental-cycles

(
    cd $TMP
    git clone https://gitlab.inria.fr/agueneau/incremental-cycles.git
    cd incremental-cycles
    git checkout $version
    dune build export/incremental_cycles.ml
)

SRC=$TMP/incremental-cycles

mkdir incremental-cycles/src
cp -v $SRC/LICENSE incremental-cycles
cp -v $SRC/_build/default/export/incremental_cycles.ml incremental-cycles/src/
cp -v $SRC/export/incremental_cycles_intf.ml incremental-cycles/src/
cp -v $SRC/export/incremental_cycles.mli incremental-cycles/src/

git checkout incremental-cycles/src/dune
git add -A .

#!/bin/bash

version=1bfe079ddfc6bff72a3b6a8ae3c0408297b04434

set -e -u -o pipefail

TMP="$(mktemp -d)"
TRAP_CMD=$(printf "rm -rf %q" "$TMP")
trap "$TRAP_CMD" EXIT

lib_name=ocaml-inotify

rm -rf "$lib_name"
mkdir -p "$lib_name/src"

(
    cd "$TMP"
    git clone https://github.com/ocaml-dune/ocaml-inotify.git "$lib_name"
    cd "$lib_name"
    git -c advice.detachedHead=false checkout "$version"
)

SRC=$TMP/$lib_name

rm "$SRC"/lib/lwt_inotify.ml{,i}
cp -v "$SRC"/lib/*.{ml,mli,c} "$lib_name"/src
cp -v "$SRC"/lib/dune "$lib_name"/src
cp -v "$SRC"/LICENSE.txt "$lib_name"/

git add -A .

#!/bin/bash

version=1.15.4
name=sha

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf $name
mkdir -p $name/src

(cd $TMP && opam source $name.$version)

SRC=$TMP/$name.$version

cp -v $SRC/*.{ml,mli,c,h} $name/

git checkout $name/dune
git add -A $name/

#!/bin/sh
set -euf

# shellcheck disable=SC2154
echo "
=============
build-test.sh
=============
.
------
Matrix
------
dkml_host_abi=$dkml_host_abi
abi_pattern=$abi_pattern
opam_root=$opam_root
exe_ext=${exe_ext:-}
.
"

# Set project directory
if [ -n "${CI_PROJECT_DIR:-}" ]; then
    PROJECT_DIR="$CI_PROJECT_DIR"
elif [ -n "${PC_PROJECT_DIR:-}" ]; then
    PROJECT_DIR="$PC_PROJECT_DIR"
elif [ -n "${GITHUB_WORKSPACE:-}" ]; then
    PROJECT_DIR="$GITHUB_WORKSPACE"
else
    PROJECT_DIR="$PWD"
fi
if [ -x /usr/bin/cygpath ]; then
    PROJECT_DIR=$(/usr/bin/cygpath -au "$PROJECT_DIR")
fi

# PATH. Add opamrun
export PATH="$PROJECT_DIR/.ci/sd4/opamrun:$PATH"

# Initial Diagnostics (optional but useful)
opamrun switch
opamrun list
opamrun var
opamrun config report
opamrun option
opamrun exec -- ocamlc -config

# Update
opamrun update

# ------------------
# Dune's build logic.
#   For now it mimics the "build" job in .github/workflows/workflow.yml
# ------------------

echo ======== Unpin Dune from DKML defaults
opamrun pin list
opamrun pin remove dune --no-action
opamrun pin remove dune-configurator --no-action

echo ======== Set git user
git config --global user.name 'github-actions[bot]'
git config --global user.email 'github-actions[bot]@users.noreply.github.com'

echo ======== Install ocamlfind-secondary and ocaml-secondary-compiler, if needed
opamrun install ./dune.opam --deps-only --with-test

case "$dkml_host_abi" in
darwin_*)
    echo ======== Install system deps on macOS
    brew install coreutils
    ;;
esac

# dune doesn't have any additional dependencies so we can build it right
# away this makes it possible to see build errors as soon as possible
echo ======== Build boot dune.exe
opamrun exec -- make _boot/dune.exe

echo ======== Ensure Dune can build itself
opamrun exec -- make bootstrap

case "$dkml_host_abi" in
windows_*)
    # Mitigate bug in opam:
    #
    #    #=== ERROR while compiling dune.3.6.0 =========================================#
    #    # path        Z:\source\dune
    #    # command     C:\Users\beckf\AppData\Local\Programs\DiskuvOCaml\tools\MSYS2\usr\bin\cp.exe -PRp Z:\source\dune\.ci\o\dkml\.opam-switch\sources\dune.3.6.0 Z:\source\dune\.ci\o\dkml\.opam-switch\build\dune.3.6.0
    #    # exit-code   1
    #    # env-file    Z:\source\dune\.ci\o\log\log-685364-345b91.env
    #    # output-file Z:\source\dune\.ci\o\log\log-685364-345b91.out
    #    ### output ###
    #    # /usr/bin/cp: cannot create regular file 'Z:\source\dune\.ci\o\dkml\.opam-switch\build\dune.3.6.0/dune': File exists
    #
    # Basically the 'cp' fallback logic in opam's src/core/opamSystem.ml:copy_dir is broken.
    # Using 'rsync' avoids that, so install it on MSYS2.
    if command -v pacman; then
        echo ======== Install rsync on Win32
        pacman --sync --needed --noconfirm rsync
    fi

    echo ======== Install deps on Win32
    opamrun install ./dune-configurator.opam --deps-only --with-test --yes

    echo ======== Run test suite on Win32
    opamrun exec -- make test-windows
    ;;
*)
    echo ======== Install deps on Unix
    opamrun install . --deps-only --with-test --yes

    # Remove DKML pins (keep alphabetical; the pins will eventually be upstreamed in DKML)
    opamrun pin remove base --no-action
    opamrun pin remove core_kernel --no-action
    opamrun pin remove ocamlformat --no-action
    opamrun pin remove ppx_expect --no-action
    opamrun pin remove time_now --no-action

    # This should have been:
    #   opam exec -- make dev-deps
    # but melange requires ocaml 4.14 (DKML hasn't been updated as of 2022-11)
    opamrun exec -- make dev-deps-sans-melange

    echo ======== Run test suite on Unix
    # This should have been:
    #   opam exec -- make test
    # but melange requires ocaml 4.14 (DKML hasn't been updated as of 2022-11)
    opamrun exec -- make test-all-sans-melange
esac

echo ======== Build configurator
opamrun install ./dune-configurator.opam --yes

#!/usr/bin/env bash

version=7c05922290adb214f834d9e0f1efe7028c4ebb85

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf opam
mkdir -p opam/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/opam.git
    cd opam
    git checkout $version
)

SRC=$TMP/opam

cat > $TMP/local-opam-changes.diff << "EOF"
diff --git a/vendor/opam/src/core/opamCoreConfigDeveloper.ml b/vendor/opam/src/core/opamCoreConfigDeveloper.ml
new file mode 100644
index 000000000..cf62a7fd2
--- /dev/null
+++ b/vendor/opam/src/core/opamCoreConfigDeveloper.ml
@@ -0,0 +1 @@
+let value = ""
diff --git a/vendor/opam/src/core/opamProcess.ml b/vendor/opam/src/core/opamProcess.ml
index 764ff18fd..ef8734625 100644
--- a/vendor/opam/src/core/opamProcess.ml
+++ b/vendor/opam/src/core/opamProcess.ml
@@ -9,6 +9,8 @@
 (*                                                                        *)
 (**************************************************************************)
 
+module Re = Dune_re
+
 let log ?level fmt =
   OpamConsole.log "PROC" ?level fmt
 
diff --git a/vendor/opam/src/core/opamStd.ml b/vendor/opam/src/core/opamStd.ml
index e04c84379..454702714 100644
--- a/vendor/opam/src/core/opamStd.ml
+++ b/vendor/opam/src/core/opamStd.ml
@@ -9,6 +9,8 @@
 (*                                                                        *)
 (**************************************************************************)
 
+module Re = Dune_re
+
 module type SET = sig
   include Set.S
   val map: (elt -> elt) -> t -> t
diff --git a/vendor/opam/src/core/opamStd.mli b/vendor/opam/src/core/opamStd.mli
index 9133f9e16..5fd460f06 100644
--- a/vendor/opam/src/core/opamStd.mli
+++ b/vendor/opam/src/core/opamStd.mli
@@ -278,7 +278,7 @@ module String : sig
   val for_all: (char -> bool) -> string -> bool
   val contains_char: string -> char -> bool
   val contains: sub:string -> string -> bool
-  val exact_match: Re.re -> string -> bool
+  val exact_match: Dune_re.re -> string -> bool
   val find_from: (char -> bool) -> string -> int -> int
 
   (** Like [String.compare], but with lowercase/uppercase variants ordered next
diff --git a/vendor/opam/src/core/opamUrl.ml b/vendor/opam/src/core/opamUrl.ml
index 16baf4c75..8234b3419 100644
--- a/vendor/opam/src/core/opamUrl.ml
+++ b/vendor/opam/src/core/opamUrl.ml
@@ -11,6 +11,8 @@
 
 open OpamStd.Op
 
+module Re = Dune_re
+
 type version_control = [ `git | `darcs | `hg ]
 
 type backend = [ `http | `rsync | version_control ]
diff --git a/vendor/opam/src/core/opamVersionInfo.ml b/vendor/opam/src/core/opamVersionInfo.ml
new file mode 100644
index 000000000..6c0168d9b
--- /dev/null
+++ b/vendor/opam/src/core/opamVersionInfo.ml
@@ -0,0 +1 @@
+let version = "2.2.0~alpha-vendored"
diff --git a/vendor/opam/src/format/opamFile.ml b/vendor/opam/src/format/opamFile.ml
index 03611c8e8..689cac471 100644
--- a/vendor/opam/src/format/opamFile.ml
+++ b/vendor/opam/src/format/opamFile.ml
@@ -18,6 +18,7 @@
       a string list list. These are mostly used internally
     - files using the "opam syntax" and lexer, parsed using OpamFormat.Pp.V
 *)
+module Re = Dune_re
 
 open OpamParserTypes.FullPos
 open OpamTypes
diff --git a/vendor/opam/src/format/opamFilter.ml b/vendor/opam/src/format/opamFilter.ml
index af7374e4b..d60abc796 100644
--- a/vendor/opam/src/format/opamFilter.ml
+++ b/vendor/opam/src/format/opamFilter.ml
@@ -8,6 +8,7 @@
 (*  exception on linking described in the file LICENSE.                   *)
 (*                                                                        *)
 (**************************************************************************)
+module Re = Dune_re
 
 open OpamTypes
 open OpamTypesBase
diff --git a/vendor/opam/src/format/opamFormula.ml b/vendor/opam/src/format/opamFormula.ml
index 6931055e4..277b17963 100644
--- a/vendor/opam/src/format/opamFormula.ml
+++ b/vendor/opam/src/format/opamFormula.ml
@@ -9,6 +9,8 @@
 (*                                                                        *)
 (**************************************************************************)
 
+module Re = Dune_re
+
 type relop = [`Eq|`Neq|`Geq|`Gt|`Leq|`Lt]
 
 let neg_relop = function
diff --git a/vendor/opam/src/format/opamSwitch.ml b/vendor/opam/src/format/opamSwitch.ml
index c6704f911..6dbd58e96 100644
--- a/vendor/opam/src/format/opamSwitch.ml
+++ b/vendor/opam/src/format/opamSwitch.ml
@@ -9,6 +9,8 @@
 (*                                                                        *)
 (**************************************************************************)
 
+module Re = Dune_re
+
 include OpamStd.AbstractString
 
 let unset = of_string "#unset#"
diff --git a/vendor/opam/src/state/opamEnv.ml b/vendor/opam/src/state/opamEnv.ml
index 8913edda0..044835ad0 100644
--- a/vendor/opam/src/state/opamEnv.ml
+++ b/vendor/opam/src/state/opamEnv.ml
@@ -15,6 +15,7 @@ open OpamStateTypes
 open OpamTypesBase
 open OpamStd.Op
 open OpamFilename.Op
+module Re = Dune_re
 
 let log fmt = OpamConsole.log "ENV" fmt
 let slog = OpamConsole.slog
diff --git a/vendor/opam/src/state/opamFileTools.ml b/vendor/opam/src/state/opamFileTools.ml
index acba124a7..0e6f051ac 100644
--- a/vendor/opam/src/state/opamFileTools.ml
+++ b/vendor/opam/src/state/opamFileTools.ml
@@ -12,6 +12,7 @@
 open OpamParserTypes.FullPos
 open OpamTypes
 open OpamTypesBase
+module Re = Dune_re
 
 let log ?level fmt = OpamConsole.log "opam-file" ?level fmt
 
diff --git a/vendor/opam/src/state/opamSysInteract.ml b/vendor/opam/src/state/opamSysInteract.ml
index 2857654a9..53f11d870 100644
--- a/vendor/opam/src/state/opamSysInteract.ml
+++ b/vendor/opam/src/state/opamSysInteract.ml
@@ -8,6 +8,8 @@
 (*                                                                        *)
 (**************************************************************************)
 
+module Re = Dune_re
+
 let log fmt = OpamConsole.log "XSYS" fmt
 
 (* Run commands *)

diff --git a/vendor/opam/src/format/opamFilter.mli b/vendor/opam/src/format/opamFilter.mli
index a2f01d179..1787b24b1 100644
--- a/vendor/opam/src/format/opamFilter.mli
+++ b/vendor/opam/src/format/opamFilter.mli
@@ -8,6 +8,7 @@
 (*  exception on linking described in the file LICENSE.                   *)
 (*                                                                        *)
 (**************************************************************************)
+module Re = Dune_re

 (** Formulas on variables, as used in opam files build scripts

EOF

for subpackage in core repository format state
do
    PKG=opam/src/$subpackage/
    mkdir -p $PKG
    set +e
    cp -v $SRC/src/$subpackage/*.{ml,mli,mll,mly,c} $PKG
    set -e
    git checkout $PKG/dune
done

git apply $TMP/local-opam-changes.diff
git add -A opam/

#!/bin/sh

version=0021887f75a6017486908cfd837b2ee126e063e7

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf pp
mkdir -p pp/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/pp.git
    cd pp
    git checkout $version
)

SRC=$TMP/pp

cp -v $SRC/src/pp.{ml,mli} pp/src
cp -v $SRC/LICENSE.md pp/

git checkout pp/src/dune
git checkout pp/LICENSE.md
git add -A .

#!/bin/bash

# CR-someday jdimino: maybe it's possible to get cmdliner to print that directly

set -e -o pipefail

CMDS=$(dune --help=plain | \
           sed -n '/COMMAND ALIASES/,/COMMON OPTIONS/p' | sed -En 's/^       ([a-z-]+) ?.*/\1/p')

for cmd in $CMDS; do
    cat <<EOF

(rule
 (with-stdout-to dune-$cmd.1
  (run dune $cmd --help=groff)))

(install
 (section man)
 (package dune)
 (files   dune-$cmd.1))
EOF
done

echo

# Create a dune-workspace file with mock repos set up for the main
# project and the odoc lockdir.
setup_odoc_workspace() {
  cat > dune-workspace <<EOF
(lang dune 3.16)
(lock_dir
 (path "dev-tools.locks/odoc")
 (repositories mock))
 (lock_dir
  (repositories mock))
(repository
 (name mock)
 (source "file://$(pwd)/mock-opam-repository"))
EOF
}

# Create a fake odoc package containing an executable that
# just prints a message.
make_mock_odoc_package() {
  mkpkg odoc <<EOF
install: [
  [ "sh" "-c" "echo '#!/bin/sh' > %{bin}%/odoc" ]
  [ "sh" "-c" "echo 'echo hello from fake odoc' >> %{bin}%/odoc" ]
  [ "sh" "-c" "chmod a+x %{bin}%/odoc" ]
]
EOF
}

#!/bin/bash

version=8474b2a29c4cb9cbf356006612c7b2b97e9c4087

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

PACKAGE=uutf

rm -rf $PACKAGE
mkdir -p $PACKAGE/src

(
    cd $TMP
    git clone https://github.com/dbuenzli/$PACKAGE.git
    cd $PACKAGE
    git checkout $version
    cd src
)

SRC=$TMP/$PACKAGE

cp -v $SRC/LICENSE.md $PACKAGE
cp -v $SRC/src/*.{ml,mli} $PACKAGE/

git checkout $PACKAGE/dune
git add -A .

#!/usr/bin/env bash

# Run this script simply as ./bench/perf.sh from the root directory.

set -e

TEST_REPO=https://github.com/ocaml-dune/dune-bench
TEST_COMMIT=b6bfaf2974ec8ee1eea92c4316ec37b9966322e3

# Some alternative benchmarks:

# TEST_REPO=https://github.com/ocaml/dune
# TEST_COMMIT=002edc11f4e0a57f11d5226cb2497c8b406027b5

# TEST_REPO=https://github.com/avsm/platform
# TEST_COMMIT=b254e3c6b60f3c0c09dfdcde92eb1abdc267fa1c

dune() {
  TIMEFORMAT=$'real %Rs\nuser %Us\nsys  %Ss\n'; time ../_build/default/bin/main.exe "$@" > /dev/null 2>&1
}

setup_test() {
  mkdir -p _perf

  cd _perf
  if [ ! -f README.md ]; then
    echo "Cloning $TEST_REPO..."
    wget $TEST_REPO/archive/$TEST_COMMIT.tar.gz
    tar -xzf $TEST_COMMIT.tar.gz --strip-components=1
  fi
  cd ..
}

pad () {
  while IFS='' read -r x; do printf "%-$1s\n" "$x"; done
}

run_test() {
  echo "Building Dune..."
  # [make release] is used for bootstrapping, but the real binary to benchmark is
  # then produced by a separate dune invocation.
  # This is done mainly because [make release] won't rebuild dune if it's stale.
  make release > /dev/null
  ./dune.exe build _build/default/bin/main.exe

  cd _perf
  rm -rf _build

  echo "Running full build..."
  dune build --release --cache=disabled 2>> $1

  echo "Running zero build..."
  dune build --release --cache=disabled 2>> $1

  cd ..
}

setup_test
CURRENT_BRANCH=$(git branch | sed -n -e 's/^\* \(.*\)/\1/p')

rm -f _perf/rows _perf/current _perf/main

echo "            " >> _perf/rows
echo "            " >> _perf/rows
echo "           |" >> _perf/rows
echo "Full build |" >> _perf/rows
echo "           |" >> _perf/rows
echo "            " >> _perf/rows
echo "           |" >> _perf/rows
echo "Zero build |" >> _perf/rows
echo "           |" >> _perf/rows

echo "Current branch" >> _perf/current
echo "==============" >> _perf/current
echo "Testing the current branch ($CURRENT_BRANCH)"
run_test current

echo " Main branch " >> _perf/main
echo "=============" >> _perf/main
git checkout main
echo "Testing main"
run_test main

git checkout $CURRENT_BRANCH

echo ""
echo "Summary for building $TEST_REPO:"
echo ""

paste -d ' ' <(pad 10 < _perf/rows) <(pad 14 < _perf/current) _perf/main

#!/bin/bash

version=430664dd13c4645bda0e9c39f7d0ec75c1534874

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf cmdliner
mkdir -p cmdliner/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/cmdliner.git
    cd cmdliner
    git checkout $version
    dune subst
    cd src
)

SRC=$TMP/cmdliner

cp -v $SRC/LICENSE.md cmdliner
cp -v $SRC/src/*.{ml,mli} cmdliner/src

git checkout cmdliner/src/dune
git add -A .

#!/usr/bin/env sh
export BENCHMARKS_RUNNER=TRUE
case "$1" in
  "dune" ) test="dune_bench"; main="main";;
  "memo" ) test="memo_bench"; main="memo_bench_main";;
  "thread_pool" ) test="thread_pool_bench"; main="thread_pool_bench_main";;
  "digest" ) test="digest_bench"; main="digest_bench_main";;
  "path" ) test="path_bench"; main="path_bench_main";;
esac
shift;
export BENCH_LIB="$test"
exec ./dune.exe exec --release -- "./bench/micro/$main.exe" -fork -run-without-cross-library-inlining "$@"

#!/usr/bin/env bash

version=b759d7c1c2f140724020f57599ead9ee394a8f7a

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

PACKAGE=opam-0install

rm -rf $PACKAGE
mkdir -p $PACKAGE

(
    cd $TMP
    git clone https://github.com/ocaml-opam/opam-0install-solver.git
    cd opam-0install-solver
    git checkout $version
)

SRC=$TMP/opam-0install-solver

cp -v $SRC/LICENSE.md $PACKAGE/
cp -v -R $SRC/lib $PACKAGE/

git checkout $PACKAGE/lib/dune
git add -A .

#!/bin/bash

version=835c49833d41221758c4ad71f07829ee259668e1

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf spawn
mkdir -p spawn/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/spawn.git
    cd spawn
    git checkout $version
)

SRC=$TMP/spawn

cp -v $SRC/src/spawn{.ml,.mli,_stubs.c} spawn/src
cp -v $SRC/LICENSE.md spawn/

git checkout spawn/src/dune
git add -A .

#!/usr/bin/env bash

version=b58af5db6afd496cfd4a5f85fb23f30ba8dfbc87

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

PACKAGE=0install-solver

rm -rf $PACKAGE
mkdir -p $PACKAGE/src

(
    cd $TMP
    git clone https://github.com/0install/0install.git
    cd 0install
    git checkout $version
)

SRC=$TMP/0install

cp -v $SRC/COPYING $PACKAGE/
cp -v -R $SRC/src/solver $PACKAGE/src/

git checkout $PACKAGE/src/solver/dune
rm -rv 0install-solver/src/solver/tests
git add -A .

#!/bin/bash

version=7a95a8c8c39ed0742284d42216106cb9559fe34e

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf notty
mkdir -p notty/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/notty.git
    cd notty
    git checkout $version
)

SRC=$TMP/notty

cp -v $SRC/LICENSE.md notty/
cp -v -R $SRC/{src,src-unix} notty/
rm notty/src/*_top*.ml

git checkout notty/{src,src-unix}/dune
git add -A .

#!/bin/bash

version=07eb8988452ad51a09d0ab7379d73a87674aba6e

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf csexp
mkdir -p csexp/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/csexp.git
    cd csexp
    git checkout $version
)

SRC=$TMP/csexp

cp -v $SRC/src/csexp.{ml,mli} csexp/src
cp -v $SRC/LICENSE.md csexp/

git checkout csexp/src/dune
git add -A .

#!/bin/bash

set -eu

usage()
{
    cat <<EOF
Usage:
    $(basename "${0}") <command> <clean_command> <name>
EOF
}

if [ $# -ne 3 ]; then
    usage
    exit 1
fi

command="${1}"
clean_command="${2}"
name="${3}"

hyperfine "${command}" \
    --show-output \
    --warmup 2 \
    --runs 3 \
    --prepare "${clean_command}" \
    --export-json bench.json \
    > /dev/null

mean_time=$(cat bench.json | jq '.results[0].mean | tostring')

cat<<EOF
[
    {
        "name": "${name}",
        "unit": "seconds",
        "value": ${mean_time}
    }
]
EOF

. ../helpers.sh

make_fake_ocamlformat() {
  version=$1
  if [ "$#" -eq "1" ]
  then
    ml_file=""
  else
    ml_file="$2"
  fi
  mkdir ocamlformat
  cat > ocamlformat/dune-project <<EOF
(lang dune 3.13)
(package (name ocamlformat))
EOF
  if [ ! "$ml_file" = "no-ml-file" ]
  then
    cat > ocamlformat/ocamlformat.ml <<EOF
let version = "$version"
let () =
  if Sys.file_exists ".ocamlformat-ignore" then
  print_endline "ignoring some files"
;;
let () = print_endline ("formatted with version "^version)
EOF
  fi
  cat > ocamlformat/dune <<EOF
(executable
 (public_name ocamlformat))
EOF
  tar cf ocamlformat-$version.tar ocamlformat
  rm -rf ocamlformat
}

make_ocamlformat_opam_pkg() {
  version=$1
  if [ "$#" -eq "2" ]
  then
    port="$2"
  else
    port=""
  fi
  if [ ! "$port" = "" ]
  then
    mkpkg ocamlformat $version <<EOF
build: [
  [
     "dune"
     "build"
     "-p"
     name
     "@install"
  ]
]
url {
  src: "http://127.0.0.1:$port"
  checksum: [
    "md5=$(md5sum ocamlformat-$version.tar | cut -f1 -d' ')"
  ]
}
EOF
  else
    mkpkg ocamlformat $version <<EOF
build: [
  [
     "dune"
     "build"
     "-p"
     name
     "@install"
  ]
]
url {
  src: "file://$PWD/ocamlformat-$version.tar"
  checksum: [
    "md5=$(md5sum ocamlformat-$version.tar | cut -f1 -d' ')"
  ]
}
EOF
  fi
}

make_project_with_dev_tool_lockdir() {
  cat > dune-project <<EOF
(lang dune 3.13)
(package
 (name foo))
EOF
  cat > foo.ml <<EOF
let () = print_endline "Hello, world"
EOF
  cat > dune <<EOF
(executable
 (public_name foo))
EOF
  cat > dune-workspace <<EOF
(lang dune 3.13)
 (lock_dir
  (path "dev-tools.locks/ocamlformat")
  (repositories mock))
  (lock_dir
   (repositories mock))
 (repository
  (name mock)
  (source "file://$(pwd)/mock-opam-repository"))
EOF
}


make_printer_lib() {
  version=$1
  mkdir printer
  cat > printer/dune-project <<EOF
(lang dune 3.13)
(package (name printer))
EOF
  if [ $version = "1.0" ]
  then
  cat > printer/printer.ml <<EOF
let print () = print_endline "formatted"
EOF
  else
  cat > printer/printer.ml <<EOF
let print () = print_endline "Hello World!"
EOF
  fi
  cat > printer/dune <<EOF
(library
 (public_name printer))
EOF
  tar cf printer.$version.tar printer
  rm -r printer
}

make_opam_printer() {
  version=$1
  mkpkg printer $version <<EOF
build: [
   [
    "dune"
    "build"
    "-p"
    name
    "@install"
   ]
 ]
 url {
 src: "file://$PWD/printer.$version.tar"
 checksum: [
  "md5=$(md5sum printer.$version.tar | cut -f1 -d' ')"
 ]
}
EOF
}

make_fake_ocamlformat_from_path() {
  mkdir .bin
  cat > .bin/ocamlformat <<EOF
#!/bin/sh
if [ -f ".ocamlformat-ignore" ]
then
  echo "ignoring some files"
fi
echo "fake ocamlformat from PATH"
EOF
  chmod +x .bin/ocamlformat
  PATH=$PWD/.bin:$PATH
}

#!/bin/bash

set -eu

usage()
{
    cat <<EOF
Usage:
    $(basename "${0}") <path_to_dune>
EOF
}

if [ $# -ne 1 ]; then
    usage
    exit 1
fi

path_to_dune="${1}"

start_dune () {
    ((${path_to_dune} build "$@" --watch @all > .#dune-output 2>&1) || (echo exit $? >> .#dune-output)) &
    DUNE_PID=$!;
}

timeout="$(command -v timeout || echo gtimeout)"

with_timeout () {
    $timeout 2 "$@"
    exit_code=$?
    if [ "$exit_code" = 124 ]
    then
        echo Timed out
        cat .#dune-output
    else
        return "$exit_code"
    fi
}

stop_dune () {
    with_timeout dune shutdown;
    wait $DUNE_PID;
    cat .#dune-output;
}

echo Breaking build
echo "let f() = 2" > ./internal/m_1_1_1_1.ml
echo "val f : unit -> int" > ./internal/m_1_1_1_1.mli

echo Starting dune
start_dune

echo Checking for error
until grep 'error' .#dune-output > /dev/null; do sleep 0.1; done

echo Found, fixing build
echo "let f() = ()" > ./internal/m_1_1_1_1.ml
echo "val f : unit -> unit" > ./internal/m_1_1_1_1.mli

echo Checking for success
until grep 'Success' .#dune-output > /dev/null; do sleep 0.1; done

echo Found, stopping dune
stop_dune

#!/bin/bash

version=426a63dadaba7f47435c9b01d8cb932c548ff2e2
name=opam-file-format

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf $name
mkdir -p $name/

(
    cd $TMP
    git clone https://github.com/ocaml-dune/$name.git
    cd $name
    git checkout $version
)

SRC=$TMP/$name

cp -v $SRC/src/*.{ml,mli,mll,mly} $SRC/LICENSE $name/

git checkout $name/dune
git add -A .

#!/bin/bash

version=0deb52fd3d644acb5ab3dc1db604937c71a542aa

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

rm -rf build_path_prefix_map
mkdir -p build_path_prefix_map/src

(
    cd $TMP
    git clone https://gitlab.com/gasche/build_path_prefix_map.git
    cd build_path_prefix_map
    git checkout $version
)

SRC=$TMP/build_path_prefix_map

cp -v $SRC/build_path_prefix_map.{ml,mli} build_path_prefix_map/src

git checkout build_path_prefix_map/src/dune
git add -A .

# setup-dkml
#   Short form: sd4
  
<#
.SYNOPSIS

Setup Diskuv OCaml (DKML) compiler on a desktop PC.

.DESCRIPTION

Setup Diskuv OCaml (DKML) compiler on a desktop PC.

.PARAMETER PC_PROJECT_DIR
Context variable for the project directory. Defaults to the current directory.

.PARAMETER FDOPEN_OPAMEXE_BOOTSTRAP
Input variable.

.PARAMETER CACHE_PREFIX
Input variable.

.PARAMETER OCAML_COMPILER
Input variable. -DKML_COMPILER takes priority. If -DKML_COMPILER is not set and -OCAML_COMPILER is set, then the specified OCaml version tag of dkml-compiler (ex. 4.12.1) is used.

.PARAMETER DKML_COMPILER
Input variable. Unspecified or blank is the latest from the default branch (main) of dkml-compiler. @repository@ is the latest from Opam.

.PARAMETER PRIMARY_SWITCH_SKIP_INSTALL
Input variable. If true no dkml-base-compiler will be installed in the 'dkml' switch.

.PARAMETER SECONDARY_SWITCH
Input variable. If true then the secondary switch named 'two' is created, in addition to the always-present 'dkml' switch. 

.PARAMETER CONF_DKML_CROSS_TOOLCHAIN
Input variable. Unspecified or blank is the latest from the default branch (main) of conf-dkml-cross-toolchain. @repository@ is the latest from Opam.

.PARAMETER DISKUV_OPAM_REPOSITORY
Input variable. Defaults to the value of -DEFAULT_DISKUV_OPAM_REPOSITORY_TAG (see below)

.PARAMETER DKML_HOME
Input variables. If specified then DiskuvOCamlHome, DiskuvOCamlBinaryPaths and DiskuvOCamlDeploymentId will be set, in addition to the always-present DiskuvOCamlVarsVersion, DiskuvOCamlVersion
and DiskuvOCamlMSYS2Dir.

# autogen from global_env_vars.
.PARAMETER DEFAULT_DKML_COMPILER
Environment variable.

.PARAMETER PIN_ALCOTEST
Environment variable.

.PARAMETER PIN_ALCOTEST_ASYNC
Environment variable.

.PARAMETER PIN_ALCOTEST_JS
Environment variable.

.PARAMETER PIN_ALCOTEST_LWT
Environment variable.

.PARAMETER PIN_ALCOTEST_MIRAGE
Environment variable.

.PARAMETER PIN_BASE
Environment variable.

.PARAMETER PIN_BASE_BIGSTRING
Environment variable.

.PARAMETER PIN_BIGSTRINGAF
Environment variable.

.PARAMETER PIN_CORE
Environment variable.

.PARAMETER PIN_CORE_KERNEL
Environment variable.

.PARAMETER PIN_CTYPES_FOREIGN
Environment variable.

.PARAMETER PIN_CTYPES
Environment variable.

.PARAMETER PIN_CURLY
Environment variable.

.PARAMETER PIN_DKML_APPS
Environment variable.

.PARAMETER PIN_DKML_EXE
Environment variable.

.PARAMETER PIN_DUNE
Environment variable.

.PARAMETER PIN_FEATHER
Environment variable.

.PARAMETER PIN_OCAMLBUILD
Environment variable.

.PARAMETER PIN_OCAMLFIND
Environment variable.

.PARAMETER PIN_OCP_INDENT
Environment variable.

.PARAMETER PIN_PPX_EXPECT
Environment variable.

.PARAMETER PIN_WITH_DKML
Environment variable.

#>
[CmdletBinding()]
param (
  # Context variables
  [Parameter(HelpMessage='Defaults to the current directory')]
  [string]
  $PC_PROJECT_DIR = $PWD,
  
  # Input variables
  [Parameter()]
  [string]
  $FDOPEN_OPAMEXE_BOOTSTRAP = "false",
  [Parameter()]
  [string]
  $CACHE_PREFIX = "v1",
  [Parameter()]
  [string]
  $OCAML_COMPILER = "",
  [Parameter()]
  [string]
  $DKML_COMPILER = "",
  [Parameter()]
  [string]
  $PRIMARY_SWITCH_SKIP_INSTALL = "false",
  [Parameter()]
  [string]
  $SECONDARY_SWITCH = "false",
  [Parameter()]
  [string]
  $CONF_DKML_CROSS_TOOLCHAIN = "@repository@",
  [Parameter()]
  [string]
  $DISKUV_OPAM_REPOSITORY = "",
  [Parameter()]
  [string]
  $DKML_HOME = ""

  # Conflicts with automatic variable $Verbose
  # [Parameter()]
  # [string]
  # $VERBOSE = "false"
    
  # Environment variables (can be overridden on command line)
  # autogen from global_env_vars.
  ,[Parameter()] [string] $DEFAULT_DKML_COMPILER = "4.14.0-v1.1.0-prerel15"
  ,[Parameter()] [string] $PIN_ALCOTEST = "1.6.0"
  ,[Parameter()] [string] $PIN_ALCOTEST_ASYNC = "1.6.0"
  ,[Parameter()] [string] $PIN_ALCOTEST_JS = "1.6.0"
  ,[Parameter()] [string] $PIN_ALCOTEST_LWT = "1.6.0"
  ,[Parameter()] [string] $PIN_ALCOTEST_MIRAGE = "1.6.0"
  ,[Parameter()] [string] $PIN_BASE = "v0.15.1"
  ,[Parameter()] [string] $PIN_BASE_BIGSTRING = "v0.15.0"
  ,[Parameter()] [string] $PIN_BIGSTRINGAF = "0.9.0+msvc"
  ,[Parameter()] [string] $PIN_CORE = "v0.15.1"
  ,[Parameter()] [string] $PIN_CORE_KERNEL = "v0.15.0"
  ,[Parameter()] [string] $PIN_CTYPES_FOREIGN = "0.19.2-windowssupport-r5"
  ,[Parameter()] [string] $PIN_CTYPES = "0.19.2-windowssupport-r5"
  ,[Parameter()] [string] $PIN_CURLY = "0.2.1-windows-env_r2"
  ,[Parameter()] [string] $PIN_DKML_APPS = "1.2.0~prerel8"
  ,[Parameter()] [string] $PIN_DKML_EXE = "1.2.0~prerel8"
  ,[Parameter()] [string] $PIN_DUNE = "3.6.2"
  ,[Parameter()] [string] $PIN_FEATHER = "0.3.0"
  ,[Parameter()] [string] $PIN_OCAMLBUILD = "0.14.0"
  ,[Parameter()] [string] $PIN_OCAMLFIND = "1.9.1"
  ,[Parameter()] [string] $PIN_OCP_INDENT = "1.8.2-windowssupport"
  ,[Parameter()] [string] $PIN_PPX_EXPECT = "v0.15.1"
  ,[Parameter()] [string] $PIN_WITH_DKML = "1.2.0~prerel8"
)

$ErrorActionPreference = "Stop"

# Reset environment so no conflicts with a parent Opam or OCaml system
if (Test-Path Env:OPAMROOT)             { Remove-Item Env:OPAMROOT }
if (Test-Path Env:OPAMSWITCH)           { Remove-Item Env:OPAMSWITCH }
if (Test-Path Env:OPAM_SWITCH_PREFIX)   { Remove-Item Env:OPAM_SWITCH_PREFIX }
if (Test-Path Env:CAML_LD_LIBRARY_PATH) { Remove-Item Env:CAML_LD_LIBRARY_PATH }
if (Test-Path Env:OCAMLLIB)             { Remove-Item Env:OCAMLLIB }
if (Test-Path Env:OCAML_TOPLEVEL_PATH)  { Remove-Item Env:OCAML_TOPLEVEL_PATH }

# Pushdown context variables
$env:PC_CI = 'true'
$env:PC_PROJECT_DIR = $PC_PROJECT_DIR

# Pushdown input variables
$env:FDOPEN_OPAMEXE_BOOTSTRAP = $FDOPEN_OPAMEXE_BOOTSTRAP
$env:CACHE_PREFIX = $CACHE_PREFIX
$env:OCAML_COMPILER = $OCAML_COMPILER
$env:DKML_COMPILER = $DKML_COMPILER
$env:PRIMARY_SWITCH_SKIP_INSTALL = $PRIMARY_SWITCH_SKIP_INSTALL
$env:SECONDARY_SWITCH = $SECONDARY_SWITCH
$env:CONF_DKML_CROSS_TOOLCHAIN = $CONF_DKML_CROSS_TOOLCHAIN
$env:DISKUV_OPAM_REPOSITORY = $DISKUV_OPAM_REPOSITORY
$env:DKML_HOME = $DKML_HOME

# Set matrix variables
# autogen from pc_vars. only windows_x86_64
$env:dkml_host_os = "windows"
$env:opam_root_cacheable = "${env:PC_PROJECT_DIR}/.ci/o"
$env:abi_pattern = "win32-windows_x86_64"
$env:msys2_system = "CLANG64"
$env:msys2_packages = "mingw-w64-clang-x86_64-pkg-config"
$env:exe_ext = ".exe"
$env:bootstrap_opam_version = "2.2.0-alpha-20221104"
$env:opam_abi = "windows_x86_64"
$env:dkml_host_abi = "windows_x86_64"
$env:opam_root = "${env:PC_PROJECT_DIR}/.ci/o"
$env:vsstudio_hostarch = "x64"
$env:vsstudio_arch = "x64"


# Set environment variables
# autogen from global_env_vars.
$env:DEFAULT_DKML_COMPILER = $DEFAULT_DKML_COMPILER
$env:PIN_ALCOTEST = $PIN_ALCOTEST
$env:PIN_ALCOTEST_ASYNC = $PIN_ALCOTEST_ASYNC
$env:PIN_ALCOTEST_JS = $PIN_ALCOTEST_JS
$env:PIN_ALCOTEST_LWT = $PIN_ALCOTEST_LWT
$env:PIN_ALCOTEST_MIRAGE = $PIN_ALCOTEST_MIRAGE
$env:PIN_BASE = $PIN_BASE
$env:PIN_BASE_BIGSTRING = $PIN_BASE_BIGSTRING
$env:PIN_BIGSTRINGAF = $PIN_BIGSTRINGAF
$env:PIN_CORE = $PIN_CORE
$env:PIN_CORE_KERNEL = $PIN_CORE_KERNEL
$env:PIN_CTYPES_FOREIGN = $PIN_CTYPES_FOREIGN
$env:PIN_CTYPES = $PIN_CTYPES
$env:PIN_CURLY = $PIN_CURLY
$env:PIN_DKML_APPS = $PIN_DKML_APPS
$env:PIN_DKML_EXE = $PIN_DKML_EXE
$env:PIN_DUNE = $PIN_DUNE
$env:PIN_FEATHER = $PIN_FEATHER
$env:PIN_OCAMLBUILD = $PIN_OCAMLBUILD
$env:PIN_OCAMLFIND = $PIN_OCAMLFIND
$env:PIN_OCP_INDENT = $PIN_OCP_INDENT
$env:PIN_PPX_EXPECT = $PIN_PPX_EXPECT
$env:PIN_WITH_DKML = $PIN_WITH_DKML

# https://patchwork.kernel.org/project/qemu-devel/patch/20211215073402.144286-17-thuth@redhat.com/
$env:CHERE_INVOKING = "yes" # Preserve the current working directory
$env:MSYSTEM = $env:msys2_system # Start a 64 bit environment if CLANG64, etc.

########################### before_script ###############################

# Troubleshooting
If ( "${env:VERBOSE}" -eq "true" ) { Get-ChildItem 'env:' }

# -----
# MSYS2
# -----
#
# https://www.msys2.org/docs/ci/
# https://patchwork.kernel.org/project/qemu-devel/patch/20211215073402.144286-17-thuth@redhat.com/

if ( Test-Path -Path msys64\usr\bin\pacman.exe ) {
  Write-Host "Re-using MSYS2 from cache."
}
else {
  Write-Host "Download the archive ..."
  If ( !(Test-Path -Path msys64\var\cache ) ) { New-Item msys64\var\cache -ItemType Directory | Out-Null }
  If ( !(Test-Path -Path msys64\var\cache\msys2.exe ) ) { Invoke-WebRequest "https://github.com/msys2/msys2-installer/releases/download/2022-09-04/msys2-base-x86_64-20220904.sfx.exe" -outfile "msys64\var\cache\msys2.exe" }

  Write-Host "Extract the archive ..."
  msys64\var\cache\msys2.exe -y # Extract to .\msys64
  Remove-Item msys64\var\cache\msys2.exe # Delete the archive again
  ((Get-Content -path msys64\etc\post-install\07-pacman-key.post -Raw) -replace '--refresh-keys', '--version') | Set-Content -Path msys64\etc\post-install\07-pacman-key.post
  msys64\usr\bin\bash -lc "sed -i 's/^CheckSpace/#CheckSpace/g' /etc/pacman.conf"

  Write-Host "Run for the first time ..."
  msys64\usr\bin\bash -lc ' '
}
Write-Host "Update MSYS2 ..."
msys64\usr\bin\bash -lc 'pacman --noconfirm -Syuu' # Core update (in case any core packages are outdated)
msys64\usr\bin\bash -lc 'pacman --noconfirm -Syuu' # Normal update
if ("${env:CI}" -eq "true") { taskkill /F /FI "MODULES eq msys-2.0.dll" } # Only safe to kill MSYS2 in CI

Write-Host "Install matrix, required and CI packages ..."
#   Packages for GitLab CI:
#     dos2unix (used to translate PowerShell written files below in this CI .yml into MSYS2 scripts)
msys64\usr\bin\bash -lc 'set -x; pacman -Sy --noconfirm --needed ${msys2_packages}  wget make rsync diffutils patch unzip git tar xz dos2unix'

Write-Host "Uninstall MSYS2 conflicting executables ..."
msys64\usr\bin\bash -lc 'rm -vf /usr/bin/link.exe' # link.exe interferes with MSVC's link.exe

Write-Host "Installing VSSetup for the Get-VSSetupInstance function ..."
Install-Module VSSetup -Scope CurrentUser -Force

Write-Host "Writing scripts ..."

# POSIX and AWK scripts

If ( !(Test-Path -Path.ci\sd4 ) ) { New-Item .ci\sd4 -ItemType Directory | Out-Null }

$Content = @'
#!/bin/sh

# ------------------------ Log Formatting ------------------------

TXT_SECTION="\e[94m" # bright blue
TXT_CLEAR="\e[0m"

if [ "${GITLAB_CI:-}" = "true" ]; then
    # https://docs.gitlab.com/ee/ci/jobs/#expand-and-collapse-job-log-sections
    print_section_start() {
        print_section_start_NAME=$1
        shift
        printf "\e[0Ksection_start:%s:%s[collapsed=true]\r\e[0K" \
            "$(date +%s)" \
            "$print_section_start_NAME"
    }
    print_section_end() {
        print_section_end_NAME=$1
        shift
        printf "\e[0Ksection_end:%s:%s\r\e[0K\n" \
            "$(date +%s)" \
            "$print_section_end_NAME"
    }
elif [ -n "${GITHUB_ENV:-}" ]; then
    # https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#grouping-log-lines
    print_section_start() {
        print_section_start_NAME=$1
        shift
        printf "::group::"
    }
    print_section_end() {
        print_section_end_NAME=$1
        shift
        printf "::endgroup::\n"
    }
else
    print_section_start() {
        print_section_start_NAME=$1
        shift
    }
    print_section_end() {
        print_section_end_NAME=$1
        shift
    }
fi

section_begin() {
    # https://docs.gitlab.com/ee/ci/yaml/script.html#add-color-codes-to-script-output
    section_NAME=$1
    shift
    section_HEADER=$1
    shift
    print_section_start "$section_NAME"
    printf "${TXT_SECTION}%s${TXT_CLEAR}\n" "$section_HEADER"
}

section_end() {
    section_NAME=$1
    shift
    print_section_end "$section_NAME"
}

# ------------------- Other Functions -----------------

transfer_dir() {
    transfer_dir_SRC=$1
    shift
    transfer_dir_DST=$1
    shift
    # Remove the destination directory completely, but make sure the parent of the
    # destination directory exists so `mv` will work
    install -d "$transfer_dir_DST"
    rm -rf "$transfer_dir_DST"
    # Move
    mv "$transfer_dir_SRC" "$transfer_dir_DST"
}

# Set TEMP variable which is used, among other things, for OCaml's
# [Filename.temp_dir_name] on Win32, and by with-dkml.exe on Windows
export_temp_for_windows() {
    if [ -x /usr/bin/cygpath ]; then
        if [ -n "${RUNNER_TEMP:-}" ]; then
            # GitHub Actions
            TEMP=$(cygpath -am "$RUNNER_TEMP")
        else
            # GitLab CI/CD or desktop
            install -d .ci/tmp
            TEMP=$(cygpath -am ".ci/tmp")
        fi
        export TEMP
    fi
}

# Fixup opam_root on Windows to be mixed case.
# On input the following variables must be present:
# - opam_root
# - opam_root_cacheable
# On output the input variables will be modified _and_ the
# following variables will be available:
# - original_opam_root
# - original_opam_root_cacheable
# - unix_opam_root
# - unix_opam_root_cacheable
fixup_opam_root() {
    # shellcheck disable=SC2034
    original_opam_root=$opam_root
    # shellcheck disable=SC2034
    original_opam_root_cacheable=$opam_root_cacheable
    if [ -x /usr/bin/cygpath ]; then
        opam_root=$(/usr/bin/cygpath -m "$opam_root")
        opam_root_cacheable=$(/usr/bin/cygpath -m "$opam_root_cacheable")
        unix_opam_root=$(/usr/bin/cygpath -u "$opam_root")
        unix_opam_root_cacheable=$(/usr/bin/cygpath -u "$opam_root_cacheable")
    else
        # shellcheck disable=SC2034
        unix_opam_root=$opam_root
        # shellcheck disable=SC2034
        unix_opam_root_cacheable=$opam_root_cacheable
    fi
}
'@
Set-Content -Path ".ci\sd4\common-values.sh" -Encoding Unicode -Value $Content
msys64\usr\bin\bash -lc 'dos2unix .ci/sd4/common-values.sh'


$Content = @'
#!/bin/sh

# ================
# checkout-code.sh
# ================
#
# Checkouts all of the git source code.
#
# This should be done outside of
# dockcross (used by Linux) since a Docker-in-Docker container can have
# difficulties doing a git checkout (the Git credentials for any private
# repositories are likely not present). We don't care about any private
# repositories for DKML but any code that extends this (ex. DKSDK) may
# need to use private repositories.

set -euf

setup_WORKSPACE_VARNAME=$1
shift
setup_WORKSPACE=$1
shift

if [ -x /usr/bin/cygpath ]; then
    setup_WORKSPACE=$(/usr/bin/cygpath -au "$setup_WORKSPACE")
fi

# ------------------------ Functions ------------------------

# shellcheck source=./common-values.sh
. .ci/sd4/common-values.sh

# Disable automatic garbage collection
git_disable_gc() {
    git_disable_gc_NAME=$1
    shift
    git -C ".ci/sd4/g/$git_disable_gc_NAME" config --local gc.auto 0
}

# Mimic the behavior of GitHub's actions/checkout@v3
# - the plus symbol in 'git fetch ... origin +REF:refs/tags/v0.0' overrides any existing REF
git_checkout() {
    git_checkout_NAME=$1
    shift
    git_checkout_URL=$1
    shift
    git_checkout_REF=$1
    shift

    if [ -e ".ci/sd4/g/$git_checkout_NAME" ]; then
        git_disable_gc "$git_checkout_NAME"
        git -C ".ci/sd4/g/$git_checkout_NAME" remote set-url origin "$git_checkout_URL"
        git -C ".ci/sd4/g/$git_checkout_NAME" fetch --no-tags --progress --no-recurse-submodules --depth=1 origin "+${git_checkout_REF}:refs/tags/v0.0"
    else
        install -d ".ci/sd4/g/$git_checkout_NAME"
        git -C ".ci/sd4/g/$git_checkout_NAME" -c init.defaultBranch=main init
        git_disable_gc "$git_checkout_NAME"
        git -C ".ci/sd4/g/$git_checkout_NAME" remote add origin "$git_checkout_URL"
        git -C ".ci/sd4/g/$git_checkout_NAME" fetch --no-tags --prune --progress --no-recurse-submodules --depth=1 origin "+${git_checkout_REF}:refs/tags/v0.0"
    fi
    git -C ".ci/sd4/g/$git_checkout_NAME" -c advice.detachedHead=false checkout --progress --force refs/tags/v0.0
    git -C ".ci/sd4/g/$git_checkout_NAME" log -1 --format='%H'
}

# ---------------------------------------------------------------------

section_begin checkout-info "Summary: code checkout"

# shellcheck disable=SC2154
echo "
================
checkout-code.sh
================
.
---------
Arguments
---------
WORKSPACE_VARNAME=$setup_WORKSPACE_VARNAME
WORKSPACE=$setup_WORKSPACE
.
------
Inputs
------
VERBOSE=${VERBOSE:-}
.
------
Matrix
------
dkml_host_abi=$dkml_host_abi
.
"

section_end checkout-info

install -d .ci/sd4/g

# dkml-runtime-distribution

#   For 'Diagnose Visual Studio environment variables (Windows)' we need dkml-runtime-distribution
#   so that 'Import-Module Machine' and 'Get-VSSetupInstance' can be run.
#   The version doesn't matter too much, as long as it has a functioning Get-VSSetupInstance.
#   commit 1a3ec82dd851751a95e6a4797387a8163c51520e = tag v0.4.0-prerel20
case "$dkml_host_abi" in
windows_*)
    section_begin checkout-dkml-runtime-distribution 'Checkout dkml-runtime-distribution'
    git_checkout dkml-runtime-distribution https://github.com/diskuv/dkml-runtime-distribution.git "1a3ec82dd851751a95e6a4797387a8163c51520e"
    section_end checkout-dkml-runtime-distribution
    ;;
esac

'@
Set-Content -Path ".ci\sd4\run-checkout-code.sh" -Encoding Unicode -Value $Content
msys64\usr\bin\bash -lc 'dos2unix .ci/sd4/run-checkout-code.sh'


$Content = @'
#!/bin/sh
set -euf

# Constants
SHA512_DEVNULL='cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e'
#   Edited by https://gitlab.com/diskuv/diskuv-ocaml/contributors/release.sh
DEFAULT_DISKUV_OPAM_REPOSITORY_TAG=638195f6db6f54f121a7211515f26d946494eb95
# Constants
DKML_VERSION=1.2.0-prerel8

setup_WORKSPACE_VARNAME=$1
shift
setup_WORKSPACE=$1
shift

if [ -x /usr/bin/cygpath ]; then
    setup_WORKSPACE=$(/usr/bin/cygpath -au "$setup_WORKSPACE")
fi

# ------------------ Variables and functions ------------------------

# shellcheck source=./common-values.sh
. .ci/sd4/common-values.sh

if [ "${VERBOSE:-}" = "true" ]; then
    do_tar_rf() {
        tar rvf "$@"
    }
else
    do_tar_rf() {
        tar rf "$@"
    }
fi

# Make the standard input work as an OCaml string.
# This currently only escapes backslashes and double quotes.
escape_arg_as_ocaml_string() {
    escape_arg_as_ocaml_string_ARG=$1
    shift
    printf "%s" "$escape_arg_as_ocaml_string_ARG" | sed 's#\\#\\\\#g; s#"#\\"#g;'
}

# Fixup opam_root on Windows to be mixed case. Set original_* and unix_* as well.
fixup_opam_root

# Set TEMP variable for Windows
export_temp_for_windows

# Load VS studio environment
if [ -e .ci/sd4/vsenv.sh ]; then
    # shellcheck disable=SC1091
    . .ci/sd4/vsenv.sh
fi

# -------------------------------------------------------------------

section_begin setup-info "Summary: setup-dkml"

# shellcheck disable=SC2154
echo "
=============
setup-dkml.sh
=============
.
---------
Arguments
---------
WORKSPACE_VARNAME=$setup_WORKSPACE_VARNAME
WORKSPACE=$setup_WORKSPACE
.
------
Inputs
------
FDOPEN_OPAMEXE_BOOTSTRAP=${FDOPEN_OPAMEXE_BOOTSTRAP:-}
DISKUV_OPAM_REPOSITORY=${DISKUV_OPAM_REPOSITORY:-}
DKML_COMPILER=${DKML_COMPILER:-}
OCAML_COMPILER=${OCAML_COMPILER:-}
CONF_DKML_CROSS_TOOLCHAIN=${CONF_DKML_CROSS_TOOLCHAIN:-}
PRIMARY_SWITCH_SKIP_INSTALL=${PRIMARY_SWITCH_SKIP_INSTALL:-}
SECONDARY_SWITCH=${SECONDARY_SWITCH:-}
MANYLINUX=${MANYLINUX:-}
DKML_HOME=${DKML_HOME:-}
VERBOSE=${VERBOSE:-}
.
-------------------
Generated Constants
-------------------
DKML_VERSION=$DKML_VERSION
DEFAULT_DISKUV_OPAM_REPOSITORY_TAG=$DEFAULT_DISKUV_OPAM_REPOSITORY_TAG
DEFAULT_DKML_COMPILER=$DEFAULT_DKML_COMPILER
.
------
Matrix
------
dkml_host_abi=$dkml_host_abi
bootstrap_opam_version=$bootstrap_opam_version
abi_pattern=$abi_pattern
opam_root=${opam_root}
opam_root_cacheable=${opam_root_cacheable}
original_opam_root=${original_opam_root}
original_opam_root_cacheable=${original_opam_root_cacheable}
unix_opam_root=${unix_opam_root}
unix_opam_root_cacheable=${unix_opam_root_cacheable}
dockcross_image=${dockcross_image:-}
dockcross_image_custom_prefix=${dockcross_image_custom_prefix:-}
dockcross_run_extra_args=${dockcross_run_extra_args:-}
docker_runner=${docker_runner:-}
in_docker=${in_docker:-}
ocaml_options=${ocaml_options:-}
.
----
Pins
----
PIN_ALCOTEST=${PIN_ALCOTEST}
PIN_ALCOTEST_ASYNC=${PIN_ALCOTEST_ASYNC}
PIN_ALCOTEST_JS=${PIN_ALCOTEST_JS}
PIN_ALCOTEST_LWT=${PIN_ALCOTEST_LWT}
PIN_ALCOTEST_MIRAGE=${PIN_ALCOTEST_MIRAGE}
PIN_BASE=${PIN_BASE}
PIN_BASE_BIGSTRING=${PIN_BASE_BIGSTRING}
PIN_BIGSTRINGAF=${PIN_BIGSTRINGAF}
PIN_CORE=${PIN_CORE}
PIN_CORE_KERNEL=${PIN_CORE_KERNEL}
PIN_CTYPES_FOREIGN=${PIN_CTYPES_FOREIGN}
PIN_CTYPES=${PIN_CTYPES}
PIN_CURLY=${PIN_CURLY}
PIN_DKML_APPS=${PIN_DKML_APPS}
PIN_DKML_EXE=${PIN_DKML_EXE}
PIN_DUNE=${PIN_DUNE}
PIN_FEATHER=${PIN_FEATHER}
PIN_OCAMLBUILD=${PIN_OCAMLBUILD}
PIN_OCAMLFIND=${PIN_OCAMLFIND}
PIN_OCP_INDENT=${PIN_OCP_INDENT}
PIN_PPX_EXPECT=${PIN_PPX_EXPECT}
PIN_WITH_DKML=${PIN_WITH_DKML}
.
"
case "$dkml_host_abi" in
windows_*)
    # shellcheck disable=SC2153
    echo "
-------------
Visual Studio
-------------
VS_DIR=$VS_DIR
VS_VCVARSVER=$VS_VCVARSVER
VS_WINSDKVER=$VS_WINSDKVER
VS_MSVSPREFERENCE=$VS_MSVSPREFERENCE
VS_CMAKEGENERATOR=$VS_CMAKEGENERATOR
.
"
    ;;
esac
section_end setup-info

do_bootstrap() {
    # Bootstrap from historical release
    runit_BOOTSTRAPPED=0

    #   Bootstrap opam from fdopen (Windows)
    if [ "$runit_BOOTSTRAPPED" = 0 ] && [ "${FDOPEN_OPAMEXE_BOOTSTRAP:-}" = "true" ]; then
        if [ -e .ci/sd4/opam64/bin/opam.exe ] && [ -e .ci/sd4/opam64/bin/opam-installer.exe ]; then
            runit_BOOTSTRAPPED=1
        else
            case "$dkml_host_abi" in
            windows_*)
                echo 'Bootstrap opam from fdopen (Windows) ...'
                install -d .ci/sd4/bs/bin
                wget -O "$setup_WORKSPACE"/.ci/sd4/opam64.tar.xz https://github.com/fdopen/opam-repository-mingw/releases/download/0.0.0.2/opam64.tar.xz

                # this stalls: tar xvCfJ "$setup_WORKSPACE"/.ci/sd4 "$setup_WORKSPACE"/.ci/sd4/opam64.tar.xz
                xz -v -d "$setup_WORKSPACE"/.ci/sd4/opam64.tar.xz
                tar xvCf .ci/sd4 .ci/sd4/opam64.tar

                rm -rf "$setup_WORKSPACE"/.ci/sd4/bs/bin/Opam.Runtime.amd64
                mv -v "$setup_WORKSPACE"/.ci/sd4/opam64/bin/Opam.Runtime.amd64/ "$setup_WORKSPACE"/.ci/sd4/bs/bin/
                mv -v "$setup_WORKSPACE"/.ci/sd4/opam64/bin/opam.exe "$setup_WORKSPACE"/.ci/sd4/bs/bin/
                mv -v "$setup_WORKSPACE"/.ci/sd4/opam64/bin/opam-installer.exe "$setup_WORKSPACE"/.ci/sd4/bs/bin/

                # diagnostics
                ldd "$setup_WORKSPACE"/.ci/sd4/bs/bin/opam.exe
                ldd "$setup_WORKSPACE"/.ci/sd4/bs/bin/opam-installer.exe

                runit_BOOTSTRAPPED=1
                ;;
            esac
        fi
    fi

    #   Bootstrap from historical release
    if [ "$runit_BOOTSTRAPPED" = 0 ] && [ "$bootstrap_opam_version" != "os" ]; then
        install -d .ci/sd4/bs
        cd .ci/sd4/bs

        if [ ! -e version ] || [ "$(cat version)" != "$bootstrap_opam_version" ]; then
            echo 'Bootstrap opam from historical release (non-Windows; Windows non-fdopen) ...'
            if command -v curl; then
                curl -L -o opam.tar.gz "https://github.com/diskuv/dkml-component-opam/releases/download/v${bootstrap_opam_version}/dkml-component-staging-opam.tar.gz"
            else
                wget -O opam.tar.gz "https://github.com/diskuv/dkml-component-opam/releases/download/v${bootstrap_opam_version}/dkml-component-staging-opam.tar.gz"
            fi
            tar tvfz opam.tar.gz
            tar xfz opam.tar.gz "./staging-files/${dkml_host_abi}/"
            rm -rf bin/
            mv "staging-files/${dkml_host_abi}/bin" .
            rm -rf "${abi_pattern}"
            printf "%s" "${bootstrap_opam_version}" >version
        fi

        rm -f opam.tar.gz
        cd ../../..

        runit_BOOTSTRAPPED=1
    fi

    #   Bootstrap from package manager or GitHub ocaml/opam release
    case "$runit_BOOTSTRAPPED,$bootstrap_opam_version,$dkml_host_abi" in
    0,os,darwin_*)
        if ! command -v opam; then
            echo 'Bootstrap opam from package manager (macOS) ...'
            brew install gpatch
            brew install opam
        fi
        runit_BOOTSTRAPPED=1
        ;;
    0,os,linux_x86)
        if [ ! -x .ci/sd4/bs/bin/opam ]; then
            echo 'Bootstrap opam from GitHub ocaml/opam release (Linux x86) ...'
            install -d .ci/sd4/bs/bin
            wget -O .ci/sd4/bs/bin/opam.tmp https://github.com/ocaml/opam/releases/download/2.1.2/opam-2.1.2-i686-linux
            sha512_check=$(openssl sha512 2>&1 </dev/null | cut -f 2 -d ' ')
            if [ "$SHA512_DEVNULL" = "$sha512_check" ]; then
                sha512=$(openssl sha512 ".ci/sd4/bs/bin/opam.tmp" 2>/dev/null | cut -f 2 -d ' ')
                check="85a480d60e09a7d37fa0d0434ed97a3187434772ceb4e7e8faa5b06bc18423d004af3ad5849c7d35e72dca155103257fd6b1178872df8291583929eb8f884b6a"
                test "$sha512" = "$check"
                chmod +x .ci/sd4/bs/bin/opam.tmp
                mv .ci/sd4/bs/bin/opam.tmp .ci/sd4/bs/bin/opam
            else
                echo "openssl 512 option unsupported."
                exit 61
            fi
        fi
        runit_BOOTSTRAPPED=1
        ;;
    0,os,linux_x86_64)
        if [ ! -x .ci/sd4/bs/bin/opam ]; then
            echo 'Bootstrap opam from GitHub ocaml/opam release (Linux x86_64) ...'
            install -d .ci/sd4/bs/bin
            wget -O .ci/sd4/bs/bin/opam.tmp https://github.com/ocaml/opam/releases/download/2.1.2/opam-2.1.2-x86_64-linux
            sha512_check=$(openssl sha512 2>&1 </dev/null | cut -f 2 -d ' ')
            if [ "$SHA512_DEVNULL" = "$sha512_check" ]; then
                sha512=$(openssl sha512 ".ci/sd4/bs/bin/opam.tmp" 2>/dev/null | cut -f 2 -d ' ')
                check="c0657ecbd4dc212587a4da70c5ff0402df95d148867be0e1eb1be8863a2851015f191437c3c99b7c2b153fcaa56cac99169c76ec94c5787750d7a59cd1fbb68b"
                test "$sha512" = "$check"
                chmod +x .ci/sd4/bs/bin/opam.tmp
                mv .ci/sd4/bs/bin/opam.tmp .ci/sd4/bs/bin/opam
            else
                echo "openssl 512 option unsupported."
                exit 61
            fi
        fi
        runit_BOOTSTRAPPED=1
        ;;
    esac
}
section_begin bootstrap-opam 'Bootstrap opam'
do_bootstrap
section_end bootstrap-opam

# Start environment distribution tarball
#   We use .tar rather than .tar.gz/.tar.bz2 because we can repeatedly add to an uncompressed .tar. But we need to
#   start with an empty tarball since some tar programs will only add ('tar rf xyz.tar') to an existing .tar.
install -d .ci/sd4/dist
tar cf .ci/sd4/dist/opam-with-env.tar -T /dev/null

do_get_dockcross() {
    if [ -n "${dockcross_image:-}" ]; then
        # The dockcross script is super-slow
        section_begin get-dockcross 'Get dockcross binary (ManyLinux)'
        install -d .ci/sd4
        #   shellcheck disable=SC2086
        docker run ${dockcross_run_extra_args:-} --rm "${dockcross_image_custom_prefix:-}${dockcross_image:-}" >.ci/sd4/dockcross.gen

        # PROBLEM 1
        # ---------
        # Super-annoying stderr output from dockcross at line:
        #    tty -s && [ -z "$MSYS" ] && TTY_ARGS=-ti
        # When there is no tty, get:
        #   tty: ignoring all arguments
        #   not a tty
        # So replace 'tty -s &&' with 'false &&'
        sed 's/tty -s &&/false \&\&/' .ci/sd4/dockcross.gen >.ci/sd4/dockcross-real
        rm -f .ci/sd4/dockcross.gen
        chmod +x .ci/sd4/dockcross-real

        # PROBLEM 2
        # ---------
        # By default dockcross for ManyLinux will chown -R all python packages; super-slow (~10 seconds)!
        # Confer: https://github.com/dockcross/dockcross/blob/master/manylinux-common/pre_exec.sh
        # That kills speed for any repetitive dockcross invocation.
        #
        # BUT it is unnecessary to chown -R when the current user is root, because inside the Docker container
        # the files are already root!
        #
        # The chown -R (within pre_exec.sh) is not run when the user ids are not passed in.
        # Confer: https://github.com/dockcross/dockcross/blob/96d87416f639af0204bdd42553e4b99315ca8476/imagefiles/entrypoint.sh#L21-L53
        #
        # So explicitly call the entrypoint if root!
        case "$dkml_host_abi" in
        #       https://github.com/dockcross/dockcross/blob/master/linux-x86/linux32-entrypoint.sh
        linux_x86)  dockcross_entrypoint=/dockcross/linux32-entrypoint.sh ;;
        *)          dockcross_entrypoint=/dockcross/entrypoint.sh ;;
        esac
        cat > .ci/sd4/dockcross <<EOF
#!/bin/bash
set -euf
BUILDER_UID="\$( id -u )"
BUILDER_GID="\$( id -g )"
if [ "\$BUILDER_UID" = 0 ] && [ "\$BUILDER_GID" = 0 ]; then
    # ---------- Start of dockcross script snippet -------
    # Verbatim from
    # https://github.com/dockcross/dockcross/blob/96d87416f639af0204bdd42553e4b99315ca8476/imagefiles/dockcross#L175-L204
    # except 1) disabling of USER_IDS

    # Bash on Ubuntu on Windows
    UBUNTU_ON_WINDOWS=\$([ -e /proc/version ] && grep -l Microsoft /proc/version || echo "")
    # MSYS, Git Bash, etc.
    MSYS=\$([ -e /proc/version ] && grep -l MINGW /proc/version || echo "")
    # CYGWIN
    CYGWIN=\$([ -e /proc/version ] && grep -l CYGWIN /proc/version || echo "")

    #if [ -z "\$UBUNTU_ON_WINDOWS" -a -z "\$MSYS" -a "\$OCI_EXE" != "podman" ]; then
    #    USER_IDS=(-e BUILDER_UID="\$( id -u )" -e BUILDER_GID="\$( id -g )" -e BUILDER_USER="\$( id -un )" -e BUILDER_GROUP="\$( id -gn )")
    #fi

    # Change the PWD when working in Docker on Windows
    if [ -n "\$UBUNTU_ON_WINDOWS" ]; then
        WSL_ROOT="/mnt/"
        CFG_FILE=/etc/wsl.conf
            if [ -f "\$CFG_FILE" ]; then
                    CFG_CONTENT=\$(cat \$CFG_FILE | sed -r '/[^=]+=[^=]+/!d' | sed -r 's/\s+=\s/=/g')
                    eval "\$CFG_CONTENT"
                    if [ -n "\$root" ]; then
                            WSL_ROOT=\$root
                    fi
            fi
        HOST_PWD=\`pwd -P\`
        HOST_PWD=\${HOST_PWD/\$WSL_ROOT//}
    elif [ -n "\$MSYS" ]; then
        HOST_PWD=\$PWD
        HOST_PWD=\${HOST_PWD/\//}
        HOST_PWD=\${HOST_PWD/\//:\/}
    elif [ -n "\$CYGWIN" ]; then
        for f in pwd readlink cygpath ; do
            test -n "\$(type "\${f}" )" || { echo >&2 "Missing functionality (\${f}) (in cygwin)." ; exit 1 ; } ;
        done ;
        HOST_PWD="\$( cygpath -w "\$( readlink -f "\$( pwd ;)" ; )" ; )" ;
    else
        HOST_PWD=\$PWD
        [ -L \$HOST_PWD ] && HOST_PWD=\$(readlink \$HOST_PWD)
    fi

    # ---------- End of dockcross script snippet -------

    # Handle: dockcross --args "-v X:Y --platform P"
    ARGS=
    if [ "\$#" -ge 1 ] && [ "\$1" = "--args" ]; then
        shift
        ARGS=\$1
        shift
    fi

    # Directly invoke entrypoint
    exec docker run --entrypoint /bin/bash \
        --rm \
        \${ARGS:-} \
         -v "\$HOST_PWD":/work \
        ${dockcross_image_custom_prefix:-}${dockcross_image:-} ${dockcross_entrypoint} "\$@"
else
    HERE=\$(dirname "\$0")
    HERE=\$(cd "\$HERE" && pwd)
    exec "\$HERE/dockcross-real" "\$@"
fi
EOF
        chmod +x .ci/sd4/dockcross

        # Bundle for consumers of setup-dkml.yml
        do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/dockcross .ci/sd4/dockcross-real

        section_end get-dockcross
    fi
}
do_get_dockcross

if [ -n "${dockcross_image:-}" ]; then
    # rsync needs to be available, even after Docker container disappears
    if [ ! -e .ci/sd4/bs/bin/rsync ]; then
        section_begin get-opam-prereqs-in-dockcross 'Get Opam prerequisites (ManyLinux)'
        install -d .ci/sd4/bs/bin
        #   shellcheck disable=SC2016
        .ci/sd4/dockcross --args "${dockcross_run_extra_args:-}" sh -c 'sudo yum install -y rsync && install $(command -v rsync) .ci/sd4/bs/bin'
        section_end get-opam-prereqs-in-dockcross
    fi
fi

# Opam prerequisites for using opam (not for installing opam)

{
    if [ -n "${docker_runner:-}" ]; then
        # rsync needs to be available, even after Docker container disappears
        if [ ! -e .ci/sd4/bs/bin/rsync.deps ]; then
            section_begin get-opam-prereqs-in-docker 'Get Opam prerequisites (Linux Docker)'
            install -d .ci/sd4/bs/bin
            ${docker_runner} sh -c '
            apt-get update &&
            apt-get install -y rsync &&
            ldd /usr/bin/rsync &&
            ls -l /lib/i386-linux-gnu/libpopt.so.0 /lib/i386-linux-gnu/libacl.so.1 /lib/i386-linux-gnu/libattr.so.1 &&
            tar cCfhz / /work/.ci/sd4/bs/bin/deps.tar.gz /usr/bin/rsync /lib/i386-linux-gnu/libpopt.so.0
        '
            touch .ci/sd4/bs/bin/rsync.deps
            section_end get-opam-prereqs-in-docker
        fi
    fi

    # Bundle Opam prerequisites (ManyLinux or Linux Docker)
    if [ -n "${docker_runner:-}" ] || [ -n "${dockcross_image:-}" ]; then
        # Bundle for consumers of setup-dkml.yml
        do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/bs/bin/rsync
    fi
}

# Get Opam Cache
do_get_opam_cache() {
    if [ "$unix_opam_root_cacheable" = "$unix_opam_root" ]; then return; fi
    if [ ! -e "$unix_opam_root_cacheable" ]; then return; fi
    section_begin get-opam-cache "Transferring Opam cache to $original_opam_root_cacheable"
    echo Starting transfer # need some output or GitLab CI will not display the section duration
    transfer_dir "$unix_opam_root_cacheable" "$unix_opam_root"
    echo Finished transfer
    section_end get-opam-cache
}
do_get_opam_cache

# Setup Opam

do_write_opam_scripts() {
    case "${FDOPEN_OPAMEXE_BOOTSTRAP:-},$dkml_host_abi" in
    true,windows_*)
        # With fdopen's opam.exe, 'os-distribution = "cygwinports"'. But native Windows opam.exe has 'os-distribution = "win32"'.
        # But on Windows we always want MSYS2 or native Windows libraries, not Cygwin. If cygwinports then
        # code like https://github.com/ocaml/opam-repository/blob/08cbb8258bd4bf30cd6f307c958911a29d537b54/packages/conf-pkg-config/conf-pkg-config.2/opam#L36
        # will fail. So always set 'os-distribution = "win32"' on Windows.
        PATCH_OS_DISTRIBUTION_WIN32=true
        # With fdopen's opam.exe, no 'exe = ".exe"' is set because Cygwin does not need file extensions.
        # Native Windows requires a .exe extension.
        PATCH_EXE_WIN32=true
        ;;
    *)
        PATCH_OS_DISTRIBUTION_WIN32=false
        PATCH_EXE_WIN32=false
        ;;
    esac

    # ---------------------
    # Empty opam repository
    # ---------------------

    install -d .ci/sd4/eor
    cat >.ci/sd4/eor/repo <<EOF
opam-version: "2.0"
browse: "https://opam.ocaml.org/pkg/"
upstream: "https://github.com/ocaml/opam-repository/tree/master/"
EOF

    # ---------------
    # Create Opam troubleshooting script
    #   Dump logs modified within the last 4 hours
    # ---------------

    cat >.ci/sd4/troubleshoot-opam.sh <<EOF
#!/bin/sh
set -euf
OPAMROOT=\$1
shift
if find . -maxdepth 0 -mmin -240 2>/dev/null >/dev/null; then
    FINDARGS="-mmin -240" # is -mmin supported? BSD (incl. macOS), MSYS2, GNU
else
    FINDARGS="-mtime -1" # use 1 day instead. Solaris
fi
printf "\n\n========= [START OF TROUBLESHOOTING] ===========\n\n" >&2
find "\$OPAMROOT"/log -mindepth 1 -maxdepth 1 \$FINDARGS -name "*.out" ! -name "log-*.out" ! -name "ocaml-variants-*.out" | while read -r dump_on_error_LOG; do
    dump_on_error_BLOG=\$(basename "\$dump_on_error_LOG")
    printf "\n\n========= [TROUBLESHOOTING] %s ===========\n\n" "\$dump_on_error_BLOG" >&2
    awk -v BLOG="\$dump_on_error_BLOG" '{print "[" BLOG "]", \$0}' "\$dump_on_error_LOG" >&2
done
printf "\nScroll up to see the [TROUBLESHOOTING] logs that begin at the [START OF TROUBLESHOOTING] line\n" >&2
EOF

    chmod +x .ci/sd4/troubleshoot-opam.sh
    do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/troubleshoot-opam.sh

    # ---------------
    # Create Opam support scripts (not needed for all platforms)
    #   The PATH to find opam must work internally in setup-dkml.yml (sd4/bs/bin) and
    #   by consumers of setup-dkml.yml (sd4/opamexe)
    # ---------------

    USER_ID=$(id -u)
    GROUP_ID=$(id -g)
    USER_NAME=$(id -un)
    GROUP_NAME=$(id -gn)

    case "${opam_root}" in
    /* | ?:*) # /a/b/c or C:\Windows
        validate_supports_docker() {
            echo "Docker only supported with relative paths for the opam root, not: ${opam_root}" >&2
            exit 3
        }
        ;;
    *) # relative path
        validate_supports_docker() {
            true
        }
        cat >.ci/sd4/opam-in-docker <<EOF
#!/bin/sh
set -euf
export PATH="/work/.ci/sd4/bs/bin:/work/.ci/sd4/opamexe:\$PATH"
export OPAMROOT=/work/${opam_root}
export OPAMROOTISOK=1
if [ "${PATCH_OS_DISTRIBUTION_WIN32}" = true ]; then export OPAMVAR_os_distribution=win32; fi
if [ "${PATCH_EXE_WIN32}" = true ]; then export OPAMVAR_exe=.exe; fi

# Reset environment so no conflicts with a parent Opam or OCaml system
unset OPAM_SWITCH_PREFIX
unset OPAMSWITCH
unset CAML_LD_LIBRARY_PATH
unset OCAMLLIB
unset OCAML_TOPLEVEL_PATH

# Optionally skip troubleshooting
troubleshooting=1
if [ "\$#" -ge 1 ] && [ "\$1" = "--no-troubleshooting" ]; then
    shift
    troubleshooting=0
fi

echo "Running inside Docker container: opam \$*" >&2
set +e
opam "\$@"
exitcode=\$?
if [ \$troubleshooting = 1 ]; then
    [ \$exitcode = 0 ] || "/work/.ci/sd4/troubleshoot-opam.sh" \$OPAMROOT
fi
exit \$exitcode
EOF
        chmod +x .ci/sd4/opam-in-docker
        ;;
    esac

    cat >.ci/sd4/deescalate <<EOF
#!/bin/sh
set -euf

if [ -e /work/.ci/sd4/bs/bin/deps.tar.gz ]; then
    tar xCfz / /work/.ci/sd4/bs/bin/deps.tar.gz
fi

groupadd -g ${GROUP_ID} ${GROUP_NAME}
useradd -l -m -u ${USER_ID} -g ${GROUP_ID} ${USER_NAME}
exec runuser -u ${USER_NAME} -g ${GROUP_NAME} -- "\$@"
EOF

    chmod +x .ci/sd4/deescalate

    # -----------------------------------
    # Create opam-with-env
    # -----------------------------------

    install -d .ci/sd4/dist

    if [ -x .ci/sd4/dockcross ]; then
        # Adding empty dockcross root volume avoids:
        #    cp: target ‘/home/root/’ is not a directory
        #    chown: cannot access ‘/home/root’: No such file or directory
        # from https://github.com/dockcross/dockcross/blob/96d87416f639af0204bdd42553e4b99315ca8476/imagefiles/entrypoint.sh#L31-L32
        install -d .ci/sd4/edr

        cat >.ci/sd4/opam-with-env <<EOF
#!/bin/sh
set -euf

HERE=\$(dirname "\$0")
HERE=\$(cd "\$HERE" && pwd)
PROJECT_DIR=\$(cd "\$HERE"/../.. && pwd)

# Optionally enable terminal if and only if '-it' option given
termargs=
if [ "\$#" -ge 1 ] && [ "\$1" = "-it" ]; then
    shift
    termargs=-it
fi

exec bash "\${PROJECT_DIR}"/.ci/sd4/dockcross --args "\${termargs} -v \${PROJECT_DIR}/.ci/sd4/edr:/home/root ${dockcross_run_extra_args:-}" /work/.ci/sd4/opam-in-docker "\$@"
EOF
        chmod +x .ci/sd4/opam-with-env

        validate_supports_docker

        # Bundle for consumers of setup-dkml.yml
        echo '__ opam-in-docker __' >&2
        cat .ci/sd4/opam-in-docker >&2
        echo '________________________' >&2
        do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/opam-with-env .ci/sd4/opam-in-docker .ci/sd4/edr

    elif [ -n "${docker_runner:-}" ]; then

        cat >.ci/sd4/opam-with-env <<EOF
#!/bin/sh
set -euf
exec ${docker_runner:-} /work/.ci/sd4/deescalate /work/.ci/sd4/opam-in-docker "\$@"
EOF
        chmod +x .ci/sd4/opam-with-env

        validate_supports_docker

        # Bundle for consumers of setup-dkml.yml
        echo '__ opam-in-docker __' >&2
        cat .ci/sd4/opam-in-docker >&2
        echo '________________________' >&2
        echo '__ deescalate __' >&2
        cat .ci/sd4/deescalate >&2
        echo '________________' >&2
        do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/opam-with-env .ci/sd4/opam-in-docker .ci/sd4/deescalate

    else

        cat >.ci/sd4/opam-with-env <<EOF
#!/bin/sh
set -euf

HERE=\$(dirname "\$0")
HERE=\$(cd "\$HERE" && pwd)
PROJECT_DIR=\$(cd "\$HERE"/../.. && pwd)

export PATH="\${PROJECT_DIR}/.ci/sd4/bs/bin:\${PROJECT_DIR}/.ci/sd4/opamexe:\$PATH"
export OPAMROOT='${opam_root}'
export OPAMROOTISOK=1
if [ "${PATCH_OS_DISTRIBUTION_WIN32}" = true ]; then export OPAMVAR_os_distribution=win32; fi
if [ "${PATCH_EXE_WIN32}" = true ]; then export OPAMVAR_exe=.exe; fi

# Reset environment so no conflicts with a parent Opam or OCaml system
unset OPAM_SWITCH_PREFIX
unset OPAMSWITCH
unset CAML_LD_LIBRARY_PATH
unset OCAMLLIB
unset OCAML_TOPLEVEL_PATH

# Optionally skip troubleshooting
troubleshooting=1
if [ "\$#" -ge 1 ] && [ "\$1" = "--no-troubleshooting" ]; then
    shift
    troubleshooting=0
fi

echo "Running: opam \$*" >&2
set +e
opam "\$@"
exitcode=\$?
if [ \$troubleshooting = 1 ]; then
    [ \$exitcode = 0 ] || "\${PROJECT_DIR}/.ci/sd4/troubleshoot-opam.sh" \$OPAMROOT
fi
exit \$exitcode
EOF
        chmod +x .ci/sd4/opam-with-env

        # Bundle for consumers of setup-dkml.yml
        do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/opam-with-env

    fi
    echo '__ opam-with-env __' >&2
    cat .ci/sd4/opam-with-env >&2
    echo '___________________' >&2

    # -------
    # opamrun
    # -------

    install -d .ci/sd4/opamrun
    cat >.ci/sd4/opamrun/opamrun <<EOF
#!/bin/sh
set -euf

HERE=\$(dirname "\$0")
HERE=\$(cd "\$HERE" && pwd)
PROJECT_DIR=\$(cd "\$HERE"/../../.. && pwd)

# Add MSVC compiler environment if available
if [ -e "\${PROJECT_DIR}/.ci/sd4/msvcenv" ]; then
    _oldpath="\$PATH"
    # shellcheck disable=SC1091
    . "\${PROJECT_DIR}/.ci/sd4/msvcenv"
    PATH="\$PATH:\$_oldpath"

    # MSVC (link.exe) needs a TMP as well.
    # Confer: https://docs.microsoft.com/en-us/cpp/build/reference/linking?view=msvc-170#link-environment-variables
    if [ -z "\${TMP:-}" ]; then
        # GitHub Actions as of 2022-10 does not set TMP. GitLab CI/CD does.
        TMP="\$RUNNER_TEMP"
    fi
    export TMP
    if [ -x /usr/bin/cygpath ]; then
        TMP=\$(/usr/bin/cygpath -aw "\$TMP")
    fi
fi

# Windows
if [ -n "\${COMSPEC:-}" ]; then
    # We must place MSYS2 in front of path so that MSYS2
    # tar.exe is used instead of Windows tar.exe.
    PATH="/usr/bin:\$PATH"
fi

exec "\${PROJECT_DIR}/.ci/sd4/opam-with-env" "\$@"
EOF
    chmod +x .ci/sd4/opamrun/opamrun

    # Bundle for consumers of setup-dkml.yml
    do_tar_rf .ci/sd4/dist/opam-with-env.tar .ci/sd4/opamrun
}
section_begin 'write-opam-scripts' 'Write opam scripts'
do_write_opam_scripts
section_end 'write-opam-scripts'

# Expose opamrun (also used for consumers of setup-dkml.yml) to GitHub
if [ -n "${GITHUB_PATH:-}" ]; then
    opamrunabs="$setup_WORKSPACE/.ci/sd4/opamrun"
    if [ -x /usr/bin/cygpath ]; then opamrunabs=$(/usr/bin/cygpath -aw "$opamrunabs"); fi
    echo "$opamrunabs" >>"$GITHUB_PATH"
    # Special case: GITHUB_PATH does not influence msys2.CMD of msys2/setup-msys2@v2, so place in real MSYS2 PATH
    if [ -n "${MSYSTEM:-}" ]; then
        install -d /usr/local/bin
        install .ci/sd4/opamrun/opamrun /usr/local/bin/opamrun
    fi
fi

# Place opamrun in the immediate PATH
PATH="$setup_WORKSPACE/.ci/sd4/opamrun:$PATH"

#   Complicated Opam sequence is because:
#   1. Opam's default curl does not work on Windows,
#      and `opam init` does not provide a way to change it (TODO: need
#      a PR!).
#   2. We have to separate the Opam download cache from the other Opam
#      caches
if [ ! -s "$opam_root/.ci.root-init" ]; then # non-empty init file so can be cached irrespective of existence
    section_begin opam-init 'Initialize opam root'

    # Clear any partial previous attempt
    rm -rf "$opam_root"

    case "$dkml_host_abi,${in_docker:-}" in
    windows_*,*)
        eor=$(cygpath -am "$setup_WORKSPACE"/.ci/sd4/eor)
        opamrun init --disable-sandboxing --no-setup --kind local --bare "$eor"
        case "$(opamrun --version)" in
        2.0.*) echo 'download-command: wget' >>"$opam_root/config" ;;
        *) opamrun option --yes --global download-command=wget ;;
        esac
        ;;
    *,true)
        opamrun init --disable-sandboxing --no-setup --kind local --bare "/work/.ci/sd4/eor"
        ;;
    *)
        opamrun init --disable-sandboxing --no-setup --kind local --bare "$setup_WORKSPACE/.ci/sd4/eor"
        ;;
    esac
    echo yes > "$opam_root/.ci.root-init"

    section_end opam-init
fi

section_begin opam-vars "Summary: opam global variables"
opamrun --no-troubleshooting var --global || true
section_end opam-vars

# Build OCaml

do_switch_create() {
    do_switch_create_NAME=$1
    shift

    section_begin "switch-create-$do_switch_create_NAME" "Create opam switch '$do_switch_create_NAME'"
    # Create, or recreate, the Opam switch. The Opam switch should not be
    # cached except for the compiler (confer docs for setup-ocaml GitHub
    # Action) which is the 'dkml' switch (or the 'two' switch).
    # Check if the switch name is present in the Opam root (which may come from cache)
    NOMINALLY_PRESENT=false
    if opamrun switch list --short | grep "^${do_switch_create_NAME}\$"; then NOMINALLY_PRESENT=true; fi

    # Check if the switch is actually present in case of cache incoherence
    # or corrupt Opam state that could result in:
    #   Error:  No config file found for switch dkml. Switch broken?
    if [ $NOMINALLY_PRESENT = true ] && [ ! -e "$opam_root/$do_switch_create_NAME/.opam-switch/switch-config" ]; then
        # Remove the switch name from Opam root, and any partial switch state.
        # Ignore inevitable warnings/failure about missing switch.
        opamrun --no-troubleshooting switch remove "$do_switch_create_NAME" --yes || true
        rm -rf "${opam_root:?}/$do_switch_create_NAME"
        NOMINALLY_PRESENT=false
    fi

    if [ $NOMINALLY_PRESENT = false ]; then
        opamrun switch create "$do_switch_create_NAME" --empty --yes
    fi
    section_end "switch-create-$do_switch_create_NAME"
}
do_switch_create dkml
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_switch_create two
else
    section_begin "switch-create-two" "Create empty opam switch 'two'"
    # Always create a secondary switch ... just empty. Avoid problems with cache content missing
    # and idempotency.
    opamrun --no-troubleshooting switch remove two --yes || true
    rm -rf "$opam_root/two"
    opamrun switch create two --empty --yes
    section_end "switch-create-two"
fi

do_switch_active() {
    section_begin "switch-active" "Set dkml as active switch"
    opamrun switch set dkml --yes
    section_end "switch-active"
}
do_switch_active

do_opam_repositories_add() {
    section_begin "opam-repo-add" "Add 'diskuv' opam repository"
    if ! opamrun --no-troubleshooting repository list -s | grep '^diskuv'; then
        opamrun repository add diskuv "git+https://github.com/diskuv/diskuv-opam-repository.git#${DISKUV_OPAM_REPOSITORY:-$DEFAULT_DISKUV_OPAM_REPOSITORY_TAG}" --yes --dont-select
    fi
    section_end "opam-repo-add"
}
do_opam_repositories_add

do_opam_repositories_config() {
    do_opam_repositories_config_NAME=$1
    shift

    section_begin "opam-repo-$do_opam_repositories_config_NAME" "Attach repositories to $do_opam_repositories_config_NAME"

    if [ ! -s "$opam_root/.ci.$do_opam_repositories_config_NAME.repo-init" ]; then # non-empty init file so can be cached irrespective of existence
        opamrun --no-troubleshooting repository remove default --switch "$do_opam_repositories_config_NAME" --yes || true
        opamrun --no-troubleshooting repository remove diskuv --switch "$do_opam_repositories_config_NAME" --yes || true
        opamrun repository add default --switch "$do_opam_repositories_config_NAME" --yes
        opamrun repository add diskuv --switch "$do_opam_repositories_config_NAME" --yes
        echo yes > "$opam_root/.ci.$do_opam_repositories_config_NAME.repo-init"
    fi

    section_end "opam-repo-$do_opam_repositories_config_NAME"
}
do_opam_repositories_config dkml
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_opam_repositories_config two
fi

do_opam_repositories_update() {
    section_begin "opam-repo-update" "Update opam repositories"
    # The default repository may be the initial 'eor' (empty) repository
    opamrun repository set-url default https://opam.ocaml.org --yes
    # Always set the `diskuv` repository url since it can change
    opamrun repository set-url diskuv "git+https://github.com/diskuv/diskuv-opam-repository.git#${DISKUV_OPAM_REPOSITORY:-$DEFAULT_DISKUV_OPAM_REPOSITORY_TAG}" --yes --dont-select
    # Update both `default` and `diskuv` Opam repositories
    opamrun update default diskuv
    section_end "opam-repo-update"
}
do_opam_repositories_update

do_pins() {
    do_pins_NAME=$1
    shift

    # dkml-base-compiler

    if [ "${DKML_COMPILER:-}" != '@repository@' ] && [ -z "${DKML_COMPILER:-}" ] && [ -z "${OCAML_COMPILER:-}" ]; then
        section_begin checkout-dkml-base-compiler "Pin dkml-base-compiler to default ${DEFAULT_DKML_COMPILER} (neither dkml-base-compiler nor OCAML_COMPILER specified) for $do_pins_NAME switch"
        opamrun pin add --switch "$do_pins_NAME" --yes --no-action dkml-base-compiler "https://github.com/diskuv/dkml-compiler.git#${DEFAULT_DKML_COMPILER}"
        section_end checkout-dkml-base-compiler
    elif [ "${DKML_COMPILER:-}" != '@repository@' ] && [ -n "${DKML_COMPILER:-}" ] && [ -z "${OCAML_COMPILER:-}" ]; then
        section_begin checkout-dkml-base-compiler "Pin dkml-base-compiler to $DKML_COMPILER (dkml-base-compiler specified; no OCAML_COMPILER specified) for $do_pins_NAME switch"
        opamrun pin add --switch "$do_pins_NAME" --yes --no-action dkml-base-compiler "https://github.com/diskuv/dkml-compiler.git#${DKML_COMPILER}"
        section_end checkout-dkml-base-compiler
    elif [ -n "${OCAML_COMPILER:-}" ]; then
        # Validate OCAML_COMPILER (OCAML_COMPILER specified)
        case "${OCAML_COMPILER:-}" in
        4.12.1) true ;;
        4.14.0) true ;;
        *)
            echo "OCAML_COMPILER version ${OCAML_COMPILER:-} is not supported" >&2
            exit 109
            ;;
        esac

        section_begin checkout-dkml-base-compiler "Pin dkml-base-compiler (OCAML_COMPILER specified) for $do_pins_NAME switch"
        opamrun pin add --switch "$do_pins_NAME" --yes --no-action dkml-base-compiler "https://github.com/diskuv/dkml-compiler.git#${OCAML_COMPILER}-v${DKML_VERSION}"
        section_end checkout-dkml-base-compiler
    fi

    # conf-dkml-cross-toolchain

    if [ "${CONF_DKML_CROSS_TOOLCHAIN:-}" != '@repository@' ]; then
        section_begin checkout-conf-dkml-cross-toolchain "Pin conf-dkml-cross-toolchain for $do_pins_NAME switch"
        opamrun pin add --switch "$do_pins_NAME" --yes --no-action conf-dkml-cross-toolchain "https://github.com/diskuv/conf-dkml-cross-toolchain.git#$CONF_DKML_CROSS_TOOLCHAIN"
        section_end checkout-conf-dkml-cross-toolchain
    fi

    # patches necessary for Windows in diskuv-opam-repository
    #
    # - ocamlfind and ocamlbuild
    #
    # - dune-configurator (and hence Dune)
    # Dune 2.9.1 and 3.0.2 will fail to build jst-config.v0.14.1 because for jst-config/discover/discover.ml Dune does:
    #   cl -nologo -O2 -Gy- -MD    -I Z:/.opam_root-cached-8/installer-ocaml/lib/ocaml -o C:\Users\beckf\AppData\Local\Temp\build_f18aec_dune\ocaml-configurator4d3858\c-test-31\test.obj -c C:\Users\beckf\AppData\Local\Temp\build_f18aec_dune\ocaml-configurator4d3858\c-test-31\test.c advapi32.lib ws2_32.lib version.lib
    # instead of
    #   cl -nologo -O2 -Gy- -MD    -I Z:/.opam_root-cached-8/installer-ocaml/lib/ocaml /FoC:\Users\beckf\AppData\Local\Temp\build_f18aec_dune\ocaml-configurator4d3858\c-test-31\test.obj -c C:\Users\beckf\AppData\Local\Temp\build_f18aec_dune\ocaml-configurator4d3858\c-test-31\test.c advapi32.lib ws2_32.lib version.lib
    # with the (irrelevant) test.c file:
    #    #include <stdio.h>
    #    #include <caml/config.h>
    #
    #    #ifdef ARCH_BIG_ENDIAN
    #    const char *s0 = "BEGIN-0-true-END";
    #    #else
    #    const char *s0 = "BEGIN-0-false-END";
    #    #endif
    #
    #    #ifdef ARCH_SIXTYFOUR
    #    const char *s1 = "BEGIN-1-true-END";
    #    #else
    #    const char *s1 = "BEGIN-1-false-END";
    #    #endif
    #
    # The actual problem is dune-configurator ... we only have patches in Diskuv
    # repository up until 2.9.3. Need to upstream fix the problem.
    #
    # - ppx_expect; only patch is for v0.14.1. Need to upstream fix the problem.
    # - base; patches for v0.14.1/2/3. Need to upstream fix the problem.
    section_begin "opam-pins-$do_pins_NAME" "Opam pins for $do_pins_NAME switch"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version alcotest "${PIN_ALCOTEST}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version alcotest-async "${PIN_ALCOTEST_ASYNC}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version alcotest-js "${PIN_ALCOTEST_JS}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version alcotest-lwt "${PIN_ALCOTEST_LWT}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version alcotest-mirage "${PIN_ALCOTEST_MIRAGE}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version base "${PIN_BASE}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version base_bigstring "${PIN_BASE_BIGSTRING}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version bigstringaf "${PIN_BIGSTRINGAF}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version core "${PIN_CORE}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version core_kernel "${PIN_CORE_KERNEL}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version ctypes "${PIN_CTYPES}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version ctypes-foreign "${PIN_CTYPES_FOREIGN}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version curly "${PIN_CURLY}"
    opamrun pin remove --switch "$do_pins_NAME"  --yes --no-action digestif # this used to be pinned, so any cached opamroot needs it unpinned
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version dkml-apps "${PIN_DKML_APPS}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version dkml-exe "${PIN_DKML_EXE}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version dune "${PIN_DUNE}"
    opamrun pin remove --switch "$do_pins_NAME"  --yes --no-action dune-configurator # this used to be pinned, so any cached opamroot needs it unpinned
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version feather "${PIN_FEATHER}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version ocamlbuild "${PIN_OCAMLBUILD}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version ocamlfind "${PIN_OCAMLFIND}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version ocp-indent "${PIN_OCP_INDENT}"
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version ppx_expect "${PIN_PPX_EXPECT}"
    opamrun pin remove --switch "$do_pins_NAME"  --yes --no-action ptime # this used to be pinned, so any cached opamroot needs it unpinned
    opamrun pin remove --switch "$do_pins_NAME"  --yes --no-action time_now # this used to be pinned, so any cached opamroot needs it unpinned
    opamrun pin add --switch "$do_pins_NAME"  --yes --no-action -k version with-dkml "${PIN_WITH_DKML}"
    section_end "opam-pins-$do_pins_NAME"
}

do_pins dkml
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_pins two
fi

do_use_vsstudio() {
    do_use_vsstudio_NAME=$1
    shift
    case "$dkml_host_abi" in
    windows_*)
        section_begin "use-vsstudio-$do_use_vsstudio_NAME" "Use Visual Studio in dkml-* Opam packages (Windows) for $do_use_vsstudio_NAME switch"

        # shellcheck disable=SC2153
        E_VS_DIR=$(escape_arg_as_ocaml_string "$VS_DIR")
        # shellcheck disable=SC2153
        E_VS_VCVARSVER=$(escape_arg_as_ocaml_string "$VS_VCVARSVER")
        # shellcheck disable=SC2153
        E_VS_WINSDKVER=$(escape_arg_as_ocaml_string "$VS_WINSDKVER")
        # shellcheck disable=SC2153
        E_VS_MSVSPREFERENCE=$(escape_arg_as_ocaml_string "$VS_MSVSPREFERENCE")
        # shellcheck disable=SC2153
        E_VS_CMAKEGENERATOR=$(escape_arg_as_ocaml_string "$VS_CMAKEGENERATOR")

        case "$(opamrun --version)" in
        2.0.*)
            if [ "${in_docker}" = "true" ]; then
                echo Opam 2.0 support in dockcross to use a portable opam var prefix not yet implemented
                exit 67
            fi
            OP=$(opamrun var prefix --switch "$do_use_vsstudio_NAME")
            OPSC=$OP/.opam-switch/switch-config
            if grep setenv: "$OPSC"; then
                echo "INFO: Updating switch-config. Old was:"
                awk '{print ">> " $0}' "$OPSC"

                awk '$1=="setenv:"{x=1} x==0{print} x==1 && $0=="]"{x=0}' "$OPSC" >"$OPSC".trimmed
                mv "$OPSC".trimmed "$OPSC"
            fi
            echo 'setenv: [' >>"$OPSC"
            echo '  [DKML_COMPILE_SPEC = "1"]' >>"$OPSC"
            echo '  [DKML_COMPILE_TYPE = "VS"]' >>"$OPSC"
            echo "  [DKML_COMPILE_VS_DIR = \"$E_VS_DIR\"]" >>"$OPSC"
            echo "  [DKML_COMPILE_VS_VCVARSVER = \"$E_VS_VCVARSVER\"]" >>"$OPSC"
            echo "  [DKML_COMPILE_VS_WINSDKVER = \"$E_VS_WINSDKVER\"]" >>"$OPSC"
            echo "  [DKML_COMPILE_VS_MSVSPREFERENCE = \"$E_VS_MSVSPREFERENCE\"]" >>"$OPSC"
            echo "  [DKML_COMPILE_VS_CMAKEGENERATOR = \"$E_VS_CMAKEGENERATOR\"]" >>"$OPSC"
            echo "  [DKML_HOST_ABI = \"${dkml_host_abi}\"]" >>"$OPSC"
            echo ']' >>"$OPSC"
            cat "$OPSC" >&2 # print
            ;;
        *)
            opamrun option --switch "$do_use_vsstudio_NAME" setenv= # reset
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+='DKML_COMPILE_SPEC = "1"'
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+='DKML_COMPILE_TYPE = "VS"'
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+="DKML_COMPILE_VS_DIR = \"$E_VS_DIR\""
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+="DKML_COMPILE_VS_VCVARSVER = \"$E_VS_VCVARSVER\""
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+="DKML_COMPILE_VS_WINSDKVER = \"$E_VS_WINSDKVER\""
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+="DKML_COMPILE_VS_MSVSPREFERENCE = \"$E_VS_MSVSPREFERENCE\""
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+="DKML_COMPILE_VS_CMAKEGENERATOR = \"$E_VS_CMAKEGENERATOR\""
            opamrun option --switch "$do_use_vsstudio_NAME" setenv+="DKML_HOST_ABI = \"${dkml_host_abi}\""
            opamrun option --switch "$do_use_vsstudio_NAME" setenv # print
            ;;
        esac

        # shellcheck disable=SC2016
        opamrun exec --switch "$do_use_vsstudio_NAME" -- sh -c 'echo $VCToolsRedistDir'

        section_end "use-vsstudio-$do_use_vsstudio_NAME"
        ;;
    esac
}
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_use_vsstudio two
fi
do_use_vsstudio dkml

# Because dune.X.Y.Z+shim (and any user DKML packages) requires DKML installed (after all, it is just
# a with-dkml.exe shim), we need either dkmlvars-v2.sexp or DKML environment
# variables. Confer: Dkml_runtimelib.Dkml_context.get_dkmlversion
#
# grep matches either:
#   [... [DiskuvOCamlVersion = "1.0.1"] ...]
#   DiskuvOCamlVersion = "1.0.1"
do_setenv() {
    do_setenv_SWITCH=$1
    shift
    section_begin "setenv-$do_setenv_SWITCH" "Set opam option for $do_setenv_SWITCH switch"
    opamrun option --switch "$do_setenv_SWITCH" setenv > ".ci/sd4/setenv.$do_setenv_SWITCH.txt"
    if ! grep -q '\(^|\[\)DiskuvOCamlVarsVersion ' ".ci/sd4/setenv.$do_setenv_SWITCH.txt"; then
        opamrun option --switch "$do_setenv_SWITCH" setenv+='DiskuvOCamlVarsVersion = "2"'
    fi
    if ! grep -q '\(^|\[\)DiskuvOCamlVersion ' ".ci/sd4/setenv.$do_setenv_SWITCH.txt"; then
        opamrun option --switch "$do_setenv_SWITCH" setenv+="DiskuvOCamlVersion = \"$DKML_VERSION\""
    fi
    if [ "$do_setenv_SWITCH" = dkml ] && [ -n "${DKML_HOME:-}" ]; then
      do_setenv_DKMLHOME_ESCAPED="$DKML_HOME"
      do_setenv_USRBIN_ESCAPED="$DKML_HOME/usr/bin"
      do_setenv_BIN_ESCAPED="$DKML_HOME/bin"
      if [ -x /usr/bin/cygpath ]; then
        do_setenv_DKMLHOME_ESCAPED=$(/usr/bin/cygpath -aw "$do_setenv_DKMLHOME_ESCAPED" | sed 's/\\/\\\\/g')
        do_setenv_USRBIN_ESCAPED=$(/usr/bin/cygpath -aw "$do_setenv_USRBIN_ESCAPED" | sed 's/\\/\\\\/g')
        do_setenv_BIN_ESCAPED=$(/usr/bin/cygpath -aw "$do_setenv_BIN_ESCAPED" | sed 's/\\/\\\\/g')
      fi
      if ! grep -q '\(^|\[\)DiskuvOCamlHome ' ".ci/sd4/setenv.$do_setenv_SWITCH.txt"; then
          opamrun option --switch "$do_setenv_SWITCH" setenv+="DiskuvOCamlHome = \"$do_setenv_DKMLHOME_ESCAPED\""
      fi
      if ! grep -q '\(^|\[\)DiskuvOCamlBinaryPaths ' ".ci/sd4/setenv.$do_setenv_SWITCH.txt"; then
          opamrun option --switch "$do_setenv_SWITCH" setenv+="DiskuvOCamlBinaryPaths = \"$do_setenv_USRBIN_ESCAPED;$do_setenv_BIN_ESCAPED\""
      fi
      if ! grep -q '\(^|\[\)DiskuvOCamlDeploymentId ' ".ci/sd4/setenv.$do_setenv_SWITCH.txt"; then
          opamrun option --switch "$do_setenv_SWITCH" setenv+="DiskuvOCamlDeploymentId = \"setup-dkml-switch-$do_setenv_SWITCH\""
      fi
    fi
    case "${dkml_host_abi}" in
    windows_*)
        if ! grep -q '\(^|\[\)DiskuvOCamlMSYS2Dir ' ".ci/sd4/setenv.$do_setenv_SWITCH.txt"; then
            if [ -x /usr/bin/cygpath ]; then
                MSYS2_DIR_NATIVE=$(/usr/bin/cygpath -aw /)
            else
                # If we are already inside MSYS2 then MSYSTEM_PREFIX should be set. But cygpath should be there as well!!
                echo "FATAL: Could not locate MSYS2: there was no cygpath" >&2
                exit 3
            fi
            MSYS2_DIR_NATIVE_ESCAPED=$(printf "%s" "$MSYS2_DIR_NATIVE" | sed 's/\\/\\\\/g')
            opamrun option --switch "$do_setenv_SWITCH" setenv+="DiskuvOCamlMSYS2Dir = \"$MSYS2_DIR_NATIVE_ESCAPED\""
        fi
    esac
    section_end "setenv-$do_setenv_SWITCH"
}
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_setenv two
fi
do_setenv dkml

do_install_compiler() {
    do_install_compiler_NAME=$1
    shift
    section_begin "install-compiler-$do_install_compiler_NAME" "Install OCaml compiler for $do_install_compiler_NAME switch"
    opamrun pin list --switch "$do_install_compiler_NAME"
    # shellcheck disable=SC2086
    opamrun upgrade --switch "$do_install_compiler_NAME" --yes dkml-base-compiler conf-dkml-cross-toolchain ${ocaml_options:-}
    section_end "install-compiler-$do_install_compiler_NAME"
}
if ! [ "${PRIMARY_SWITCH_SKIP_INSTALL:-}" = "true" ]; then
    do_install_compiler dkml
fi
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_install_compiler two
fi

do_summary() {
    do_summary_NAME=$1
    shift
    section_begin "summary-$do_summary_NAME" "Summary: $do_summary_NAME switch"
    opamrun var --switch "$do_summary_NAME"
    opamrun exec --switch "$do_summary_NAME" -- ocamlc -config
    section_end "summary-$do_summary_NAME"
}
do_summary dkml
if [ "${SECONDARY_SWITCH:-}" = "true" ]; then
    do_summary two
fi

'@
Set-Content -Path ".ci\sd4\run-setup-dkml.sh" -Encoding Unicode -Value $Content
msys64\usr\bin\bash -lc 'dos2unix .ci/sd4/run-setup-dkml.sh'

$Content = @'
# MSVC environment variables:
# 1. https://docs.microsoft.com/en-us/cpp/build/reference/cl-environment-variables?view=msvc-170
# 2. https://docs.microsoft.com/en-us/cpp/build/reference/linking?view=msvc-170#link-environment-variables (except TMP)
# 3. VCToolsRedistDir: https://docs.microsoft.com/en-us/cpp/windows/redistributing-visual-cpp-files?view=msvc-170#locate-the-redistributable-files
BEGIN{FS="="}
$1=="CL"||$1=="_CL_"||$1=="INCLUDE"||$1=="LIBPATH" {print "export " $0}
$1=="LINK"||$1=="_LINK_"||$1=="LIB"||$1=="PATH"    {print "export " $0}
$1=="VCToolsRedistDir"                             {print "export " $0}

'@
Set-Content -Path ".ci\sd4\msvcenv.awk" -Encoding Unicode -Value $Content
msys64\usr\bin\bash -lc 'dos2unix .ci/sd4/msvcenv.awk'


$Content = @'
{
    # trim leading and trailing space
    sub(/^ */, "");
    sub(/ *$/, "");

    print "export PATH='" $0 "'";
}
'@
Set-Content -Path ".ci\sd4\msvcpath.awk" -Encoding Unicode -Value $Content
msys64\usr\bin\bash -lc 'dos2unix .ci/sd4/msvcpath.awk'

# PowerShell (UTF-16) and Batch (ANSI) scripts


$Content = @'
# Diagnose Visual Studio environment variables (Windows)
# This wastes time and has lots of rows! Only run if "VERBOSE" GitHub input key.
if ( "${env:VERBOSE}" -eq "true" ) {
    if (Test-Path -Path "C:\Program Files (x86)\Windows Kits\10\include") {
        Get-ChildItem "C:\Program Files (x86)\Windows Kits\10\include"
    }
    if (Test-Path -Path "C:\Program Files (x86)\Windows Kits\10\Extension SDKs\WindowsDesktop") {
        Get-ChildItem "C:\Program Files (x86)\Windows Kits\10\Extension SDKs\WindowsDesktop"
    }

    $env:PSModulePath += "$([System.IO.Path]::PathSeparator).ci\sd4\g\dkml-runtime-distribution\src\windows"
    Import-Module Machine

    $allinstances = Get-VSSetupInstance
    $allinstances | ConvertTo-Json -Depth 5
}

# Make export expression [SN]NAME=[SV]VALUE[EV]
# where [SN] is start name and [SV] and [EV] are start and end value
if (("${env:GITLAB_CI}" -eq "true") -or ("${env:PC_CI}" -eq "true")) {
    # Executed immediately in POSIX shell, so must be a real POSIX shell variable declaration
    $ExportSN = "export "
    $ExportSV = "'"
    $ExportEV = "'"
    $ExportExt = ".sh"
} else {
    # Goes into $env:GITHUB_ENV, so must be plain NAME=VALUE
    $ExportSN = ""
    $ExportSV = ""
    $ExportEV = ""
    $ExportExt = ".github"
}

# Locate Visual Studio (Windows)
if ("${env:vsstudio_dir}" -eq "" -and (!(Test-Path -Path .ci/sd4/vsenv${ExportExt}) -or !(Test-Path -Path .ci/sd4/vsenv.ps1))) {
    $env:PSModulePath += "$([System.IO.Path]::PathSeparator).ci\sd4\g\dkml-runtime-distribution\src\windows"
    Import-Module Machine

    $CompatibleVisualStudios = Get-CompatibleVisualStudios -ErrorIfNotFound
    $CompatibleVisualStudios
    $ChosenVisualStudio = ($CompatibleVisualStudios | Select-Object -First 1)
    $VisualStudioProps = Get-VisualStudioProperties -VisualStudioInstallation $ChosenVisualStudio
    $VisualStudioProps

    Write-Output "${ExportSN}VS_DIR=${ExportSV}$($VisualStudioProps.InstallPath)${ExportEV}" > .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_VCVARSVER=${ExportSV}$($VisualStudioProps.VcVarsVer)${ExportEV}" >> .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_WINSDKVER=${ExportSV}$($VisualStudioProps.WinSdkVer)${ExportEV}" >> .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_MSVSPREFERENCE=${ExportSV}$($VisualStudioProps.MsvsPreference)${ExportEV}" >> .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_CMAKEGENERATOR=${ExportSV}$($VisualStudioProps.CMakeGenerator)${ExportEV}" >> .ci/sd4/vsenv${ExportExt}

    Write-Output "`$env:VS_DIR='$($VisualStudioProps.InstallPath)'" > .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_VCVARSVER='$($VisualStudioProps.VcVarsVer)'" >> .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_WINSDKVER='$($VisualStudioProps.WinSdkVer)'" >> .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_MSVSPREFERENCE='$($VisualStudioProps.MsvsPreference)'" >> .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_CMAKEGENERATOR='$($VisualStudioProps.CMakeGenerator)'" >> .ci/sd4/vsenv.ps1
}

# Link to hardcoded Visual Studio (Windows)
if ("${env:vsstudio_dir}" -ne "") {
    Write-Output "${ExportSN}VS_DIR=${ExportSV}${env:vsstudio_dir}${ExportEV}" > .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_VCVARSVER=${ExportSV}${env:vsstudio_vcvarsver}${ExportEV}" >> .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_WINSDKVER=${ExportSV}${env:vsstudio_winsdkver}${ExportEV}" >> .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_MSVSPREFERENCE=${ExportSV}${env:vsstudio_msvspreference}${ExportEV}" >> .ci/sd4/vsenv${ExportExt}
    Write-Output "${ExportSN}VS_CMAKEGENERATOR=${ExportSV}${env:vsstudio_cmakegenerator}${ExportEV}" >> .ci/sd4/vsenv${ExportExt}

    Write-Output "`$env:VS_DIR='${env:vsstudio_dir}'" > .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_VCVARSVER='${env:vsstudio_vcvarsver}'" >> .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_WINSDKVER='${env:vsstudio_winsdkver}'" >> .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_MSVSPREFERENCE='${env:vsstudio_msvspreference}'" >> .ci/sd4/vsenv.ps1
    Write-Output "`$env:VS_CMAKEGENERATOR='${env:vsstudio_cmakegenerator}'" >> .ci/sd4/vsenv.ps1
}

'@
Set-Content -Path ".ci\sd4\config-vsstudio.ps1" -Encoding Unicode -Value $Content


$Content = @'
@ECHO OFF

REM The OCaml dkml-base-compiler will compile fine but any other
REM packages (ocamlbuild, etc.) which
REM need a native compiler will fail without the MSVC compiler in the
REM PATH. There isn't a `with-dkml.exe` alternative available at
REM this stage of the GitHub workflow.
SET VSCMD_DEBUG=2
SET VSCMD_SKIP_SENDTELEMETRY=1
call "%VS_DIR%\Common7\Tools\VsDevCmd.bat" -no_logo -host_arch=%vsstudio_hostarch% -arch=%vsstudio_arch% -vcvars_ver=%VS_VCVARSVER% -winsdk=%VS_WINSDKVER%
if %ERRORLEVEL% neq 0 (
    echo.
    echo.The "%VS_DIR%\Common7\Tools\VsDevCmd.bat" command failed
    echo.with exit code %ERRORLEVEL%.
    echo.
    exit /b %ERRORLEVEL%
)

REM VsDevCmd.bat turns off echo; be explicit if we want it on or off
@echo OFF

REM MSVC environment variables in Unix format.
echo %PATH% > .ci\sd4\msvcpath


REM * We can't use `bash -lc` directly to query for all MSVC environment variables
REM   because it stomps over the PATH. So we are inside a Batch script to do the query.
msys64\usr\bin\bash -lc "set | grep -v '^PATH=' | awk -f .ci/sd4/msvcenv.awk > .ci/sd4/msvcenv"
'@
Set-Content -Path ".ci\sd4\get-msvcpath-into-msys2.bat" -Encoding Default -Value $Content

msys64\usr\bin\bash -lc "sh .ci/sd4/run-checkout-code.sh PC_PROJECT_DIR '${env:PC_PROJECT_DIR}'"
if ($LASTEXITCODE -ne 0) {
  throw "run-checkout-code.sh failed"
}

# Diagnose Visual Studio environment variables (Windows)
# This wastes time and has lots of rows! Only run if "VERBOSE" GitHub input key.

If ( "${env:VERBOSE}" -eq "true" ) {
  if (Test-Path -Path "C:\Program Files (x86)\Windows Kits\10\include") {
    Get-ChildItem "C:\Program Files (x86)\Windows Kits\10\include"
  }
  if (Test-Path -Path "C:\Program Files (x86)\Windows Kits\10\Extension SDKs\WindowsDesktop") {
    Get-ChildItem "C:\Program Files (x86)\Windows Kits\10\Extension SDKs\WindowsDesktop"
  }

  $env:PSModulePath += "$([System.IO.Path]::PathSeparator).ci\sd4\g\dkml-runtime-distribution\src\windows"
  Import-Module Machine

  $allinstances = Get-VSSetupInstance
  $allinstances | ConvertTo-Json -Depth 5
}
.ci\sd4\config-vsstudio.ps1
msys64\usr\bin\bash -lc "dos2unix .ci/sd4/vsenv.sh"
Get-Content .ci/sd4/vsenv.sh
Get-Content .ci/sd4/vsenv.ps1

# Capture Visual Studio compiler environment
& .ci\sd4\vsenv.ps1
& .ci\sd4\get-msvcpath-into-msys2.bat
msys64\usr\bin\bash -lc "cat .ci/sd4/msvcpath | tr -d '\r' | cygpath --path -f - | awk -f .ci/sd4/msvcpath.awk >> .ci/sd4/msvcenv"    
msys64\usr\bin\bash -lc "tail -n100 .ci/sd4/msvcpath .ci/sd4/msvcenv"

msys64\usr\bin\bash -lc "sh .ci/sd4/run-setup-dkml.sh PC_PROJECT_DIR '${env:PC_PROJECT_DIR}'"
if ($LASTEXITCODE -ne 0) {
  throw "run-setup-dkml.sh failed"
}

########################### script ###############################

Write-Host @"
Finished setup.

To continue your testing, run in PowerShell:
  `$env:CHERE_INVOKING = "yes"
  `$env:MSYSTEM = "$env:msys2_system"
  `$env:dkml_host_abi = "$env:dkml_host_abi"
  `$env:abi_pattern = "$env:abi_pattern"
  `$env:opam_root = "$env:opam_root"
  `$env:exe_ext = "$env:exe_ext"

Now you can use 'opamrun' to do opam commands like:

  msys64\usr\bin\bash -lc 'PATH="`$PWD/.ci/sd4/opamrun:`$PATH"; opamrun install XYZ.opam'
  msys64\usr\bin\bash -lc 'PATH="`$PWD/.ci/sd4/opamrun:`$PATH"; opamrun exec -- bash'
  msys64\usr\bin\bash -lc 'sh ci/build-test.sh'
"@

#!/usr/bin/env bash

version=4175d4cc6bb2a99b93e993cdb47e43fc8d27acfa

set -e -o pipefail

TMP="$(mktemp -d)"
trap "rm -rf $TMP" EXIT

PACKAGE=fmt

rm -rf $PACKAGE
mkdir -p $PACKAGE/src

(
    cd $TMP
    git clone https://github.com/ocaml-dune/$PACKAGE.git
    cd $PACKAGE
    git checkout $version
)

SRC=$TMP/$PACKAGE

cp -v $SRC/LICENSE.md $PACKAGE/
cp -v -R $SRC/src/fmt.ml{i,} $PACKAGE/src/

git checkout $PACKAGE/src/dune
git add -A .

---
title: "Vignette for Dune: merging clusters to improve replicability through ARI merging"
author: "Hector Roux de Bézieux"
output: 
  rmarkdown::html_document:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Dune Vignette}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Installation

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}
BiocManager::install("Dune")
```


We use a subset of the Allen Smart-Seq nuclei dataset. Run `?Dune::nuclei` for more details on pre-processing.

```{r setup}
suppressPackageStartupMessages({
  library(RColorBrewer)
  library(dplyr)
  library(ggplot2)
  library(tidyr)
  library(knitr)
  library(purrr)
  library(Dune)
})
data("nuclei", package = "Dune")
theme_set(theme_classic())
```

# Initial visualization

We have a dataset of $1744$ cells, with the results from 3 clustering algorithms: Seurat3, Monocle3 and SC3. The Allen Institute also produce hand-picked cluster and subclass labels. Finally, we included the coordinates from a t-SNE representation, for visualization.


```{r}
ggplot(nuclei, aes(x = x, y = y, col = subclass_label)) +
  geom_point()
```

We can also see how the three clustering algorithm partitioned the dataset initially:

```{r, fig.hold='hold', out.width="33%", fig.height=9}
walk(c("SC3", "Seurat", "Monocle"), function(clus_algo){
  df <- nuclei
  df$clus_algo <- nuclei[, clus_algo]
  p <- ggplot(df, aes(x = x, y = y, col = as.character(clus_algo))) +
    geom_point(size = 1.5) +
    # guides(color = FALSE) +
    labs(title = clus_algo, col = "clusters") +
    theme(legend.position = "bottom")
  print(p)
})
```

# Merging with **Dune**
## Initial ARI

The adjusted Rand Index between the three methods can be computed.

```{r}
plotARIs(nuclei %>% select(SC3, Seurat, Monocle))
```

As we can see, the ARI between the three methods is initially quite low.

## Actual merging

We can now try to merge clusters with the `Dune` function. At each step, the algorithm will print which clustering label is merged (by its number, so `1~SC3` and so on), as well as the pair of clusters that get merged.

```{r}
merger <- Dune(clusMat = nuclei %>% select(SC3, Seurat, Monocle), verbose = TRUE)
```

The output from `Dune` is a list with four components:

```{r}
names(merger)
```

`initialMat` is the initial matrix. of cluster labels. `currentMat` is the final matrix of cluster labels. `merges` is a matrix that recapitulates what has been printed above, while `ImpARI` list the ARI improvement over the merges.

## ARI improvement

We can now see how much the ARI has improved:

```{r}
plotARIs(clusMat = merger$currentMat)
```

The methods now look much more similar, as can be expected.

We can also see how the number of clusters got reduced.

```{r}
plotPrePost(merger)
```

For SC3 for example, we can visualize how the clusters got merged:

```{r}
ConfusionPlot(merger$initialMat[, "SC3"], merger$currentMat[, "SC3"]) +
  labs(x = "Before merging", y = "After merging")
```

Finally, the __ARIImp__ function tracks mean ARI improvement as pairs of clusters get merged down.

```{r}
ARItrend(merger)
```

# Session

```{r}
sessionInfo()
```

---
name: Bug report
about: File an issue to help us improve
title: ''
labels: ''
assignees: ''

---

<!-- Thank you for filing an issue to help us improve Dune! -->
## Expected Behavior


## Actual Behavior


## Reproduction

<!-- If possible, please open a PR with a test that reproduces the problem.
      Adding tests is easy, see  https://github.com/ocaml/dune/blob/master/doc/hacking.rst -->

- PR with a reproducing test: 

<!-- Otherwise, please detail how the behavior can be reproduced -->

1.
1.
1.

## Specifications

- Version of `dune` (output of `dune --version`):
- Version of `ocaml` (output of `ocamlc --version`)
- Operating system (distribution and version):

<!-- remove the following if not used -->
## Additional information 

- Link to gist with verbose output (run `dune` with the `--verbose` flag):

# Hacking on Dune

This section is for people who want to work on Dune itself.

## Bootstrap

In order to build itself, Dune uses an OCaml script
([bootstrap.ml](bootstrap.ml)) that dumps most of the sources of Dune into a
single `boot.ml` file. This file is built using `ocamlopt` or `ocamlc`
and used to build everything else.

To get the development version bootstrapped and installed, run the following
while in the project's root directory:

```sh
$ make          # create the dune.exe file and bootstrap the dune build
$ make install  # install the newly built dune
```

Note that we don't include all of the sources in boot.ml. We skip a
few parts to speed up the build. In particular:
- vendored libraries are replaced by simpler implementations taken
  from `vendor/boot`
- a few files in `src` have an alternative version. These alternatives
  versions are named `XXX.boot.EXT`. For instance: `glob_lexer.boot.ml`

## OCaml compatibility test

Install opam switches for all the entries in the
[dune-workspace.dev](dune-workspace.dev) file and run:

```sh
$ make all-supported-ocaml-versions
```

## Repository organization

- `vendor/` contains dependencies of Dune, that have been vendored
- `plugin/` contains the API given to `dune` files that are OCaml
  scripts
- `src/` contains the core of `Dune`
- `bin/` contains the command line interface
- `doc/` contains the manual and rules to generate the manual pages

## Design

Dune (nee "JBuilder") was initially designed to sort out the public release of
Jane Street packages which became incredibly complicated over time. It is still
successfully used for this purpose.

One necessary feature to achieve this is the ability to precisely
report the external dependencies necessary to build a given set of
targets without running any command, just by looking at the source
tree. This is used to automatically generate the `<package>.opam`
files for all Jane Street packages.

To implement this, the build rules are described using a build arrow,
which is defined in [src/build.mli](src/build.mli). In the end it makes the
development of the internal rules of Dune very composable and
quite pleasant.

To deal with process multiplexing, Dune uses a simplified
Lwt/Async-like monad, implemented in [src/fiber/fiber.mli](src/fiber/fiber.mli).

## Tests

Dune uses [cram style](https://blog.janestreet.com/testing-with-expectations/)
tests for its test suite. The test suite is contained in
[test/blackbox-tests](test/blackbox-tests). A single test consists of a test
directory in the test-cases/ sub directory which contains a run.t file defining
the test.

An example `run.t` file:

```
A description of the test. The command running the tests is preceeded by
two spaces and a $. The expected output of the command is also indented by
two spaces and is right below the command (note the absence of a $)

  $ echo "the expected output is below"
  the expected output is below
```

Running the entire test suite is done with `$ make test`. A particular test can
be executed with `$ dune build @<test-name>`.

Running the test will execute the command after the `$` and its output will be
compared against the expected output right below the command. Any differences
will result in a test failure.

### Adding a Test

Simply add a new directory in test/blackbox-tests/test-cases and then `$ make`
generate the rules for the test , followed by `$ make promote` to accept the new
rules.

### Accepting Corrections

A failing expect test will generate a diff between the expected and actual
output. If the new output generate by the command is satisfactory, it can be
*promoted* with the `$ make promote` command.

## Code flow

- [src/dune_file.mli](src/dune_file.mli) contains the internal representation
  of `dune` files and the parsing code
- [src/jbuild_load.mli](src/jbuild_load.mli) contains the code to scan
  a source tree and build the internal database by reading
  the `dune` files
- [src/gen_rules.mli](src/gen_rules.mli) contains all the build rules
  of Dune
- [src/build_system.mli](src/build_system.mli) contains a trivial
  implementation of a Build system. This is what Jenga will provide
  when implementing the bridge

---
name: Feature request
about: Suggest an idea to improve dune
title: ''
labels: ''
assignees: ''

---

<!-- Thank you for filing an issue to help us improve Dune! -->
## Desired Behavior

<!-- Please describe the desired behavior in as much detail as you can. This
     should include the context in which you think this would be useful. -->

## Example

<!-- Please provide a concrete example of the proposed functionality. This would
     include a minimal working example of the files you're trying to build with.
     -->

# Custom Repositories

By default when locking package versions, Dune looks up packages from two
sources:

  1. The upstream, community maintained `opam-repository` at
     [ocaml/opam-repository](https://github.com/ocaml/opam-repository) for most
     packages
  2. An overlay repository with patched software to allow usage in Dune at
     [ocaml-dune/opam-overlays](https://github.com/ocaml-dune/opam-overlays)

To change the presets, the `dune-workspace` file has to be edited (and created
if it didn't exist):

::::{dropdown} `dune-workspace`
:icon: file-code

:::{literalinclude} repos/dune-workspace
:language: dune
:::

::::

In this case, we want to select a specific revision of the community repository
instead of always using the most recent one as it would do by default. We
define a new repository and configure the lock directory to use this
repository.

When relocking the dependencies, the list of packages that are found as
dependencies changes accordingly:

```
$ dune pkg lock
Solution for dune.lock:
- base-unix.base
- fmt.0.9.0
- ocaml.5.0.0
- ocaml-base-compiler.5.0.0
- ocaml-config.3
- ocamlbuild.0.15.0+dune
- ocamlfind.1.9.6+dune
- topkg.1.0.
```

Compared to before, the OCaml compiler version is older, which shows
that we did indeed pick an older version of the package repository for locking.

:::{note}
This feature can also be used to make sure the locked dependencies are
reproducible, as fixing all the package repository versions will lead to
equivalent locking results.
:::

# Rule production

This document describes how rule production works in Dune. It was originally
written by Jérémie Dimino as part of the
[streaming RFC](https://github.com/ocaml/dune/pull/5251), but moved
into the dev documentation as it provides a great overview on how this part of
Dune works at present.

## How does rule production works?

### `Dune_engine.Load_rules`

The production of rules is driven by the module `Load_rules` in the
`dune_engine` library. This library is the build system core of
Dune. It is meant as a general purpose library for writing build
systems, and the Dune software is built on top of it. In theory,
`dune_engine` shouldn't know about `dune` or `dune-project`
files. However, for historical reason this is not the case yet and
`dune_engine` still knows some things about them.

As we work on Dune, we expect that `dune_engine` will become more and
more agnostic. Even though it is not completely agnostic, we have
successfully been using it to build Jane Street code base, using the
Jane Street rules on top of this core. So it's already more general
than Dune itself.

For the purpose of this design doc, we will treat `dune_engine` as a
completely general library that doesn't know about `dune` files.

The main feature of `Load_rules` is the `Load_rules.load_dir` function:

```ocaml
val load_dir : dir:Path.t -> Loaded.t Memo.t
```

`Loaded.t` represents a "loaded" set of rules for a particular
directory. It can also be thought as a "compiled" set of rules. A
`Loaded.t` contains all the rules in existence that produce targets in
`dir`. For instance, given a `Loaded.t` we can figure out all the
files that would be produced in `dir` if we were building everything
that could be built. While `dir` is a build directory, this also
includes files present in the source tree. This is because
`Load_rules.load_dir` implicitly adds copy rules for all source files
present in the source directory that correspond to the build directory
`dir`. For instance, if `dir` is `_build/default/src`, Dune will
implicitly add rules to copy files in `src` to `_build/default/src`.
Except for rules that have the special `promote` or `fallback` modes.

This is in fact how Dune evaluates globs during the build. Indeed,
when writing `dune` files we work in an imaginary world where both the
source files and the generated files are present. So when we write
`(deps (glob_files *.txt))`, this `*.txt` denotes both `.txt` files
that are present on disk in the source tree but also as the ones that
can be generated by the build.

In practice, to evaluate `(glob_files *.txt)` in directory `d`, Dune
calls `Load_rules.load_dir ~dir:d` and filter the list of files that can
be built. Similarly, when Dune needs to build a file
`_build/default/src/x`, it first calls `Load_rules.load_dir` with
`_build/default/src` and then looks up a rule that has `x` has
target in the returned `Loaded.t`. The `Load_rules.load_dir` is
memoised, so it can be called multiple times during the build without
guilt.

While `Load_rules` is responsible for driving the production of rules,
it is part of `dune_engine` which doesn't know about `dune` files and
doesn't know about OCaml libraries or OCaml compilation in general. So
it is not responsible for actually producing the build rules that
allow to build Dune projects. Instead, `Load_rules` defers the actual
production of rules to a callback that it obtains via
`Build_config`. Inside Dune, this callback is implemented by the
`Gen_rules` module inside the `dune_rules` library. `dune_rules` is
the library that is responsible for parsing, interpreting and
compiling `dune` files down to low-level build rules.

### `Dune_rules.Gen_rules`

The entry of `Dune_rules.Gen_rules` is the `gen_rules` function. Its
API looks like:

```ocaml
val gen_rules :
  Build_config.Context_or_install.t ->
  dir:Path.Build.t ->
  string list ->
  Build_config.gen_rules_result Memo.t
```

Where `Build_config.gen_rules_result` is, in most cases —when the value
returned is `Build_config.Rules _`—, a "raw" set of rules. Raw in the sense
that there is no overlap checks or any other checks. During the rule
production phase, we merely accumulate a set of rules that is later
processed. The API of `gen_rules` is in fact a bit more complex, but
the above definition is enough for the purpose of this document.

The first thing `gen_rules` does is analyse the directory it is
given. If the directory corresponds to a source directory with a `dune`
file, `gen_rules` will dispatch the call to the part of `dune_rules`
that parses and interprets the `dune` file. This is the simplest case,
but even in this case there are some things worth mentioning.

For instance, when compiling an OCaml library dune stores the
artifacts for the library in generated dot-directories. For instance,
the cmi files for library `foo` living in source directory `src` will
end up in `_build/default/src/.foo.objs/byte`. We could produce these
rules when `gen_rules` is called with directory
`_build/default/src/.foo.objs/byte`, however that would spread out the
logic for interpreting `library` stanzas. It is much simpler to
produce all the build rules corresponding to a `library` stanza in one
go. This is what is happening at the moment: when called with
directory `_build/default/src`, `gen_rules` will not only produce
rules for this directory but will also produce rules for
`_build/default/src/.foo.objs/byte` and various other directories.

`Load_rules` doesn't know anything about this. And in particular, it
doesn't know that it is the `gen_rules` call for directory
`_build/default/src/` that will produce the rules for the dot
subdirectories. When `Load_rules` loads the rules for the
`.../.foo.objs/byte` sub-directory, it simply calls `gen_rules` with
this directory. It is `gen_rules` that "redirects" the call to the
`_build/default/src` directory by calling
`Load_rules.load_dir_and_produce_its_rules`. This function simply
calls `Load_rules.load_dir` and re-emits all the raw rules that were
returned by the corresponding `gen_rules` call.

This works because `Load_rules.load_dir` accepts the facts that
`gen_rules` produces rules for many directory at once. It simply
filters out the result. But for things to behave well, the unwritten
following invariant must hold: `gen_rules ~dir:d` is allowed to
generate rules for directory `d'` iff `gen_rules ~dir:d'` emits a call
to `Load_rules.load_dir ~dir:d`.

This scenario happens in a number of cases. All these cases share a
common pattern: the redirections are always to an ancestor
directory. At the moment, there is one exception to this pattern in
the odoc rules, however it is easy to remove.

Finally, the `copy_files` stanza creates another form of dependency
between directory. In order to calculate the targets produced by
`copy_files`, which needs to be known at rule production time, we need
to evaluate the glob given to `copy_files`. Which requires doing a
call to `Load_rules.load_dir` as previously described. Contrary to the
other form of dependency we just describe, this ones can go from any
directory to any other directory. For instance, the following stanza
in `src/dune`:

```
(copy_files foo/*.txt)
```

would create a dependency from `_build/default/src` to
 `_build_default/src/foo`.

So in the end, if we were looking at the internal computation graph of
Dune and narrowing it to just the calls to `Load_rules.load_dir`, we
would see a graph with many edges going from a directory to one of its
ancestor. These would mostly be between generated dot-subdirectories
and their first ancestor that has a corresponding directory in the
source tree. Plus a few other arbitrary ones for each `copy_files`
stanza.

## Directory targets

Before directory targets, answering the question "what rules produces
file X?" was easy. Dune would just call `Load_rules.load_dir` and
lookup `X` in the result. With directory targets, things are a bit
more complicated. Indeed, `X` might also be produced by a directory
target in an ancestor directory. This means that `Load_rules.load_dir`
now need to look in parent directories as well, which introduce more
dependencies from directories to their parents and can create cycles
because of `copy_files` stanza that create dependencies in the other
direction.

At a result, some combinations of `copy_files` and directory targets
don't produce the expected result. This is documented in the test
suite.

# Dune Monorepo Benchmarking

Dune is continuously benchmarked for performance regressions. This includes micro-benchmarks run with `make bench` (refer to the [hacking guide](../hacking.rst)), and macro-benchmarks designed to test performance on large projects, known as "Monorepo benchmarks."

These benchmarks simulate real-world complexities by assembling a monorepository from a large subset of co-installable opam packages. They are essential for assessing the impact of changes on large codebases and providing a realistic performance profile. The metrics monitored include but are not limited to execution times and memory usage.

This document guides you on how to use and visualise the Monorepo benchmarks and details the assembly process of the monorepo.

*Note: For `current-bench` documentation (the tool running Dune benchmarks), visit [current-bench's repository](https://github.com/ocurrent/current-bench/tree/main/doc).*

## Accessing Benchmarks

Benchmark results are available [here](https://bench.ci.dev/ocaml/dune/branch/main?worker=fermat&image=bench%2Fmonorepo%2Fbench.Dockerfile).

[`current-bench`](https://github.com/ocurrent/current-bench/) is configured to automatically run Monorepo benchmarks on every new PR and commit to `main` in the `dune` repository. The dashboard defaults to the `main` branch, and you can adjust the time window in the UI's top right corner for your desired analysis duration.

### Benchmarks on Pull Requests

For every new PR, the benchmarks are automatically run. In the list of checks
on that PR's GitHub page, the benchmark runner adds a link to the dashboard
with the title `ocaml-benchmarks (bench/monorepo/bench.Dockerfile;fermat)`.
You can also use the [direct link to the
dashboard](https://bench.ci.dev/ocaml/dune/branch/main?worker=fermat&image=bench%2Fmonorepo%2Fbench.Dockerfile)
and find the PR number in the left sidebar.

The UI displays the PR's last commit hash used for the comparison.  It's a good
idea to verify that this is the commit you want to compare.

A table of comparison is displayed on the PR branch dashboard, which gives a
comparison of each metric's value on the PR's last commit vs the
`main` branch value. Plus, a delta percentage of the values that would
help you notice any significant changes to drill further into.

There's also a graph comparing metric values that could make it easier to
notice sharp changes in the metrics to further investigate and understand what
causes them.

For more information on how to use `current-bench`, refer to the [current-bench
User
Manual](https://github.com/ocurrent/current-bench/blob/main/doc/user_manual.md).

### Posting Results to the PR

You can opt-in to post benchmark results in the pull request directly. To do so, you can label the pull request with `notify-benchmark-results`. When a PR is tagged with this label, `current-bench` will post benchmark results after every commit.

See [this PR](https://github.com/ocaml/dune/pull/8596#issuecomment-1734763937) as an example.

## How to Run the Benchmarks Locally

To run the benchmarks locally, you need to set up a Docker environment for the
Monorepo benchmark, then generate and mount a `duniverse` directory.

1.  **Build the Docker Image**: From the root directory of your local `dune`
    repository, build the Docker image using the
    `bench/monorepo/bench.Dockerfile`

    ``` console
    $ docker build . -f bench/monorepo/bench.Dockerfile --tag=dune-monorepo-benchmark
    ```

2.  **Clone the Benchmark Repository**: Clone the `ocaml-monorepo-benchmark`
    repository to obtain necessary scripts and files.

    ``` console
    $ git clone https://github.com/ocaml-dune/ocaml-monorepo-benchmark.git
    $ cd ocaml-monorepo-benchmark
    ```

3.  **Generate the `duniverse` Directory**: Use the `generate-duniverse.sh`
    script to create the `duniverse` directory at a desired location (e.g.,
    `/tmp`).

    ``` console
    $ ./generate-duniverse.sh /tmp
    ```

    This will create a directory `/tmp/duniverse` containing all the Monorepo's packages.

4.  **Run the Benchmark**: Execute the Docker container, mounting the
    `duniverse` directory as a volume.

    ``` console
    $ docker run -it --volume=/tmp/duniverse:/home/opam/bench-dir/current-bench-data/duniverse dune-monorepo-benchmark bash --login -c 'make bench'
    ```

The results from the Monorepo benchmark are provided in a JSON format, which can
be parsed by [`current-bench`](https://bench.ci.dev).

## For Developers

Detailed documentation on how the Monorepo is constructed and updated is
available
[here](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/README.md).

### Descriptions of Existing Benchmarks

The benchmark runner runs two types of benchmarks: **one shot benchmarks** and
**watch mode benchmarks**.

#### One-Shot Benchmarks

These run `dune build <TARGET>` and record the time it takes for the command to
complete. Under normal operation, there are two one-shot benchmarks.

##### Build From Scratch

This builds the target assuming a clean state. As this
benchmark can take a long time (> 15 minutes) compared to the expected
noise in benchmark results, it's run only once.

##### Null Build
This builds the target assuming it's starting with the target already built.
This benchmark runs several times to reduce noise. The number of repetitions is
controlled by the `--num-short-job-repeats` argument.

#### Watch Mode Benchmarks

After running the one-shot benchmarks, the benchmark runner will start Dune in
watch mode. Once the initial build is complete, the watch mode benchmarks are
performed. Each watch mode benchmark makes a change to a file in the Monorepo
(which will trigger a rebuild) and measures how long it takes to rebuild the
target.

Each file is changed 3 times. The benchmark runner waits for a rebuild after each change:
 - A benign change is made to a source file.
 - A compile error is introduced into a source file.
 - The compile error is fixed.

There are 2 files which are changed in watch mode benchmarks. The intention was
to change a file in a library with few reverse dependencies and to change a
file in a library with many reverse dependencies:
 - The file `src/path.ml` in the library `file_path`, which has few reverse
   dependencies (within the monorepo)
 - The file `src/list.ml` in the library `base`, which has many reverse
   dependencies (within the monorepo)


### How to Add New Benchmarks

One-shot benchmarks are defined ad-hoc, so there's no structured way to add more,
but see the use of the function `measure_one_shot_build` in [this
file](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/dune-monorepo-benchmark-runner/src/main.ml)
to learn how one might add more one-shot benchmarks.

To add new watch mode benchmarks, edit the file
[dune-monorepo-benchmark-runner/src/watch_mode_scenarios.ml](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/dune-monorepo-benchmark-runner/src/watch_mode_scenarios.ml)
to add new entries to the list. The fields of list entries are documented
[here](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/dune-monorepo-benchmark-runner/src/watch_mode_file_to_change.mli).
Each entry specifies a file to change, some text in the file to be replaced and
two replacement text options -- something which compiles and something which
does not compile during the watch mode benchmark.

#### Make a new release of the monorepo benchmarks
To have Dune start running the new benchmark in `current-bench`, do a new release
of the
[ocaml-monorepo-benchmark](https://github.com/ocaml-dune/ocaml-monorepo-benchmark)
by pushing a tag to the repo named the current date like `YYYY-MM-DD.N` where
`N` is the current day's release number, starting with 0.

Then update the `dune` repository, editing the
[bench/monorepo/bench.Dockerfile](https://github.com/ocaml/dune/blob/main/bench/monorepo/bench.Dockerfile)
file, setting the value of the `MONOREPO_BENCHMARK_TAG` variable to the tag
you just pushed.

Future PRs against Dune will now run the new benchmark.


### Maintenance

Whenever something goes wrong that prevents the benchmarks from running, update
the [Postmortems File](https://github.com/ocaml-dune/ocaml-monorepo-benchmark/blob/main/postmortems.md)
in the [OCaml Monorepo Benchmark](https://github.com/ocaml-dune/ocaml-monorepo-benchmark)
repo by detailing the problem and the fix. It's also worth reading that
file in case the current problem is similar to one that's already been solved
(still update the file in this case though!).

Here are some common problems that affect the monorepo benchmark.

#### Dune Gets a Backwards-Incompatible Update

Sometimes Dune gets an update that changes its contract with projects (e.g.,
[dune/9472](https://github.com/ocaml/dune/pull/9472),
[dune/8293](https://github.com/ocaml/dune/pull/8293)).
Such changes are rare and usually affect a small number of packages, but the
benchmark monorepo depends on a large number of packages, and if a single one
fails to build, the monorepo benchmark will not run.

The fix is to patch any affected packages so that they can be built with Dune
following the breaking change. Add a file to the "benchmark/patches" directory
of the [OCaml Monorepo Benchmark](https://github.com/ocaml-dune/ocaml-monorepo-benchmark).
Patches must be named `x.diff` where `x` is the name of the subdirectory in the Duniverse 
directory that contains the affected package. The patch will be applied
from the package's Duniverse subdirectory before the benchmarks are run.

Then [make a new release of the monorepo
benchmark and updates Dune's monorepo bench Docker file](#make-a-new-release-of-the-monorepo-benchmarks)
so thatDdune can see the new changes.

Be sure to also upstream the patch to the affected package! 

Note that even
though the fix may be upstreamed, the monorepo benchmark still needs the patch
file. This is because it uses a snapshot of the opam repo, so isn't exposed to
package updates.

#### The Source Archive of a Package is no Longer Available

This is only an issue when generating the monorepo Duniverse directory, as the
source for all packages is cached on the benchmark server in a persistent
Duniverse directory.

The source archives for opam packages are not stored in a centralized location.
Each package is free to choose an arbitrary URL for opam to download the source archive
from. This leads to some packages sources becoming unavailable over time due to
sites hosting package sources becoming unavailable, repos on GitHub being
renamed, etc. Due to the number of packages in the monorepo and the infrequency
with which the Duniverse directory is generated, the chance of at least one
package being unavailable while generating the Duniverse directory turns out to
be quite high.

Congratulations, you've just found a broken package that could potentially
affect opam users in the wild.

The fix will be to somehow find the package's source archive and upload it
to [opam Source Archives](https://github.com/ocaml/opam-source-archives) repo.
This repo stores the source archives of package whose original source archive
has become unavailable.

To find the package source archive, first check if it's in the opam cache. If
it's there, it will be at the URL
`https://opam.ocaml.org/cache/<hash-algo>/<first-two-chars-of-hash>/<full-hash>`.
For example:
`https://opam.ocaml.org/cache/md5/04/04f0113150261bae2075381c18c6e60c`.
The hash of a package is included in the output of `opam show <package>`.

If it's not in the opam cache, it might still be in your local opam cache
(assuming you've downloaded the package before). Assuming a standard opam
installation on Unix this will be located at
`~/.opam/download-cache/<hash-algo>/<first-two-chars-of-hash>/<full-hash>`.
If it's not on your machine, ask around to see if it's on someone else's.

Note that to add the package to the opam Source Archives, you will need to name
it according to the convention there. If you got the package source archive from
a cache it will be named after its hash.

Once the PR adding the source archive to the opam Source Archives is merged,
manually update the `monorepo-bench.opam.locked` file in the
`benchmark` directory of the [OCaml Monorepo
Benchmark](https://github.com/ocaml-dune/ocaml-monorepo-benchmark) to use the
new URL. Make sure to use the file's raw URL (see existing examples in
that file).

If no cached copy of the package is available, you'll need to get creative.
Its Git repo might be public and included in the output of `opam show <package>`.
Worst case, you'll need to remove the package from the benchmark monorepo by
removing its entries from the `monorepo-bench.opam`,
`monorepo-bench.opam.locked`, and `dune-project` files in the `benchmark`
directory of the [OCaml Monorepo
Benchmark](https://github.com/ocaml-dune/ocaml-monorepo-benchmark) repo (note
that the `dune-project` file lists libraries, not packages). The package's
reverse dependencies will also need to be removed in this way.

---
geometry: "left=2cm,right=2cm,top=2.4cm,bottom=2.1cm"
bibliography: refs.bib
csl: refs.csl
---

# Memo: an incremental computation library that powers Dune

Andrey Mokhov, Arseniy Alekseyev

*Jane Street, London, United Kingdom*

### Abstract

We present Memo, an incremental computation library that supports a new, faster
and more scalable, file-watching build mode in Dune 3.0. The requirements from
the build systems domain make Memo a unique point in the design space of
incremental computation libraries. Specifically, Memo needs to cope with
concurrency, dynamic dependencies, dependency cycles, and non-determinism;
provide support for efficiently collecting and reporting user-friendly errors;
and scale to computation graphs containing tens of millions of incremental
nodes.

## Introduction

The OCaml build system Dune [@dune] supports a continuous file-watching build
mode, where rebuilds are triggered automatically as the user is editing source
files. A simple and naive way to implement this functionality is to restart Dune
on any file change but that is too slow for large projects. To avoid re-scanning
the project's source tree, re-parsing all build specification files, and
re-generating all build rules from scratch every time, Dune uses an incremental
computation library called Memo. Below we briefly introduce the Memo's API.

Memo provides a monadic API built on top of the *structured concurrency monad*
`Fiber`. Dune uses `Fiber` to speed up builds by executing external commands in
parallel, e.g., running multiple instances of `ocamlopt` to compile independent
source files. To inject a *fiber* into the `Memo` monad, and to extract it back,
one can use the following pair of functions:

```ocaml
  val of_fiber : 'a Fiber.t -> 'a Memo.t
  val run : 'a Memo.t -> 'a Fiber.t
```

More interestingly, functions in the `Memo` monad can be memoized and cached
between different build runs[^1]:

[^1]: The function `create` takes a few more arguments, e.g., for reporting good
error messages, which we omit for the sake of clarity.

```ocaml
  val create : ('i -> 'o Memo.t) -> ('i, 'o) Memo.Table.t
  val exec : ('i, 'o) Memo.Table.t -> 'i -> 'o Memo.t
```

Here `Memo.Table.t` is a table that stores the input/output mapping computed in
the current run, along with the *dependencies* that are automatically captured
when memoized functions call one another. Such explicit memoization, instead of,
e.g., memoizing every `Memo.map` and `Memo.bind` call, makes it easy to control
the degree of incrementality.

Finally, Memo provides a way to *invalidate* a specific input/output pair, or a
*cell*, via the following API:

```ocaml
  val cell : ('i, 'o) Memo.Table.t -> 'i -> ('i, 'o) Cell.t
  val invalidate : ('i, 'o) Cell.t -> Invalidation.t (* [Invalidation.t]s can be combined *)
  val restart : Invalidation.t -> unit
```

When Dune receives new events from the file-watching backend, it cancels the
build run by interrupting the currently running external commands (if any), and
then calls `restart` to let Memo know which files changed and initiate a new
build run. When evaluating future calls to `run`, Memo will consider all outputs
that transitively depend on the invalidated cells as out of date, and will
recompute them when/if needed.

## Key features

This section discusses the most interesting features of Memo and some aspects of
the implementation.

Firstly, to *capture dependencies*, Memo maintains a call stack, where stack
frames correspond to `exec` calls. The call stack is also used for reporting
good error messages: when a user-supplied function raises an exception, we
extend it with the current Memo stack trace using human-readable annotations
provided to `create` via an optional argument.

Memo supports two ways of *error reporting*: *early* (to show errors to the user
as soon as they occur during a build), and *deterministic* (to provide a stable
error summary at the end of the build). To speed up rebuilds, Memo also supports
*error caching*: by default, it doesn't recompute a `Memo` function that
previously failed if its dependencies are up to date. This behaviour can be
overridden for *non-reproducible errors* that should not be cached, for example,
the errors that occur while Dune cancels the current build by interrupting the
execution of external commands.

One of the most interesting features of Memo compared to other incremental
computation libraries is *dependency cycle detection*.
Since Dune supports *dynamic build dependencies* [@mokhov2020build], the
dependency graph is not known before the build starts. During the build, new
computation nodes and dependency edges are discovered concurrently, and Memo
uses an incremental cycle detection algorithm [@gueneau2019cycles] to detect and
report dependency cycles as soon as they are created. This is a unique feature
of Memo: Incremental [@incremental] and Adapton [@hammer2014adapton] libraries
do not support concurrency; the Tenacious library (used by Jane Street's
internal build system Jenga) does support concurrency but it detects cycles by
"stopping the world" and traversing the "frozen" dependency graph to see if
concurrent computations might have deadlocked by waiting for each other. The
approach used by Tenacious is conceptually simple but introduces a delay between
the creation of a dependency cycle and its detection. Memo reports dependency
cycle errors without delay, and uses the human-readable annotations supplied to
`create` to make it easier to understand and debug cycles.

Compared to the incremental computation libraries mentioned above, the current
implementation of Memo makes one unusual design choice. Incremental, Adapton and
Tenacious are all *push based*, i.e., they trigger recomputation starting from
the leaves of the computation graph. Memo is *pull based*: `invalidate` calls
merely mark leaves (cells) as out of date, but recomputation is driven from the
top-level `run` calls. The main drawback of our approach is that Memo traverses
the whole graph on each rebuild. The main benefits are: (i) Memo
doesn't need to store reverse dependencies, which saves space and eliminates
various garbage collection pitfalls; (ii) the pull based approach makes it
easier to avoid "spurious" recomputations, where a node is recomputed but
subsequently becomes unreachable from the top due to a new dependency structure.
Our current experiments show that traversing the whole graph on each rebuild
isn't prohibitively expensive even for large builds where computation graphs
contain tens of millions of nodes. We may rethink this design decision in future
as Dune and Memo need to scale to larger projects. Having said that, we consider
the pull based approach to incremental computation to be under-researched, and
are keen to investigate how far we can take it in practice.

## Development status

Memo is still in active development and we welcome feedback from the OCaml
community on how to make it better. While the current implementation is tied to
Dune's lightweight concurrency library Fiber, the core functionality can be made
available as a functor over an arbitrary concurrency monad, making it usable
with Async and Lwt.

## Acknowledgements

We thank Jeremie Dimino for driving the design of Memo and for his work on
incrementalising Dune. We are also grateful to Rudi Horn, Rudi Grinberg, Emilio Jesús Gallego
Arias and other Dune developers for their many contributions, and to Armaël
Guéneau for helping us integrate the incremental cycle detection library in
Memo.

# References

<!-- References to be generated by Pandoc. -->

The Development Cycle
=====================

Our calculator now has a better structure, but it still does not do much.

In this chapter, we are going to make several iterations where we add a cram
test, see it fail, implement the missing feature, and repeat. This is how many
OCaml projects are developed (including Dune itself).

:::{note}
There are many ways to do this: tests can be written first or last; some prefer
to `promote` only the fixed version.

In this tutorial, we're going to write tests first and `promote` the erroneous
version but as you become more familiar with Dune, you'll be able to explore
variations of this loop.
:::

## Display Errors

So far, our calculator is not doing any error handling.

We're going to create a test with an error, see how the calculator behaves, and
add a better error message.

### Create a Test

Add a new test in `test/calc.t` with the following content. Note that we do
not specify any output for the command. As before, make sure to include two
spaces before the `$` sign.

```{code-block} cram
:emphasize-lines: 4
  $ calc -e '1+2'
  3

  $ calc -e '1+'
```

Next, run the tests using `dune runtest`.
Dune will display a diff with the actual output:

```diff
   $ calc -e '1+'
+  calc: internal error, uncaught exception:
+        Calc.Parser.MenhirBasics.Error
+        
+  [125]
```

Run `dune promote` and observe that the error message is part of the test.
At that point, running `dune runtest` will succeed.

Our goal for the rest of this section is to make this test display a nice error message.

### Handle the Exception

The message points to an uncaught exception.
Indeed, `Parser.main` can raise `Parser.Error`.
Let's catch this exception in `eval_lb` and display the location of the error
in the input file.

Edit `lib/cli.ml`:

```{code-block} ocaml
:emphasize-lines: 2,5-6
let eval_lb lb =
  try
    let e = Parser.main Lexer.token lb in
    Printf.printf "%d\n" (eval e)
  with Parser.Error ->
    Printf.printf "parse error near character %d" lb.lex_curr_pos
```

:::{note}
This is just an excerpt from the file.
Edit the `eval_lb` to make it look like this.
The modified parts are highlighted.
:::

Now, run the tests again with `dune runtest`. It displays the following diff:

```diff
   $ calc -e '1+'
-  calc: internal error, uncaught exception:
-        Calc.Parser.MenhirBasics.Error
-
-  [125]
+  parse error near character 2
```

Run `dune promote` and note that `test/calc.t` has changed.

Run `dune runtest`. Nothing is displayed, indicated that the test has passed.

:::{note}
This is similar to what happened at the end of {doc}`the previous chapter
<structure>`.

The first `dune runtest` compares the *expected output*
(from `test/calc.t`, the uncaught exception message) to the *actual output*
(our new error message) and displays the diff. So, the uncaught exception
appears as deleted lines (prefixed with `-`) and the new error message appears
as added lines (prefixed with `+`).

Running `dune promote` will copy the last *actual output* to `calc/test.t`.

Running `dune runtest` a second time will compare the *expected output* (the
new message) with the *actual output* of the command (the new message) and find
no difference.
:::

## Add Floats

At this stage, our calculator only supports integers. In this section, we are
going to add support for floating-point numbers and operations.

### Add a Test

First, add a test at the end of `test/calc.t`:

```console
$ calc -e '1+2.5'
```

Run `dune runtest`: this displays an error.

```diff
   $ calc -e '1+2.5'
+  calc: internal error, uncaught exception:
+        Failure("lexing: empty token")
+        
+  [125]
```

Run `dune promote` to update the failing test.
Our goal for the rest of this section is to change that test to print `3.5`.

### Add a `Float` constructor

We need to add a new kind of expression. Let's extend the `exp` type in
`lib/ast.ml` to add a new `Float` constructor.

```{code-block} ocaml
:emphasize-lines: 4
type exp =
| Int of int
| Add of exp * exp
| Float of float
```

With this new constructor, we can represent the `2.5` part as `Float 2.5`.

### Lexing and Parsing

We also need to extend our lexer to produce a new token type for floats, and a
production rule in the grammar.

Let's first add a token type in `lib/parser.mly`:

```{code-block} ocaml
:emphasize-lines: 3
%token<int> Int
%token Plus
%token<float> Float
```

A new rule in `lib/lexer.mll`:

```{code-block} ocaml
:emphasize-lines: 7
rule token = parse
    | eof { Parser.Eof }
    | space { token lexbuf }
    | '\n' { Parser.Eof }
    | '+' { Parser.Plus }
    | digit+ { Parser.Int (int_of_string (Lexing.lexeme lexbuf)) }
    | digit+ '.' digit+ { Parser.Float (float_of_string (Lexing.lexeme lexbuf)) }
```

And a new rule in `lib/parser.mly`:

```{code-block} ocaml
:emphasize-lines: 4
expr:
| Int { Int $1 }
| expr Plus expr { Add ($1, $3) }
| Float { Float $1 }
```

### Evaluation

Let's run `dune build`.

With the new constructor, the compiler is now complaining that the pattern
matching in our `eval` function is incomplete.

```
File "lib/cli.ml", line 1, characters 15-70:
1 | let rec eval = function Ast.Int n -> n | Add (a, b) -> eval a + eval b
                   ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Error (warning 8 [partial-match]): this pattern-matching is not exhaustive.
Here is an example of a case that is not matched:
Float _
```

To fix this, we need to tweak `lib/cli.ml` a bit. Instead of returning an
`int`, let's introduce a `value` type that can represent either `int` or
`float`:

```ocaml
type value = VInt of int | VFloat of float

let value_to_string = function
  | VInt n -> string_of_int n
  | VFloat f -> Printf.sprintf "%.6g" f
```

And update the `eval_lb` function to use this printer:

```{code-block} ocaml
:emphasize-lines: 3-5
let eval_lb lb =
  try
    let expr = Parser.main Lexer.token lb in
    let v = eval expr in
    Printf.printf "%s\n" (value_to_string v)
  with Parser.Error ->
    Printf.printf "parse error near character %d" lb.lex_curr_pos
```

Finally, for `eval` itself, we'll use `(+)` or `(+.)` if both values have the
same type, or convert integers to floats if needed:

```ocaml
let rec eval = function
  | Ast.Int n -> VInt n
  | Float f -> VFloat f
  | Add (a, b) -> (
      match (eval a, eval b) with
      | VInt na, VInt nb -> VInt (na + nb)
      | VFloat fa, VFloat fb -> VFloat (fa +. fb)
      | VInt na, VFloat fb -> VFloat (float na +. fb)
      | VFloat fa, VInt nb -> VFloat (fa +. float nb))
```

With this implementation done, call `dune runtest` and notice that it changes
the output. Call `dune promote` to update the test.

## Pi

In this section, we're going to add named constants, and `pi` in particular.

### Create a test

Add a new test in `test/calc.t`:

```console
  $ calc -e '1+pi'
```

Run `dune runtest` and see the failure.
Run `dune promote` to add the failure to the test file.
Our goal in the rest of the section is to change the output of this test.

### Add a Constructor

Let's add an new constructor in `lib/ast.ml`:

```{code-block} ocaml
:emphasize-lines: 5
type exp =
  | Int of int
  | Add of exp * exp
  | Float of float
  | Ident of string
```

That way, `pi` is going to be represented as `Ident "pi"`.

### Lexing and Parsing

We now need to parse these constants.

Let's add a new token in `lib/parser.mly`:

```{code-block} ocaml
:emphasize-lines: 5
%token Eof
%token<int> Int
%token<float> Float
%token Plus
%token<string> Ident
```

Produce it using a new lexing rule in `lib/lexer.mll`:

```{code-block} ocaml
:emphasize-lines: 3-5,14
let digit = ['0'-'9']

let letter = ['a'-'z']

let ident = letter+

rule token = parse
    | eof { Parser.Eof }
    | space { token lexbuf }
    | '\n' { Parser.Eof }
    | '+' { Parser.Plus }
    | digit+ { Parser.Int (int_of_string (Lexing.lexeme lexbuf)) }
    | digit+ '.' digit+ { Parser.Float (float_of_string (Lexing.lexeme lexbuf)) }
    | ident { Parser.Ident (Lexing.lexeme lexbuf) }
```

And handle it as a new derivation in `lib/parser.mly`:

```{code-block} ocaml
:emphasize-lines: 5
expr:
| Int { Int $1 }
| Float { Float $1 }
| expr Plus expr { Add ($1, $3) }
| Ident { Ident $1 }
```

### Evaluation

Finally, we'll have to update our evaluation function to take the new
constructor into account.

OCaml does not have a value for `pi`, but it has a `Stdlib.acos` function.
Since {math}`\cos{\frac{\pi}{2}} = 0`, we can define `pi` as {math}`2 * \arccos
0` .

```{code-block} ocaml
:emphasize-lines: 10-11
let rec eval = function
  | Ast.Int n -> VInt n
  | Float f -> VFloat f
  | Add (a, b) -> (
      match (eval a, eval b) with
      | VInt na, VInt nb -> VInt (na + nb)
      | VFloat fa, VFloat fb -> VFloat (fa +. fb)
      | VInt na, VFloat fb -> VFloat (float na +. fb)
      | VFloat fa, VInt nb -> VFloat (fa +. float nb))
  | Ident "pi" -> VFloat (2. *. Stdlib.acos 0.)
  | Ident _ -> failwith "unknown ident"
```

Finally, let's run our test. `dune runtest` will display a diff with the new
value. Run `dune promote` to accept it.

## Multiplication

### Create a Test

Let's add a new test in `test/calc.t`:

```cram
  $ calc -e '1+2*3'
```

Run the test with `dune runtest`. Notice the error message.

Let's run `dune promote` to add it to the file.

Now, our goal for the rest of this section is to change that to the expected
result.

### Add a Constructor

Let's update the AST `lib/ast.ml`: we're generalizing addition to binary
operations, and create a new `Mul` operation (notice that we remove the line corresponding to the `Add` expression).

```{code-block} ocaml
:emphasize-lines: 1-3,9
type op =
  | Add
  | Mul

type exp =
  | Int of int
  | Float of float
  | Ident of string
  | Op of op * exp * exp
```

### Lexing and Parsing

Let's add a new token for `*` in `lib/parser.mly`:

```{code-block} ocaml
:emphasize-lines: 3
%token<string> Ident
%token Plus
%token Star
```

Then, we'll produce it using a new rule in `lib/lexer.mll`:

```{code-block} ocaml
:emphasize-lines: 4
    | space { token lexbuf }
    | '\n' { Parser.Eof }
    | '+' { Parser.Plus }
    | '*' { Parser.Star }
```

And use that token in a new rule in `lib/parser.mly` (also modifying the `Add`
rule):

```{code-block} ocaml
:emphasize-lines: 4,5
| Int { Int $1 }
| Float { Float $1 }
| Ident { Ident $1 }
| expr Plus expr { Op (Add, $1, $3) }
| expr Star expr { Op (Mul, $1, $3) }
```

We have a last edit to do here, which is to add a precedence annotation:

```{code-block} ocaml
:emphasize-lines: 2
%left Plus
%left Star
```

### Evaluation

Now, we can update our evaluation function in `lib/cli.ml`:

```{code-block} ocaml
:emphasize-lines: 1-6,13-14
let eval_number_op f_int f_float va vb =
  match (va, vb) with
  | VInt na, VInt nb -> VInt (f_int na nb)
  | VFloat fa, VFloat fb -> VFloat (f_float fa fb)
  | VInt na, VFloat fb -> VFloat (f_float (float_of_int na) fb)
  | VFloat fa, VInt nb -> VFloat (f_float fa (float_of_int nb))

let rec eval = function
  | Ast.Int n -> VInt n
  | Float f -> VFloat f
  | Ident "pi" -> VFloat (2. *. Stdlib.acos 0.)
  | Ident _ -> failwith "unknown ident"
  | Op (Add, a, b) -> eval_number_op ( + ) ( +. ) (eval a) (eval b)
  | Op (Mul, a, b) -> eval_number_op ( * ) ( *. ) (eval a) (eval b)
```

With these updates done, let's run `dune runtest`. It displays that the result
is `7`. Call `dune promote` to update the test.

## Division

This section is going to be very similar to the previous one. Instead we're
adding the division operator.

### Create a Test

Add a test in `test/calc.t`:

```cram
  $ calc -e '4/2'
```

Call `dune runtest`, note the error, call `dune promote`.

### Add a Constructor

Add a constructor in `lib/ast.ml`:

```{code-block} ocaml
:emphasize-lines: 4
type op =
  | Add
  | Mul
  | Div
```

### Lexing and Parsing

Update `lib/parser.mly`:

```{code-block} ocaml
:emphasize-lines: 4
%token<string> Ident;
%token Plus
%token Star
%token Slash
```

Then add the right precedence:

```{code-block} ocaml
:emphasize-lines: 2
 %left Plus
+%left Star Slash
```

And the corresponding rule:

```{code-block} ocaml
:emphasize-lines: 4
| Ident { Ident $1 }
| expr Plus expr { Op (Add, $1, $3) }
| expr Star expr { Op (Mul, $1, $3) }
| expr Slash expr { Op (Div, $1, $3) }
```

Finally, add a lexing rule in `lib/lexer.mll`:

```{code-block} ocaml
:emphasize-lines: 3
    | '+' { Parser.Plus }
    | '*' { Parser.Star }
    | '/' { Parser.Slash }
```

### Evaluation

Add a new case in `lib/cli.ml`:

```{code-block} ocaml
:emphasize-lines: 8
let rec eval = function
  | Ast.Int n -> VInt n
  | Float f -> VFloat f
  | Ident "pi" -> VFloat (2. *. Stdlib.acos 0.)
  | Ident _ -> failwith "unknown ident"
  | Op (Add, a, b) -> eval_number_op ( + ) ( +. ) (eval a) (eval b)
  | Op (Mul, a, b) -> eval_number_op ( * ) ( *. ) (eval a) (eval b)
  | Op (Div, a, b) -> eval_number_op ( / ) ( /. ) (eval a) (eval b)
```

Now, running `dune runtest` should display the right result. Run `dune promote`
to accept it.

## Sine

### Create a Test

Create a new test in `test/calc.t`:

```cram
  $ calc -e 'sin (pi / 6)'
```

Call `dune runtest`, note the error, call `dune promote`.

### Add a Constructor

Add a constructor in `lib/ast.ml`:

```{code-block} ocaml
:emphasize-lines: 6
type exp =
  | Int of int
  | Float of float
  | Ident of string
  | Op of op * exp * exp
  | Call of string * exp
```

### Lexing and Parsing

Update `lib/parser.mly` by defining new tokens for parentheses:

```{code-block} ocaml
:emphasize-lines: 5
%token<string> Ident;
%token Plus
%token Star
%token Slash
%token Lpar Rpar
```

And a production for calls:

```{code-block} ocaml
:emphasize-lines: 5
| Ident { Ident $1 }
| expr Plus expr { Op (Add, $1, $3) }
| expr Star expr { Op (Mul, $1, $3) }
| expr Slash expr { Op (Div, $1, $3) }
| Ident Lpar expr Rpar { Call ($1, $3) }
```

Update `lib/lexer.mll`:

```{code-block} ocaml
:emphasize-lines: 4-5
    | '+' { Parser.Plus }
    | '*' { Parser.Star }
    | '/' { Parser.Slash }
    | '(' { Parser.Lpar }
    | ')' { Parser.Rpar }
```

### Evaluation

We'll need a `float` to pass to `Stdlib.sin`, so let's introduce a conversion functin. We'll also add the corresponding cases in `lib/cli.ml`:

```{code-block} ocaml
:emphasize-lines: 1-3,13-14
let as_float = function
  | VInt n -> float_of_int n
  | VFloat f -> f

let rec eval = function
  | Ast.Int n -> VInt n
  | Float f -> VFloat f
  | Ident "pi" -> VFloat (2. *. Stdlib.acos 0.)
  | Ident _ -> failwith "unknown ident"
  | Op (Add, a, b) -> eval_number_op ( + ) ( +. ) (eval a) (eval b)
  | Op (Mul, a, b) -> eval_number_op ( * ) ( *. ) (eval a) (eval b)
  | Op (Div, a, b) -> eval_number_op ( / ) ( /. ) (eval a) (eval b)
  | Call ("sin", e) -> VFloat (Stdlib.sin (as_float (eval e)))
  | Call _ -> failwith "unknown function"
```

Now, run the tests using `dune runtest`.
Accept the correction with `dune promote`.

## Conclusion

We've added several features to our calculator, and added tests in the meantime.
To do so, we've used `dune runtest` and `dune promote`, two of the most useful
Dune commands.

::::{dropdown} Checkpoint
:icon: location

This is how the project looks like at the end of this chapter.

:::{literalinclude} introduction/dune-project
:caption: dune-project (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/dune
:caption: bin/dune (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/calc.ml
:caption: bin/calc.ml (unchanged)
:language: ocaml
:::

:::{literalinclude} structure/lib/dune
:caption: lib/dune (unchanged)
:language: dune
:::

:::{literalinclude} development-cycle/lib/ast.ml
:caption: lib/ast.ml
:language: ocaml
:::

:::{literalinclude} development-cycle/lib/cli.ml
:caption: lib/cli.ml
:language: ocaml
:::

:::{literalinclude} development-cycle/lib/lexer.mll
:caption: lib/lexer.mll
:language: ocaml
:::

:::{literalinclude} development-cycle/lib/parser.mly
:caption: lib/parser.mly
:language: ocaml
:::

:::{literalinclude} structure/test/dune
:caption: test/dune (unchanged)
:language: dune
:::

:::{literalinclude} development-cycle/test/calc.t
:caption: test/calc.t
:language: cram
:::

::::

Introduction
============

The goal of this first chapter is to get to a place where you have a working
Dune project with a skeleton of a calculator.

This is is a [tutorial](https://diataxis.fr/tutorials/): it is meant to be
followed in order, but you can stop at any point. You can also restart from any
chapter, using these sections that are present at the end of the previous
chapter.

::::{dropdown} Checkpoint
:icon: location
:open:

This will contain the project at the end of each chapter.
::::

## Installing Packages

First, you'll need to have a working Opam installation. This is described in
{doc}`/howto/install-dune`.

Then, create an empty directory somewhere, say `~/dune-calc`. In this tutorial,
we will only create files in this directory.

Let's first make sure you have a working opam installation. Run this command:

```sh
opam --version
```

It should display something like "2.2.0". Anything greater than 2.0.0 is fine.

:::{important}
When asked to type a command, you can click {octicon}`copy` to copy the full
command to your clipboard.

In this tutorial, all commands will be typed at the root of your project, like
`~/dune-calc`.
:::

Let's create a local switch: `cd` to this directory and run the following command.
This can take a few minutes.

```sh
opam switch create ./ 5.1.1
```

This command has created a directory named `_opam` in the current directory.
Now, let's install some packages by running:

```sh
opam install dune.3.15.3 menhir.20231231
```

You can confirm that `opam` is correctly setup by typing `dune --version`,
which should display 3.15.3. Otherwise, please refer to
{doc}`/howto/install-dune`.

:::{note}
The instructions use precise version numbers in `opam install` command. This is
to ensure that the error messages will exactly map what you're seeing, but
it is very likely to work with any version.
:::

## The Calculator Skeleton

Now that we have an opam switch and some packages installed, let's create the
various files.

For each file, click {octicon}`chevron-down` to reveal the file contents and
click {octicon}`copy` to copy the contents to your clipboard. Open `file.txt`
in a text editor and paste the contents there.

::::{dropdown} `dune-project`
:icon: file-code

:::{literalinclude} introduction/dune-project
:language: dune
:::

This file contains metadata about the project:

- the version of the dune language we're using
- the extensions we're {doc}`using </reference/dune-project/using>`
- the {doc}`package </reference/dune-project/package>` we're defining

:::{seealso}
{doc}`/reference/dune-project/index`
  Reference documentation about `dune-project` files
:::
::::

::::{dropdown} `dune`
:icon: file-code

:::{literalinclude} introduction/dune
:language: dune
:::

This file contains a description of what's in our project:

- an {doc}`/reference/dune/executable` stanza defining our calculator binary
- an {doc}`/reference/dune/ocamllex` stanza, setting up rules to compile
  `lexer.mll` to a `Lexer` module
- a {doc}`/reference/dune/menhir` stanza, to similarly use `parser.mly` as a `Parser` module

::::

::::{dropdown} `parser.mly`
:icon: file-code

:::{literalinclude} introduction/parser.mly
:language: ocaml
:::

This contains definitions of our tokens and grammar rules, using
[Menhir](https://gallium.inria.fr/~fpottier/menhir/).
::::

::::{dropdown} `lexer.mll`
:icon: file-code

:::{literalinclude} introduction/lexer.mll
:language: ocaml
:::

This is our lexer, using [ocamllex](https://ocaml.org/manual/5.2/lexyacc.html).
::::

::::{dropdown} `ast.ml`
:icon: file-code

:::{literalinclude} introduction/ast.ml
:language: ocaml
:::

This contains a definition of the arithmetic expressions manipulated by the calculator.

:::{note}
This is in a separate file from `calc.ml` to avoid module cycles, since `Calc`
depends on `Parser`, which depends on the expression type.
:::
::::

:::{dropdown} `calc.ml`
:icon: file-code

:::{literalinclude} introduction/calc.ml
:language: ocaml
:::

This is the "business logic" of our app, in which we:
- display a prompt
- call the lexer and parser to get an expression
- evaluate the expression
- display the result
:::

At this stage, we have the skeleton of a calculator.

Run the following command to build and execute the calculator:

```sh
dune exec calc
```

You can enter additions, such as `1+2` followed by {kbd}`Enter`.
Exit with {kbd}`Ctrl+C`.

Initially, only addition is supported and anything else triggers an
exception terminating the execution.

Note that a `_build` directory is now present. This is where Dune will store
all compiled artifacts.

You can safely remove this directory - that's actually what the `dune clean`
command does. But that's not usually necessary since Dune will keep track of
dependencies and what is up to date.

The `_opam` directory is where your dependencies are located. It is managed by
opam. If it gets removed by accident or something is corrupted in there, it is
safe to remove it and recreate it by running `opam switch` and `opam install`
as described above.

Improving Structure
===================

Our calculator is fairly monolithic at this stage.
Instead of a single executable, we're going to extract a library and create
some tests.
For now, this is just a cram test that will call the executable, but this
structure will later allow adding unit tests for the library.

## Extract a Library

Create folders `bin`, `lib` and `test`, for binary, library, and tests,
respectively.

Run the following command to install [cmdliner](https://ocaml.org/p/cmdliner/1.3.0/doc/index.html):

```sh
opam install cmdliner.1.3.0
```

Let's create a library. Create `lib/dune` with the `(ocamllex)` and `(menhir)` stanzas from the original `dune` file and a new `(library)` stanza:

:::{literalinclude} structure/lib/dune
:language: dune
:emphasize-lines: 1-3
:::

:::{note}
This is the whole contents of the file. The `(library)` part is highlighted to
show that it's the part that we've just added.
:::

:::{note}

We're defining a {doc}`library </reference/dune/library>` that depends on the
`cmdliner` library.

Libraries can either be defined in your project, or provided by an opam
package. In the case of `cmdliner`, this is the latter, since we've installed
it just before.

{doc}`The OCaml Ecosystem </explanation/ocaml-ecosystem>` covers the difference
between packages, libraries, and modules.
:::

Move `ast.ml`, `lexer.mll`, and `parser.mly` to the `lib` directory.

Now we're going to move `calc.ml` to `lib/cli.ml` and replace it by the following:

:::{literalinclude} structure/lib/cli.ml
:language: ocaml
:::

:::{note}
Two things are happening here.

We are adding a second code path to evaluate a `string` directly, so we extract
an `eval_lb` function that operates on a `lexbuf` (the "source" a lexer can
read from).

We are also moving to `cmdliner` for command-line parsing. This consists in:
- an `info` value (of type `Cmdliner.Cmd.info`) which contains metadata for the program (used in help, etc)
- a `term` value (of type `unit Cmdliner.Term.t`) which sets up arguments and calls `eval_lb` with the right `lexbuf`
- a `cmd` value (of type `unit Cmdliner.Cmd.t`) grouping `info` and `term` together 
- a `main` function of type `unit -> 'a` to run `cmd`
:::

## Extract an Executable

Let's create an executable in `bin`. To do so, create a `bin/dune` file with the following contents:

:::{literalinclude} structure/bin/dune
:language: dune
:::

And `bin/calc.ml` with a single function call:

:::{literalinclude} structure/bin/calc.ml
:language: ocaml
:::

Delete the `dune` at the root.

## Create a Test

Create `test/calc.t` with the following contents.

:::{important}
In {doc}`cram tests </reference/cram>`, commands start with two spaces, a
dollar sign, and a space.

Make sure to include **two spaces** at the beginning of the line.
:::

:::{literalinclude} structure/test/calc.t
:language: cram
:lines: 1
:::

Now create `test/dune` to inform Dune that cram tests will use our `calc`
executable and need to be executed again when it changes:

:::{literalinclude} structure/test/dune
:language: dune
:::

At this stage, we're ready to run our test.

Let's do this with `dune runtest`.

It's displaying a diff:

```diff
   $ calc -e '1+2'
+  3
```

Now, run `dune promote`. The contents of `test/calc.t` have changed. Most
editors will pick this up automatically, but it might be necessary to reload
the file to see the change.

Finally, run `dune runtest`. Nothing happens.

Now, run the calculator by running `dune exec calc` to confirm that the
interactive mode still works.

:::{note}
What happened here? This Dune feature, where some tests can edit the source
file, is called {doc}`promotion </concepts/promotion>`.

{doc}`Cram tests </reference/cram>` contain both commands and their expected input.
We did not include any output in the initial cram test. When running `dune
runtest` for the first time, Dune executes the commands, and calls `diff`
between the *expected output* (in `test/calc.t`: no output at all) and the *actual
output* (from running the command: the line "3"), which will display added
lines with a `+` sign and deleted lines with a `-` sign.

Running `dune promote` replaces the input file (`test/calc.t`) with the last
*actual output*. So this includes the line with "3".

Running `dune runtest` again will execute the test again and compare the
*expected output* (`test/calc.t` with the "3" line in it) with the *actual
output* and finds no difference. This means that the test passes.
:::

::::{dropdown} Checkpoint
:icon: location

This is how the project looks like at the end of this chapter.

:::{literalinclude} introduction/dune-project
:caption: dune-project (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/dune
:caption: bin/dune
:language: dune
:::

:::{literalinclude} structure/bin/calc.ml
:caption: bin/calc.ml
:language: ocaml
:::

:::{literalinclude} structure/lib/dune
:caption: lib/dune
:language: dune
:::

:::{literalinclude} introduction/ast.ml
:caption: lib/ast.ml (unchanged)
:language: ocaml
:::

:::{literalinclude} structure/lib/cli.ml
:caption: lib/cli.ml
:language: ocaml
:::

:::{literalinclude} introduction/lexer.mll
:caption: lib/lexer.mll (unchanged)
:language: ocaml
:::

:::{literalinclude} introduction/parser.mly
:caption: lib/parser.mly (unchanged)
:language: ocaml
:::

:::{literalinclude} structure/test/dune
:caption: test/dune
:language: dune
:::

:::{literalinclude} structure/test/calc.t
:caption: test/calc.t
:language: cram
:::

::::

Using a PPX Preprocessor
========================

Our calculator is pretty much opaque: we feed it a string, and it displays a
result (on an error message), but we have now way to know what the internal expression looks like.

In this chapter, we're going to use a `ppx` deriver to generate a `pp_expr`
function that can display expressions.

## Prerequisites

Install [ppx_deriving](https://github.com/ocaml-ppx/ppx_deriving) by running:

```sh
opam install ppx_deriving.5.2.1
```

## Create a test

Add a new test in `test/calc.t`:

```console
$ calc --debug-ast -e '2 * sin (pi / 2)'
```

Run `dune runtest` and see the failure.
Run `dune promote` to add the failure to the test file.
Our goal in the rest of the chapter is to change the output of this test.

## Use `ppx_deriving.show`

Add an `[@@deriving show]` attribute on types in `lib/ast.ml`:

```{code-block} ocaml
:emphasize-lines: 5,13
type op =
  | Add
  | Mul
  | Div
[@@deriving show]

type expr =
  | Int of int
  | Float of float
  | Ident of string
  | Op of op * expr * expr
  | Call of string * expr
[@@deriving show]
```

This will generate `pp_op`, `show_op`, `pp_expr`, and `show_expr` functions.

To do do we need to instruct Dune to use `ppx_deriving.show` as a preprocessor by updating `lib/dune`:

```{code-block} dune
:emphasize-lines: 4-5
(library
 (name calc)
 (libraries cmdliner)
 (preprocess
  (pps ppx_deriving.show))
 (foreign_stubs
  (language c)
  (names calc_stubs)))
```

## Add a `--debug-ast` Flag

We have a few edits to make to `lib/cli.ml`.

First, add a new `cmdliner` flag to parse the command-line and pass it to
`repl` and `eval_lb`:

```{code-block} ocaml
:emphasize-lines: 6-8,11-12
let term =
  let open Cmdliner.Term.Syntax in
  let+ expr_opt =
    let open Cmdliner.Arg in
    value & opt (some string) None & info [ "e" ]
  and+ debug_ast =
    let open Cmdliner.Arg in
    value & flag & info [ "debug-ast" ]
  in
  match expr_opt with
  | Some s -> eval_lb ~debug_ast (Lexing.from_string s)
  | None -> repl ~debug_ast
```

Then, forward it from `repl` to `eval_lb`:

```{code-block} ocaml
:emphasize-lines: 1,5
let repl ~debug_ast =
  while true do
    Printf.printf ">> %!";
    let lb = Lexing.from_channel Stdlib.stdin in
    eval_lb ~debug_ast lb
  done
```

Finally, update `eval_lb` to use it:

```{code-block} ocaml
:emphasize-lines: 1,4
let eval_lb ~debug_ast lb =
  try
    let expr = Parser.main Lexer.token lb in
    if debug_ast then Format.eprintf "[debug] %a\n" Ast.pp_exp expr;
    let v = eval expr in
    Printf.printf "%s\n" (value_to_string v)
  with Parser.Error ->
    Printf.printf "parse error near character %d" lb.lex_curr_pos
```

Run the tests again with `dune runtest`.
At that point, the output should be correct.
Call `dune promote` to update the expected output.

::::{dropdown} Checkpoint
:icon: location

This is how the project looks like at the end of this chapter.

:::{literalinclude} introduction/dune-project
:caption: dune-project (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/dune
:caption: bin/dune (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/calc.ml
:caption: bin/calc.ml (unchanged)
:language: ocaml
:::

:::{literalinclude} using-ppx/lib/dune
:caption: lib/dune
:language: dune
:::

:::{literalinclude} using-ppx/lib/ast.ml
:caption: lib/ast.ml
:language: ocaml
:::

:::{literalinclude} interfacing-with-c/lib/calc_stubs.c
:caption: lib/calc_stubs.c (unchanged)
:language: c
:::

:::{literalinclude} using-ppx/lib/cli.ml
:caption: lib/cli.ml
:language: ocaml
:::

:::{literalinclude} interfacing-with-c/lib/lexer.mll
:caption: lib/lexer.mll
:language: ocaml
:::

:::{literalinclude} development-cycle/lib/parser.mly
:caption: lib/parser.mly (unchanged)
:language: ocaml
:::

:::{literalinclude} structure/test/dune
:caption: test/dune (unchanged)
:language: dune
:::

:::{literalinclude} using-ppx/test/calc.t
:caption: test/calc.t
:language: cram
:::

::::

---
author: Marek Kubica
---

OCaml Package Management With Dune
==================================

:::{warning}
Dune Package Management is not final yet and details are still subject to
change.
:::

In this tutorial we will be looking at how to use Dune for managing project
dependencies. This enables users to install the compiler as well as third-party
dependencies using a single tool which takes care of building code and
dependencies.

To get started you only need Dune. Head to {doc}`setup` to begin the setup.

:::{toctree}
:hidden:
:maxdepth: 1
setup
dependencies
pinning
repos
:::

---
author: Etienne Millon
---

Developing with Dune
====================

:::{warning}
This tutorial is a work in progress.
:::

In this tutorial, will start with a small Dune project, and extend it by using
common features:

- the {doc}`/reference/dune/executable`, {doc}`/reference/dune/library`, and
  {doc}`/reference/dune/test` {term}`stanzas <stanza>`;
- {doc}`cram </reference/cram>` tests;
- bindings to C code using {doc}`foreign stubs </reference/foreign-stubs>`;
- using a [ppx deriver](https://ocaml.org/docs/metaprogramming).

By doing so, you'll interact with `dune runtest` and `dune promote`, and will
use the most common {term}`stanzas <stanza>` in Dune files.

Start the tutorial with the {doc}`introduction`.

:::{toctree}
:hidden:
:maxdepth: 1
introduction
structure
development-cycle
interfacing-with-c
using-ppx
unit-tests
conclusion
:::

Dune is an community orientated open source project. It was originally
developed at [Jane Street][js] and is now maintained by Jane Street,
[Tarides][tarides] as well as several developers from the OCaml
community.

Contributions to Dune are welcome and should be submitted via GitHub
pull requests against the `main` branch. See [./doc/hacking.rst][hack]
for a guide to getting started on the code base.

Dune is distributed under the MIT license and contributors are
required to sign their work in order to certify that they have the
right to submit it under this license. See the following section for
more details.

Signing contributions
---------------------

We require that you sign your contributions. Your signature certifies
that you wrote the patch or otherwise have the right to pass it on as
an open-source patch. The rules are pretty simple: if you can certify
the below (from [developercertificate.org][dco]):

```
Developer Certificate of Origin
Version 1.1

Copyright (C) 2004, 2006 The Linux Foundation and its contributors.
1 Letterman Drive
Suite D4700
San Francisco, CA, 94129

Everyone is permitted to copy and distribute verbatim copies of this
license document, but changing it is not allowed.


Developer's Certificate of Origin 1.1

By making a contribution to this project, I certify that:

(a) The contribution was created in whole or in part by me and I
    have the right to submit it under the open source license
    indicated in the file; or

(b) The contribution is based upon previous work that, to the best
    of my knowledge, is covered under an appropriate open source
    license and I have the right under that license to submit that
    work with modifications, whether created in whole or in part
    by me, under the same open source license (unless I am
    permitted to submit under a different license), as indicated
    in the file; or

(c) The contribution was provided directly to me by some other
    person who certified (a), (b) or (c) and I have not modified
    it.

(d) I understand and agree that this project and the contribution
    are public and that a record of the contribution (including all
    personal information I submit with it, including my sign-off) is
    maintained indefinitely and may be redistributed consistent with
    this project or the open source license(s) involved.
```

Then you just add a line to every git commit message:

```
Signed-off-by: Joe Smith <joe.smith@email.com>
```

Use your real name (sorry, no pseudonyms or anonymous contributions.)

If you set your `user.name` and `user.email` git configs, you can sign
your commit automatically with `git commit -s`.

It is possible to set up `git` so that it signs off automatically by using a
prepare-commit-msg hook in git. See <https://stackoverflow.com/a/46536244> for
details. As noted in the manual for `format.signOff`, note that adding the
`Signed-off-by` trailer should be a conscious act and means that you certify
you have the rights to submit this work under the same open source license.

[dco]: http://developercertificate.org/
[js]: https://www.janestreet.com/
[tarides]: https://tarides.com/
[hack]: ./doc/hacking.rst

Coding style
------------

- wrap lines at 80 characters,
- use `[Ss]nake_case` over `[Pp]ascalCase`.

Conclusion
==========

This tutorial has covered the parts of Dune that users are the most likely to
interact with:

- the {doc}`/reference/dune/executable`, {doc}`/reference/dune/library`, and
  {doc}`/reference/dune/test` stanzas;
- how to use {doc}`cram tests</reference/cram>` and the workflow of
  {doc}`promotion </concepts/promotion>`;
- bindings to C code using {doc}`foreign stubs </reference/foreign-stubs>`;
- using a ppx deriver.

## Where to go from here

You might be interested in:

- our {doc}`how-to guides </howto/index>` to apply this to your projects;
- various {doc}`explanations </explanation/index>` about how Dune works.

# How Package Management Works

This document explains how Dune's package management works under the hood. It
requires a bit of familiarity with how opam repositories work and how Dune
builds packages. Thus it is aimed at people who want to understand how the
feature works, not how it is used.

For a tour on how to apply package management to a project, refer to the
{doc}`/tutorials/dune-package-management/index` tutorial.

## Motivation

A core part of modern programming is using existing code to save time. The
OCaml package ecosystem has quite a long history with many projects building
upon each other over many years. A significant step forward was the creation of
the OCaml Package Manager, opam, along with the establishment of a public
package repository which made it a lot more feasible to share code between
people and projects.

Over time, best practices have evolved, and while opam has incorporated some
changes, it couldn't adopt all the modern workflows due to its existing user
base and constraints.

Thus the Dune Package Management has been designed with a few core goals in
mind:

* No global state visible to users, everything is local to projects
* Package management is configured through files (`dune-project` and optionally
  `dune-workspace`)
* Repositories are automatically kept up to date unless explicitly configured
  to use specific versions
* Builds can only access packages they have declared dependencies on
* Reproducible builds through lockfiles

Dune plays well with the existing OCaml ecosystem and does not introduce a new
type of packages. Rather, it uses the same package repository and Dune packages
stay installable with opam.

## Package Management in a Project

This section describes what happens in a Dune project using the package
management feature.

## Dependency Selection

The first step is to determine which packages need to be installed.
Traditionally this has been defined in the `depends` field of a project's opam
file(s).

Since version 1.10 Dune has supported {doc}`opam file generation
</howto/opam-file-generation>` by specifying the package dependencies in the
`dune-project`.

The package management feature uses the same metadata, as Dune will determine
the list of packages to install from the `depends` field in the `dune-project`
file. This allows projects to completely omit generation of `.opam` files, as
long as they use Dune for package management. Thus all dependencies on OCaml
packages are only declared in one single file.

For compatibility with a larger amount of existing projects, Dune will also
collect dependencies from `.opam` files in the project. So while recommended,
there is no obligation to switch to declaring dependencies in the
`dune-project`. Likewise the generation of `.opam` files continues to be
supported.

## Locking

Given the list of the project's dependencies and their version
constraints, the next steps are:

1. Find the transitive dependencies and figure out a version for each
   dependency that satisifies the constraints
2. For each dependency, download it, build it, and make it available to the
   project

In opam, `opam install` does both of these.

In Dune, these are separate steps: the first one is `dune pkg lock`, and the
second one happens implicitly as part of [building](#building).

The idea of doing the first step and recording it for later is popular in other
programming language package managers like NPM and is usually called locking.
Creating a lock file ensures that the dependencies to be installed are
always the same - unless that lock file is updated of course.

:::{note}
`opam` also supports creating lock files. However, these are not as central to
the opam workflow as they are in the case of package management in Dune, which
always requires a set of locked packages.
:::

In the most general sense, a lock file is just a set of specific packages
and their versions to be installed.

Instead of a lock file, Dune writes this information to a directory (the "lock
directory") with files that describe the dependencies. It includes the
package's name and version. Unlike many other package managers, the files
include a lot of other information as well, such as the location of the source
archives to download (since there is no central location for all archives), the
build instructions (since each package can use its own way of building), and
additional metadata like the system packages it depends upon.

The information is stored in a directory (`dune.lock` by default) as separate
files, to reduce potential merge conflicts and simplify code review. Storing
additional files like patches is also simpler this way.

### Package Repository Management

To find a valid solution that allows a project to be built, it is necessary to
know what packages exist, what versions of these packages exist, and what other
packages these depend on, etc.

In opam, this information is tracked in a central repository called
[`opam-repository`](https://github.com/ocaml/opam-repository), which contains
all the metadata for published packages.

It is managed using Git; opam typically uses a snapshot to find the
dependencies when searching for a solution that satisfies the constraints.

Likewise, Dune uses the same repository; however, instead of snapshots of the
contents, it uses the Git repository directly.

:::{note}
Dune maintains a shared internal cache containing all Git repositories that
projects use. This way updates and checkouts are very fast because only new
revisions have to be retrieved. The downside is that to be included in the
cache, all the Git repos have to be cloned first which depending on the size of
the repositories can take a bit of time.
:::

On every call to `dune pkg lock`, Dune will update the metadata repository
first (hence why efficiently updating that repository matters). This means that
each `dune pkg lock` will use the newest set of packages available.

However, it is also possible to declare specific revisions of the repositories,
to get a reproducible solution. Due to using Git, any previous revision of the
repository can be used by specifying a commit hash.

Dune uses two repositories by default:

* `upstream` refers to the default branch of `opam-repository`, which contains
  all the publicly released packages.
* `overlay` refers to
  [opam-overlay](https://github.com/ocaml-dune/opam-overlays), which defines
  packages patched to work with package management. The long-term goal is to
  have as few packages as possible in this repository as more and more packages
  work within Dune Package Management upstream. Check the
  [compatibility](#compatibility) section for details.

### Solving

After Dune has read the constraints and loaded set of candidate packages, it is
necessary to determine which packages and versions should be selected for the
package lock.

To do so, Dune uses
[`opam-0install-solver`](https://github.com/ocaml-opam/opam-0install-solver),
which is a variant of the [`0install`](https://github.com/0install/0install)
solver to find solutions for opam packages.

Contrary to opam, the Dune solver always starts from a blank slate; it assumes
nothing is installed and everything needs to be installed. This has the
advantage that solving is now simpler, and previous solver solutions don't
interfere with the current one. Thus, given the same inputs, it should always
come up with the same result; no state is held between the solver runs.

This can lead to more packages being installed (as opam won't install new
package versions by default if the existing versions satisfy the constraints),
but it avoids interference from already installed packages that lead to
potentially different solutions.

After solving is done, the solution gets written into the lock directory with
all the metadata necessary to build and install the packages. From this point
on, there is no need to access the package metadata repositories.

:::{note}
Solving and locking does not download the package sources. These are downloaded
in the build step.
:::

(building)=
## Building

When building, Dune will read the information from the lock directory and set
up rules for the packages. Check {doc}`/explanation/mental-model` for details
about rules.

The rules that the package management sets up include:

* Fetch rules to download and unpack the source archives, and also download any
additional sources such as patches
* Build rules to execute the build instructions stored in the lock directory
* Install rules to put the artifacts that were built into the appropriate
  Dune-managed folders

Creating these processes as rules mean that they will only be executed on
demand, so if the project has already downloaded the sources, it does not need
to download them again. Likewise, if packages are installed, they stay
installed.

The results of the rules are stored in the project's `_build` directory and
managed automatically by Dune. Thus, when cleaning the build directory, the
installed packages are cleaned as well and will be reinstalled at the next
build.

(compatibility)=
## Packaging for Dune Compatibility

Dune can build and install most packages as dependencies, even if they are not
built with Dune themselves. Dune will execute the build instructions from the
lock directory, very similar to opam.

However, packages must adhere to certain rules to be compatible with Dune.

The most important one is that the packages must not use absolute paths to
refer to files. That means they cannot read the path they are being built or
installed in and expect this path to remain the same. Dune builds packages in a
sandbox location, and after the build has finished, it moves the files to the
actual destination.

:::{note}
Unlike opam, Dune at the moment does not wrap the build in sandboxing tools
like [Bubblewrap](https://github.com/containers/bubblewrap).
:::

To comply with these restrictions the usual solution is to use relative paths,
as Dune guarantees that packages installed into different sections are
installed in a way where their relative location stays the same.

The `overlay` repository exists specifically to make currently non-compliant
packages compatible with Dune's package management. It does so by supplying
releases of packages where the current upstream releases don't support Dune
package management yet.

# Runtime RPC versioning implementation notes

This document describes the versioning protocol used to ensure two-way
compatibility between different versions of the API.

The approach is loosely inspired by the `Both_converts` model used by
[Versioned_rpc](https://ocaml.janestreet.com/ocaml-core/latest/doc/async_rpc_kernel/Async_rpc_kernel/index.html#module-Versioned_rpc)
in `Async`, in which both parties maintain a "menu" of supported RPC
versions, which is used to negotiate a common protocol for each
method.

This is a working document and will be updated as the design evolves.

## Terms

- A **procedure** is a common term encompassing *notifications*
    (one-way messages) and *requests* (a communication to which
    a response is expected).

- A **model** is the logical payload type for one direction of
    a procedure. Note that this is a per-actor entity; the ultimate
    goal of runtime RPC versioning is to allow clients and servers to
    disagree on a model type without preventing them from interacting.

- A **wire type** for a procedure is the logical type sent "over the
    wire" for one direction of a procedure. The end result of
    negotiating a version for a procedure is to select a wire type
    known to both the client and the server. Typically, the older of
    the two model types will be chosen as the wire type,

- A **generation** of a procedure is the set of wire types
    corresponding to each direction of a procedure, along with the
    de/serialisation logic and upgrade and downgrade functions
    transforming the wire types to the model types and vice versa.
    Each generation is associated with a *version number*, which
    should be unique within a procedure.

- The **menu** is a mapping from method names to the particular
    generations that will be used for each procedure for a particular
    session. In the source code, this term is overloaded to also refer
    to a mapping from method names to *all known* generations of
    a procedure.

- The **declaration** of a procedure lists its model types and all
    known generations, along with its method name. Multiple
    declarations of the same procedure is allowed, so long as they do
    not overlap version numbers.

- The **implementation** of a declaration is the actual behavior of
    a procedure, which acts on the model types. Typically, this will
    be on the server, but in the future there may also be a use for
    server-to-client requests. This document is not concerned with the
    internals of any given implementation, only whether such an
    implementation exists at all.

## Background

Previously, there was no distinction between model and wire types.
This meant that any change to a model type required both build servers
and clients to upgrade in lockstep, as otherwise the receiver would be
unable to deserialize the payload of a procedure.

Unfortunately, most lighter-weight solutions (such as modifying the
de/serialization logic to be resilient to, e.g., extra/missing fields
or variants in types) are insufficient. Early designs of the
diagnostic API, for example, reported targets as strings, but was
changed to give structured information instead.

Similarly, requirements like "the client must always be older than the
server" (or the reverse) don't work in environments like Jane Street,
where the same editor plugin must be able to interact seamlessly with
multiple iterations of Dune (which may be older or newer than the
editor plugin itself).

The main goal of the system, then, is to ensure that both the server
and client applications can be programmed against the current model
types for each procedure, with all backwards- or forwards- conversions
happening under the hood.

## Protocol

At session initialization time, the client will first send an
initialization request to the server containing a single version
number corresponding to the overall RPC version the client will use.
If this number is determined to be versioning-compatible (see
[Session versioning](#session-versioning)), the server will respond
with a token instructing the client to initiate version negotiation.
Otherwise, the server will respond with an error.

Upon receiving this token, the client will initiate version
negotiation by sending a list of `(method-name, generations)`
pairs, where `method-name` is the name of each declared procedure, and
`generations` is the list of version numbers for that procedure's
generations.

Upon receiving a list of supported versions from the client, the
server will compare it to its list of *implemented* versions,
selecting the greatest common generation for each procedure. If the
client and server do not share any generations for a procedure, it is
omitted entirely. If there is at least one method for which a common
version exists, then the server responds with a list of `(method-name,
selected-version)` pairs, where `selected-version` is the version
number of the greatest common generation. This list is then used by
both parties to construct the version menu. Otherwise, if there are no
common versions for any methods, an error is returned to the client
and the session is invalidated.

Note that we do not currently require declared/implemented versions to
span a contiguous range of version numbers. This can have a few uses,
such as preventing clients from using a known-bugged generation of
a procedure.

When executing a procedure, the sender first looks up the correct
generation in the menu (see [Error handling](#error-handling)), and
downgrades the payload from the sender-side model type to the wire
type. Upon receipt, the server performs the same lookup to deserialize,
then upgrade the payload to the receiver-side model type, then the
procedure implementation is performed, producing a response in the
case of requests. If necessary, the same transformations are then
performed in reverse, sending the value back to the sender, completing
the procedure.

Barring strange circumstances (such as a client declaring a generation
with a newer version number than the type exposed in `dune_rpc.mli`), it
is always the case that the transformation from wire to model types will
be the identity function on the side that is older.

## Miscellaneous implementation notes

### Session versioning

In addition to version numbers existing for each procedure version,
there are two further version numbers associated with the session as
a whole which are sent as part of session initialization.

The first is the version of Dune each side purports to be as
a `MAJOR.MINOR` number (serialized as an `int * int` pair). This is
not currently checked.

Next is a version of the initial handshake protocol to be used. This
takes the form of a single `int`. In the future, if the initial
negotiation protocol changes, this value can be adjusted and checked to
account for this.

### Error handling

Handling of versioning errors has become more complex, as we need to
distinguish between "no such method exists" and "the server and client
do not share any common generations for this method". Secondly, this
means that the initiation of a procedure can now fail, which
complicates one-way communications (for example, the server must
swallow errors and clients must be upgraded to handle version errors
on notifications, which were previously infallible).

Finally, the versioning protocol itself must be either versioned
separately or stabilised (see [Session versioning](#session-versioning)).

### Tweaks

- We currently send the entire version menu from client to server and
    back twice, once for the client to inform the server of all
    supported versions, and again for the server to inform the client
    of the common versions. This can lead to large messages being
    passed at session initialization, which may become a performance
    bottleneck.

    - The size of the version negotiation messages is proportional to
        the number of all known generations for all procedures, which
        can be approximated by `number-of-procedures` times
        `number-of-supported-generations`. In practice, I do not
        expect this number to be large (I would be surprised if this
        number is ever on the order of 100).

    - One alternative is to perform per-procedure negotiation, where
        the initiator of a procedure first sends its known version
        ranges, the recipient sends the selected version (or an
        error), then the procedure proceeds as before. This approach
        trades startup and lookup overhead for a constant
        per-communication overhead. It also makes distinguishing "no
        such method exists" and "no common versions" simpler.

# Setting Up Package Management With Dune

The idea of package management with Dune has been as unobtrusive as
possible. Thus most projects can easily be built with just the minimum of
changes.

In this tutorial we will create a simple project to use the integrated package
management feature for the very first time.

## Declare Dependencies

The best way to work with the package management is to declare your
dependencies in the `dune-project` file.

::::{dropdown} `dune-project`
:icon: file-code

:::{literalinclude} setup/dune-project
:language: dune
:emphasize-lines: 6-7
:::

We define a project called `test` and declare that to build it we need an OCaml
compiler that is at least version 4.14.

This is exactly the same information that is used to generate opam files using
the `generate_opam_files` stanza as described in
{doc}`/howto/opam-file-generation`.

::::

::::{dropdown} `test.ml`
:icon: file-code

:::{literalinclude} setup/test.ml
:language: ocaml
:::

To show that the build works, this simple program will be built and executed.

::::


::::{dropdown} `dune`
:icon: file-code

:::{literalinclude} setup/dune
:language: dune
:::

To declare our module an executable we need a little bit of configuration, so we
just define the module as an executable.

::::

After our project skeleton is set up, we can proceed to the next step.

## Locking Dependencies

After declaring the dependencies, you will need to tell Dune which package
versions to use for your project. This is done by creating a lock directory.
This is easily done with a new Dune command:

```
$ dune pkg lock
Solution for dune.lock:
- ocaml.5.2.0
- ocaml-base-compiler.5.2.0
- ocaml-config.3
```

This will update all the required opam repositories, use the newest version of
each and try to find a set of packages and versions that satisfy the
constraints that your project dependencies declare.

:::{note}
The versions that get locked might be different from this tutorial, as we only
specified the lower bound of `ocaml`; barring any additional configuration, Dune
will pick the newest possible version for each dependency.
:::

## Build Project

To build the project, you can just use the regular Dune commands.

```sh
dune build
```

This will download, build, and install all your locked dependencies and then use
those to build your project. This means that the first time building it will take
longer than usual, as the dependencies need to be built first. Subsequent
builds where all dependencies have been built before will be just as fast as
before.

We can show that the package has been built successfully and works as expected:

```
$ dune exec ./test.exe
Hello, OCaml, Rust!
```

:::{note}
If you want to only build and fetch the project dependencies, you can use
the `@pkg-install` alias like so

```shell
$ dune build @pkg-install
```

See {doc}`/reference/aliases/pkg-install` for more information.
:::

## Conclusion

In this section we learned how to set up a Dune project that picks a compiler
and installs it without the need for any additional tooling.

In the next section {doc}`dependencies` we will look on how to add third party
dependencies.

# Rule streaming

This document describes a new design for the production of build rules
in Dune. The new design aims to be more natural, easier to reason
about and to make existing features work well with newer ones such as
directory targets.

It was originally written by Jérémie Dimino as part of the
[streaming RFC](https://github.com/ocaml/dune/pull/5251), and later on moved
into the dev documentation.


## Problem

The [rule production](./rule-production.md) document exposes a concrete problem
with directory targets, but there is also a general sense of messiness in the
way things work. Generating rules for multiple directories at once is
natural, but the current encoding is odd.

## Proposal

The proposal is to add the following rule: `gen_rules ~dir` is allowed
to produced rules in `dir` or any of its descendant only. It is not
allowed to produce rules anywhere else.

`Load_rules.load_dir ~dir` will then always call itself recursively on
the parent of `dir` and take the union of the rules produced by
`gen_rules` for `dir` and the ones produced by the recursive
call. `gen_rules` will no longer have to redirect a call via
`Load_rules.load_dir_and_produce_its_rules`, which we would simply
remove.

This introduces a cycle with all `copy_files` stanza that copy files
from a sub-directory. We propose the break this cycle by introducing
laziness in the rule production code.

### Generating rules with a mask

The idea is that when we produce rules, we will produce rules under
a current active "mask" that tells us where we are allowed to generate
files or directories.  Trying to produce a rule with targets not
matched by this mask will be a runtime error.

When entering `gen_rules ~dir`, the initial mask will be: "any files
and directories that is a descendant of directory `dir`".

We can then narrow the mask to a sub-mask:

```ocaml
val narrow : Target_mask.t -> unit Memo.t -> unit Memo.t
```

With `narrow mask m`, `m` would only be allowed to produce rules whose
target are matched by the intersection of `mask` and the current
mask. `m` wouldn't be evaluated eagerly. Instead, `gen_rules` would
now return a set of direct rules as well as a list of
`(Target_mask.t * unit Memo.t)`. Let's call such a pair a
suspension. A suspension can be forced by evaluation its second
component. Doing so will yield a list of rules matched by the mask and
a new list of suspension.

### Staged rules loading

The next step is to stage `Load_rules.load_dir`. In addition to taking
a directory, `load_dir` will now also take a mask and will return the
set of rules for this mask. To do that, it might need to force a bunch of
suspensions recursively.


### How does that help?

We will put `copy_rules` under a `narrow <only file targets in current
dir>`. In order to determine if a directory is part of a directory
target in an ancestor directory, we wouldn't need to force this
suspension.

### Difficulties

Interpreting a `library` stanza requires knowing the set of `.ml`
files in the current directory. Knowing this requires interpreting
`copy_files` in the current directory. So the interpretation of
`library` stanzas will need to go under a `narrow` as well.

# Dune cache: design and implementation notes

This document describes main ideas behind the Dune cache as well as a few
subtleties of the implementation. This is a working document and it will be
updated as part of on-going development. Some of the described features are
currently still in development or unused, in particular, here we describe
support for two types of cache entries &ndash; artifacts and values &ndash; but
Dune currently doesn't store any value entries in the cache.

The design and implementation are based on a non-trivial assumption that there
are no hash collisions, for instance, that we will never come across two files
with different contents but with the same content hash. While this assumption is
common in the world of build systems and package managers, and is highly
unlikely to be violated by chance, one can manufacture hash collisions on
purpose, especially when using a weak hash like MD5.

## What we store in the cache

The cache stores build _artifacts_ and _values_.

* An _artifact_ is a file produced by a build rule. As any file, it has a name
  as well as content. Note that we treat a file's executable permission bit as
  part of its content.

* A _value_ is anything else produced during a build that is not written to a
  file but which is worth storing persistently between successive builds. A
  common example is a string written to the standard output by a build action.
  Such output-producing actions may run as part of a build rule or at an earlier
  stage when generating rules. Unlike artifacts, values have no names, yet they
  still have content.

## How we use the cache

The build system uses the cache to _store_ and _restore_ build artifacts and
values. Here is a typical interaction sequence for the case of artifacts:

* The build system is executing a build rule and has already identified all of
  its dependencies, thereby obtaining the _rule hash_. It uses it to make a
  _restore request_ to the cache.

* Now there are two cases:

    - The cache successfully restores all the artifacts of the rule, placing
      them into the build directory and returning their content hashes. The
      build system can skip building the artifacts and can use the obtained
      content hashes for deciding whether any dependent build rules need to be
      rerun. **This successful scenario is the only reason we use the cache.**

    - The cache fails to restore the artifacts, either because of an error or
      because it doesn't have an entry for the given rule hash. In this case,
      the build system needs to build the artifacts itself. On completion, it
      makes a _store request_ to the cache, providing a list of build artifacts
      (_file names_ in the build directory) as well as their _content hashes_.
      The cache stores the artifacts and after that the build system is allowed
      to continue with the build (but not before, since the cache requires
      exclusive access to the artifacts in the build directory). There can be a
      rare situation where the store request is declined because the given entry
      is already in the cache. How could this happen? We did try to restore it
      first! The reason is that the cache can be populated concurrently by
      multiple build systems and by the distributed cache daemon, so there can
      be a race between multiple systems adding the same entry to the cache. In
      this case, the cache will perform _deduplication_ of build artifacts by
      replacing them with hard links to the copies already stored in the cache
      (this is an example where the exclusive access is needed).

Build values are handled similarly; the main difference is that to identify a
cache entry we use the corresponding _action hash_ rather than the _rule hash_.
The only difference between rule hashes and action hashes is that the former
include the names of the produced artifacts into the hash, while the latter
do not, since values have no names.

## Cache storage format

Let `root` stand for the cache root directory. It has three main subdirectories.

* `root/meta/v3` stores _metadata files_, one per each historically executed
  build rule or value-producing action. (While this is a convenient mental
  model, in reality we need to occasionally remove some outdated metadata files
  to free disk space &ndash; see the section on cache trimming.)
  <br/><br/>
  A metadata file corresponding to a build rule is named by the rule hash and
  stores file names and content hashes of all artifacts produced by the rule.
  <br/><br/>
  A metadata file corresponding to a value-producing action is named by the
  action hash and stores the hash of the resulting value.
  <br/><br/>
  It is important to guarantee that rule and action hashes do not accidentally
  overlap, which may happen if one simply hashes their in-memory representations
  because a rule and an action might happen to be represented by the same
  sequence of bytes in memory.

* `root/files/v3` is a storage for artifacts, where files named by content
  hashes store the matching contents. We will create hard links to these files
  from build directories and rely on the hard link count, as well as on the last
  change time as useful metrics during cache trimming.

* `root/values/v3` is a storage for values. As in the case of `files`, we store
  the values in the files named by their content hashes. However, these files
  will always have the hard link count equal to 1, because they do not appear
  anywhere in build directories. By storing them in a separate directory, we
  simplify the job of the cache trimmer.

* `root/temp` contains temporary files used for atomic file operations needed
   when adding new entries to the cache, as will be described below.

Note that since this document was first written, some of the above paths have
changed due to version bumps (to `v4` and beyond).

## Adding entries to the cache

To add entries to the cache, we use the functions `store_artifacts` and
`store_value` described in the corresponding sections below. Setting possible
errors aside, these functions can succeed in two ways.

* They return `Stored` if the given entry is new and it has been successfully
  stored in the cache.

* They return `Already_present` if the given entry has already been present in
  the cache and can therefore be discarded. This is a rare scenario where
  multiple systems race to add the same entry, and only one of them will receive
  the glory of the `Stored` response.

### Atomic writing to the cache

As mentioned above, the cache can be modified concurrently by multiple systems,
so to prevent collisions on individual files, we need to create new files
atomically. To do that, we first create a temporary file in the `temp`
directory, then create a hard link to it from the cache (this operation will
fail if another process managed to create the cache entry earlier), and then
unlink the temporary file.

From now on, whenever we say "create a file", we mean create a file atomically.
If two systems attempt to create a file with the same name simultaneously, one
of them will win the competition and the contents it writes will remain in the
cache until it is deleted during cache trimming.

Note that it is possible for a metadata file with a given name to have multiple
possible contents due to _non-determinism_, and the cache implementation should
not assume otherwise.

### Storing artifacts

To store artifacts produced by a build rule, we perform the following sequence
of steps.

* Create a metadata file in the `meta` directory, listing all the artifacts.
  If the file already exists (which should be a rare case), verify that it
  contains the expected list of artifacts (both file names and content hashes).
  If it doesn't, we have found a non-deterministic build rule and report an
  error.

* For each artifact, we store it in the `files` directory using the artifact's
  content hash as the name. In each case, there are two scenarios:

    - If the artifact is already in the cache, we perform
      deduplication by replacing the artifact in the build directory
      with a hard link to the file stored in the cache. We assume that
      the build system will wait for `store_artifacts` to complete before
      starting any further actions that might read these artifacts from
      the build directory and thus interfere with the deduplication.

    - Otherwise, we create a hard link to the artifact from the `files`
      directory.

The function returns `Already_present` if the metadata file and all of the
artifacts were already in the cache; otherwise, it returns `Stored`.

### Storing values

Storing a value is simpler than storing artifacts because there is no need
for deduplication. The steps are:

* Create a metadata file in the `meta` directory, recording the value's hash.
  If the file already exists (which should be a rare case), verify that it
  contains the same hash. If it doesn't, we have found a non-deterministic build
  action and report an error.

* Store the value as a file in the `values` directory using the value's hash as
  the file name. If the file is already in the cache, we don't need to do
  anything.

The function returns `Already_present` if the metadata file and the value were
already in the cache; otherwise, it returns `Stored`.

## Restoring entries from the cache

To restore entries from the cache, we use the functions `restore_artifacts` and
`restore_value` described in the corresponding sections below. Setting possible
errors aside, these functions either fail to find the entry in the cache and
return `Not_found_in_cache`, or succeed and return `Restored` along with some
information about the restored entry.

### Restoring artifacts

Given a rule hash, the function `restore_artifacts` performs the following
steps.

* Look up the corresponding metadata file in the `meta` directory. If it doesn't
  exist, return `Not_found_in_cache`. Otherwise, read the list of artifacts,
  i.e. the list of file names and their content hashes from the metadata file.

* For each artifact, lookup the content hash in the `files` directory. If it
  doesn't exist, return `Not_found_in_cache`. Otherwise: (i) delete the
  corresponding (most likely stale) file in the build directory, and then (ii)
  create a hard link with the same name, pointing to the file in the cache.

If the above succeeds for every artifact in the list, the function returns
`Restored` along with the obtained list of file name and content hash pairs.

### Restoring values

Given an action hash, the function `restore_value` performs the following steps.

* Look up the corresponding metadata file in the `meta` directory. If it doesn't
  exist, return `Not_found_in_cache`. Otherwise, read the hash of the value.

* Look up the hash in the `values` directory. If it doesn't exist, return
  `Not_found_in_cache`. Otherwise, return `Restored` along with the value read
  from the stored file.

## Trimming the cache

Storing all historically produced artifacts and values is infeasible, so the
cache needs to be regularly trimmed. The current trimming algorithm performs the
following steps.

* Scan the `files` directory to find all currently unused artifact entries. An
  artifact is _unused_ if its hard link count is equal to 1. There is no point in
  trimming other entries, since they appear in at least one build directory. In
  fact, trimming them is potentially harmful because if the same entries were to
  be added to the cache again from a new directory, we would have been unable to
  perform the deduplication, thus losing some sharing opportunities.

* Scan the `values` directory to find all value entries. We have no information
  about their current usage, so we conservatively allow all of them to be
  trimmed and recomputed in the next build if needed.

* Sort the entries according to the following criteria:

    - Type: artifacts precede values in the trimming list since artifacts are
      generally larger and we know for sure that they are unused;

    - The time of last change: entries that became unused more recently go later
      in the list.

* Traverse the list and delete the corresponding entries until the trimming goal
  has been met. Right before deleting an artifact entry, double check that its
  hard link count is still equal to 1. A build system running concurrently might
  have created a hard link to it after we collected the information, so deleting
  this file from the cache could lead to a loss of sharing between different
  build directories.

* Finally, traverse the `meta` directory and remove all _broken_ metadata files,
  i.e. the files that refer to content hashes with no corresponding entries in
  the `files` and `values` directories. This step does not need to be done on
  every trimming. It is expensive but metadata files are generally small and
  there is no harm in keeping broken metadata files in the cache. In fact, the
  information contained in broken metadata files can be utilised by the build
  system for so-called _shallow builds_ where intermediate build artifacts are
  not materialised on the disk and it is sufficient to only know their hashes,
  which are listed in the metadata files.

To enable more sophisticated trimming strategies, we could augment the metadata
stored in the cache with information about the _cost_ of producing cache
entries, i.e. the time it would take to execute the corresponding rule or action
to restore the entry if needed. For deterministic build rules, we can do _local
cost reasoning_, i.e. we do not need to take the cost of rebuilding their
dependents into account, since such rebuilding would be unnecessary due to the
early cut-off optimisation.

Another promising idea is to add support for incremental cache trimming where
the build system informs the cache that a previously added entry has become
obsolete, letting the cache trim it early if it meets the trimming criteria.

### Interaction with the previous cache versions

Note also that as the new cache format evolves further and we, for example, move
from `files/v3` to `files/v4`, the cache trimmer will need to evolve too, to be
able to cope with entries of all currently supported versions.

# Directory targets

A **directory target** corresponds to a file tree rooted at a specified root
directory, for example, `docs/*`. Typical examples of rules producing directory
targets are: unpacking an archive, running `make` in a vendored package, and
building files with non-deterministic names (e.g. including the current date).

## Declaring a directory target

To declare a directory target `docs/*`, use the syntax `(target (dir docs))` in
a rule stanza. The corresponding rule should create the directory `docs` and is
free to populate it with an arbitrary number of files and/or subdirectories.

Like file targets, directory targets can be promoted to the source tree by
adding `(mode promote)` to the rule stanza.

## Depending on a directory target

There are two ways to depend on a directory target:

* An **opaque dependency** on the whole file tree `docs/*`. Opaque dependencies
  are invalidated if the contents of the tree is changed in any way. To declare
  an opaque dependency on `docs/*`, use the syntax `(dep (dir docs))` in a rule
  stanza.

* A **projection dependency** on a specific file in the tree, e.g.
  `docs/html/index.html`. A projection dependency is declared using the standard
  syntax `(dep docs/html/index.html)` and works like a dependency on a normal
  file target. For example, if the `docs/*` directory is rebuilt and only
  `docs/html/logo.png` is modified, then the dependency on `docs/html/index.html`
  is considered to be up-to-date. Note that it is easy to make a mistake with
  such projection dependencies, for example, by forgetting that `index.html`
  actually does include the image `docs/html/logo.png`. In such cases,
  sandboxing will help since only the requested projection dependencies will be
  available in the sandbox (i.e., not the whole directory target).

## Building a directory target

Users can request building whole directory targets or individual files via
`dune build docs` and `dune build docs/html/index.html` commands.

## Current limitations

* It is not allowed to have two rules with the same directory target. That is,
  like file targets, directory targets are **exclusive** (but see _shared
  directory targets_ below).

* Directory targets cannot have nested file or directory targets, i.e. other
  rules are not allowed to declare targets within the file tree of a directory
  target.

## Possible future extensions

Here are some possible extensions to consider:

* **Opaque directory targets**: a rule may declare that its directory target is
  opaque, in which case projection dependencies on its content will be
  disallowed. One can also consider only partially opaque directory targets,
  where the contents of the directory is only partially visible.

* **Shared directory targets**: we can allow multiple rules to write to the same
  directory target, as long as they do not write to the same files. In this
  case, depending on a directory target would mean depending on all of the rules
  that declare it as a target.

# Pinning Projects

When Dune is looking up packages to lock, it uses the (pre)configured OCaml
package repositories. However it is also possible to manually specify the
sources of packages; for example, if the package is not released in a package
repository. This is called "pinning".

## Installing Packages From a Pin

To pin a package, a new `pin` has to be declared in the `dune-project` file.

::::{dropdown} `dune-project`
:icon: file-code

:::{literalinclude} pinning/dune-project
:language: dune
:emphasize-lines: 4-6,12
:::

This will create a pin on the `fmt` package and use the specified Git repository
URL to retrieve the sources.

Don't forget to remove the version constraints from `fmt` in the list of
dependencies.

::::

The next time the package is locked, Dune will use this repository instead of
the information from the selected package repositories.

```
$ dune pkg lock
Solution for dune.lock:
- base-unix.base
- fmt.dev
- ocaml.5.0.0
- ocaml-base-compiler.5.0.0
- ocaml-config.3
- ocamlbuild.0.15.0+dune
- ocamlfind.1.9.6+dune
- topkg.1.0.7
```

Unlike previously, the version of the `fmt` library that is picked is `dev`, to
signify a development version.

The next time the project is built, the `fmt` package will be built from the
source in the specified Git repository rather than from the source tarball
released in the `opam-repository`.

```
$ dune exec ./test.exe
Hello, OCaml, Rust!
```

The Revision Store
==================

The revision store is the place where Git data that is relevant to the Dune
package management is cached.

The Concepts
------------

The revision store uses Git in the way of its original slogan, as a
content-addressable file system. A lot of data (code) and meta-data (opam files)
is stored in Git repositories that are often forked from each other, hence to
save space Dune has a Git object cache.

Git is implemented as a way to store revisions and being able to address them.
However, these revisions do not have to have a common ancestor and given Git
uses SHA1 hashes for addressing it is possible to join multiple repositories in
one single Git repository without clashing. A fairly common usecase outside of
Dune are `gh-pages` branches to serve documentation on Github, which do not
share a common parent with the main branch of the repository.

The revision store exploits this feature by putting all revisions of all
repositories into one single large repository to take advantage of caching
effects.

The Advantages
--------------

This way of organizing means that all revisions shared between multiple forks
of a Git repository can reuse the same Git objects that are in common and don't
need to download them nor store them again. Updating repositories can be done
incrementally, as Git knows which revisions are available locally and which
ones need to be fetched.

It is also possible to refer to previous states easily as the commits are part
of a Git repo and checking out an older version is as simple as checking out
the current version of the files.

Considerations and Compromises
------------------------------

An important consideration was that the management of the revision store should
be entirely transparent to the user, they should not need to do any steps to
create nor maintain it. It should get created automatically if needed and all
the steps that are necessary to keep it updated should happen in the
background. The store should always work like a cache that can be discarded
safely without causing data loss.

The revision store should always give out the most recent version of data,
unless explicitly instructed otherwise. This means that :

  * If only a Git source is specified, then the revision store will
    automatically get the newest revision
  * If the source specifies a tag or branch, then the revision store will
    automatically update to the newest revision
  * It a revision is specified, the revision store will only update if the
    revision is not yet cached, otherwise the cached version can be used

The final consideration means that an offline usage is possible if all
repositories specified are specified with their hash.

Due to the fact that the revision store is a Git repository it means that the
data sources that can be added to it also have to be available via Git. This
means that adding repositories that use different version control systems
aren't supported at the moment nor are plain HTTP sources supported.

Support for other kinds of VCSes is a possible extension by replicating similar
concepts with other version control systems, provided they allow for similar
flexibility as the Git way of storing revisions. However at the moment most
users have settled on using Git, hence this version should be able to
accommodate the needs for most users.

Another compromise is that old repositories with long histories and large sizes
have to be cloned before use, thus increasing the size of the initial download
compared to the same metadata downloaded as a compressed tarball. Despite Git
compressing objects, the history of the repositories to be added does increase
the overhead.

A solution to this could be shallow clones which only contain the latest
revisions, however these have [shown to be
problematic](https://blog.cocoapods.org/Master-Spec-Repo-Rate-Limiting-Post-Mortem/)
thus for time being we are fetching the complete histories.

Implementation
--------------

This section describes the current way the revision store is implemented.

As the revision store is not project specific, it is stored in the user's cache
directory (using the [freedesktop.org](https://www.freedesktop.org/wiki/)
specifications, the directory specified by `XDG_CACHE_HOME`), with all dune
instances sharing one single revision store.

The revision store itself is a `bare` Git repository without a worktree. This
is because all repositories in the revision store are equal and checking out
one particular revision would be a waste of disk space as the Git tooling can
be used to construct any revisions out of the bare repository anyway.

Thus every source that is added to the revision store as a remote that tracks
the default branch (or, if a branch is specified explicitly, then that
branch) and fetched, thus storing the required revisions in the revision store.

The implementation of these features is a mix of calling the `git` binary and
implementing parts in OCaml. This means that the `git` binary is required on
the system. Possible future improvements could be using
[ocaml-git](https://github.com/mirage/ocaml-git) to avoid the dependency.

# Managing Dependencies

The OCaml ecosystem has a wealth of third-party packages that are available for
use. In this section we will look into how to use them with Dune.

## Adding Dependencies

Much like in regular projects, to add a library we need to add a dependency to
it. For simplicity we will use the popular `fmt` library as an example, but any
package from the [package repository](https://ocaml.org/packages) can be used.

First we update the `dune-project` file to add a dependeny on the opam package.

::::{dropdown} `dune-project`
:icon: file-code

:::{literalinclude} dependencies/dune-project
:language: dune
:emphasize-lines: 8
:::

::::

After this change to our project dependencies, we need to relock dependencies
to update our lock directory with the new packages.

```
$ dune pkg lock
Solution for dune.lock:
- base-unix.base
- fmt.0.9.0
- ocaml.5.2.0
- ocaml-base-compiler.5.2.0
- ocaml-config.3
- ocamlbuild.0.15.0+dune
- ocamlfind.1.9.6+dune
- topkg.1.0.7
```

You can see a lot of new dependencies, among these `fmt`.

:::{note}
The list of packages being output includes all dependencies of your project,
including transitive dependencies.
:::

This will take care of installing the dependencies, but we still need to add it to
our build as a library as usual:

::::{dropdown} `dune`
:icon: file-code

:::{literalinclude} dependencies/dune
:language: dune
:emphasize-lines: 3
:::

Adding a library dependency to our `dune` file via the `libraries` stanza.

::::

This will allow us to use the `Fmt` module in our OCaml code.

::::{dropdown} `test.ml`
:icon: file-code

:::{literalinclude} dependencies/test.ml
:language: ocaml
:emphasize-lines: 4
:::

We update the code to define an `Fmt.t` pretty-printer for the list of strings
and then use it to print the value.

::::

To build it we just call `build` again.

```
$ dune build
```

which will download and install the new dependencies and build our project as
before.

As we see, the code works and uses `fmt` to do the pretty-printing:

```
$ dune exec ./test.exe
Hello, OCaml, Rust!
```

### Dependency Constraints

Packages are often only compatible with some versions of dependencies. To
specify a version range, use the regular Dune dependency syntax
used for opam dependencies in the `dune-project` file.

::::{dropdown} `dune-project`
:icon: file-code

:::{literalinclude} dependencies/constraints
:language: dune
:emphasize-lines: 7-8
:::

::::

This ensures the `fmt` package to install will be compatible with
our request. These constraints will be taken into account the next time the
package is locked:

```
$ dune pkg lock
Solution for dune.lock:
- base-unix.base
- fmt.0.9.0
- ocaml.5.2.0
- ocaml-base-compiler.5.2.0
- ocaml-config.3
- ocamlbuild.0.15.0+dune
- ocamlfind.1.9.6+dune
- topkg.1.0.7
```

The version of `fmt` picked is indeed between `0.6` and `1.0`.

## Removing Dependencies

Given all dependencies are defined in the `dune-project` file, removing a
dependency means to remove the dependency from the `depends` field of your
`dune-project` and relocking the project.

The new lock directory will not depend on the package anymore, and in future
builds, the package will not be accessible as `library` anymore.

:::{note}
The removed dependency might still be part of the lock directory if some other
dependency of your project depends on it.
:::

Unit Tests
==========

We're testing our calculator using cram tests, but in some cases, it can be more comfortable to use unit tests to test internal behaviors.

In this chapter, we're going to extract the function that deals with conversion
to floats and put it in a test harness.

## Install Dependencies

Run `opam install alcotest.1.7.0`.

## Refactor

Let's refactor our `eval_number_op` function in `lib/cli.ml` so that it uses the `as_float` function. Move `as_float` above it so that is can be used.

```{code-block} ocaml
let eval_number_op f_int f_float va vb =
  match (va, vb) with
  | VInt na, VInt nb -> VInt (f_int na nb)
  | _ ->
      let fa = as_float va in
      let fb = as_float vb in
      VFloat (f_float fa fb)
```

## Create a Test Suite

We're first going to move our cram tests so that they can be in `test/cram/`,
while our unit tests will be in `test/unit/`.

Move the contents of the folder `test` into a fresh folder `test/cram/`.

Create a folder `test/unit/`.

Create `test/unit/dune` with the following contents:

```dune
(test
 (name test_calc)
 (libraries alcotest calc))
```

And `test/unit/test_calc.ml`:

```ocaml
open Calc

let test_as_float =
  let test ~name expression ~expected =
    ( Printf.sprintf "as_float(%s)" name,
      `Quick,
      fun () ->
        let got = Cli.as_float expression in
        Alcotest.check
          (Alcotest.float Stdlib.epsilon_float)
          __LOC__ expected got )
  in
  [
    test ~name:"int" (VInt 2) ~expected:2.;
    test ~name:"float" (VFloat 3.5) ~expected:3.5;
  ]

let suite = [ ("as_float", test_as_float) ]

let () = Alcotest.run __FILE__ suite
```

Now run `dune runtest`. In addition to the existing cram tests, this also runs
unit tests.

Interfacing with C
==================

In this chapter, we're going to extend our calculator with a new function.
The difference with `sin` is that we're going to use a C implementation of the
function because it is not available in `Stdlib`. To do so, we're going to use
{doc}`(foreign_stubs) </reference/foreign-stubs>` to implement the function in
C.

## Create a test

Add a new test in `test/calc.t`:

```console
  $ calc -e 'log10(123456)'
```

Run `dune runtest` and see the failure.
Run `dune promote` to add the failure to the test file.
Our goal in the rest of the chapter is to change the output of this test.

## Lexing

We have a tiny change to make in `lib/lexer.mll`: extend function names so that
they can contain numbers (but not at the beginning).

```{code-block} ocaml
let ident = letter (letter | digit)+
```

## Evaluation

Let's extend our `eval` function in `lib/cli.ml` to handle a `log10` function.
Instead of implementing the function in OCaml, we declare it as an `external`
(with its type).

```{code-block} ocaml
:emphasize-lines: 1,12
external log10_c : float -> float = "calc_log10"

let rec eval = function
  | Ast.Int n -> VInt n
  | Float f -> VFloat f
  | Ident "pi" -> VFloat (2. *. Stdlib.acos 0.)
  | Ident _ -> failwith "unknown ident"
  | Op (Add, a, b) -> eval_number_op ( + ) ( +. ) (eval a) (eval b)
  | Op (Mul, a, b) -> eval_number_op ( * ) ( *. ) (eval a) (eval b)
  | Op (Div, a, b) -> eval_number_op ( / ) ( /. ) (eval a) (eval b)
  | Call ("sin", e) -> VFloat (Stdlib.sin (as_float (eval e)))
  | Call ("log10", e) -> VFloat (log10_c (as_float (eval e)))
  | Call _ -> failwith "unknown function"
```

## Create Foreign Stubs

The final part is to implement `calc_log10` as a C function and link it with
our library.

Let's create a file `lib/calc_stubs.c`:

:::{literalinclude} interfacing-with-c/lib/calc_stubs.c
:language: c
:::

:::{note}

The interface between C and OCaml code uses a C type called `value`.

Values of this type can be converted from and to `double` using `Double_val`
and `caml_copy_double`.

They need to be registered with the garbage collector using the `CAMLparam1`
and `CAMLreturn` macros.
:::

And we finally we specify to Dune that this file is part of the library in
`lib/dune`:

```{code-block} dune
:emphasize-lines: 4-6
(library
 (name calc)
 (libraries cmdliner)
 (foreign_stubs
  (language c)
  (names calc_stubs)))
```

Run the tests again with `dune runtest`.
At that point, the output should be correct.
Call `dune promote` to update the expected output.

## Conclusion

In this chapter, we've extended our library with some C code. The mechanism to
do so is called {doc}`foreign stubs </reference/foreign-stubs>`.

::::{dropdown} Checkpoint
:icon: location

This is how the project looks like at the end of this chapter.

:::{literalinclude} introduction/dune-project
:caption: dune-project (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/dune
:caption: bin/dune (unchanged)
:language: dune
:::

:::{literalinclude} structure/bin/calc.ml
:caption: bin/calc.ml (unchanged)
:language: ocaml
:::

:::{literalinclude} interfacing-with-c/lib/dune
:caption: lib/dune
:language: dune
:::

:::{literalinclude} development-cycle/lib/ast.ml
:caption: lib/ast.ml (unchanged)
:language: ocaml
:::

:::{literalinclude} interfacing-with-c/lib/calc_stubs.c
:caption: lib/calc_stubs.c
:language: c
:::

:::{literalinclude} interfacing-with-c/lib/cli.ml
:caption: lib/cli.ml
:language: ocaml
:::

:::{literalinclude} interfacing-with-c/lib/lexer.mll
:caption: lib/lexer.mll
:language: ocaml
:::

:::{literalinclude} development-cycle/lib/parser.mly
:caption: lib/parser.mly (unchanged)
:language: ocaml
:::

:::{literalinclude} structure/test/dune
:caption: test/dune (unchanged)
:language: dune
:::

:::{literalinclude} interfacing-with-c/test/calc.t
:caption: test/calc.t
:language: cram
:::

::::

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{clusMat}
\alias{clusMat}
\title{A clustering matrix used to demonstrate the ari-merging process.}
\format{
An object of class \code{matrix} (inherits from \code{array}) with 100 rows and 5 columns.
}
\usage{
clusMat
}
\description{
A clustering matrix used to demonstrate the ari-merging process.
}
\details{
This matrix has 100 samples with 5 cluster labels. Cluster labels 2
trought 5 are modified versions of cluster label 1, where some clusters from
label 1 where broken down into smaller clusters.
It is just a toy dataset that can be re-generated with the code in
https://github.com/HectorRDB/Pipeline_Brain/blob/master/Sandbox/createToyDataset.R
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{functionTracking}
\alias{functionTracking}
\title{Track the evolution of a function along merging}
\usage{
functionTracking(merger, f, p = 1, n_steps = NULL, ...)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{f}{the function used. It must takes as input a clustering matrix and
return a value}

\item{p}{A value between 0 and 1. We stop when the metric used for merging has
improved by p of the final total improvement. Default to 1 (i.e running the full merging).}

\item{n_steps}{Alternatively, you can specifiy the number of merging steps to
do before stopping.}

\item{...}{additional arguments passed to f}
}
\value{
a vector of length the number of merges
}
\description{
For a given ARI merging, compute the evolution on the function f
}
\examples{
# Return the number of clusters for the fourth cluster label
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
f <- function(clusMat, i) dplyr::n_distinct(clusMat[, i])
functionTracking(merger, f, i = 4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{nuclei}
\alias{nuclei}
\title{Cluster labels for a subset of the allen Smart-Seq nuclei dataset}
\format{
An object of class \code{data.frame} with 1744 rows and 7 columns.
}
\usage{
nuclei
}
\description{
Cluster labels for a subset of the allen Smart-Seq nuclei dataset
}
\details{
This matrix of clusters was obtained by running 3 clustering algorithms
on a brain snRNA-Seq dataset from Tasic et .al (https://doi.org/10.1038/s41586-018-0654-5).
This dataset was then subsetted to the GABAergic neurons.
Code to reproduce all this can be found in the github repository from the
Dune paper (https://github.com/HectorRDB/Dune_Paper).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{clusterConversion}
\alias{clusterConversion}
\title{clusterConversion}
\usage{
clusterConversion(merger, p = 1, average_n = NULL, n_steps = NULL)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{p}{A value between 0 and 1. We stop when the metric used for merging has
improved by p of the final total improvement. Default to 1 (i.e running the full merging).}

\item{average_n}{Alternatively, you can specify the average number of clusters you
want to have.}

\item{n_steps}{Finally, you can specify the number of merging steps to
do before stopping.}
}
\value{
A list containing a matrix per clustering method, with a column for the old labels
and a column for the new labels.
}
\description{
Find the conversion between the old cluster and the final clusters
}
\details{
If more than one of \code{p},\code{average_n} and \code{n_steps} is specified,
then the order of preference is \code{n_steps}, then \code{average_n} then \code{p}.
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
clusterConversion(merger)[[2]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{ARIs}
\alias{ARIs}
\title{ARI Matrix}
\usage{
ARIs(clusMat, unclustered = NULL)
}
\arguments{
\item{clusMat}{The clustering matrix with a row per cell and a column per
clustering label type}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}
}
\value{
The ARI matrix
}
\description{
ARI Matrix
}
\details{
In the ARI matrix where each cell \strong{i,j} is the adjusted Rand Index
between columns \strong{i} and \strong{j} of the original \code{clusMat}.
If \code{unclustered} is not NULL, the cells which have been assigned to the
\code{unclustered} cluster will not be counted towards computing the ARI.
}
\examples{
data("clusMat", package = "Dune")
ARIs(clusMat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{ARIImp}
\alias{ARIImp}
\title{ARI improvement}
\usage{
ARIImp(merger, unclustered = NULL)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}
}
\value{
a vector with the mean ARI between methods at each merge
}
\description{
Compute the ARI improvement over the ARI merging procedure
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
plot(0:nrow(merger$merges), ARIImp(merger))
}
\seealso{
ARItrend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/Dune.R
\name{Dune}
\alias{Dune}
\alias{Dune,matrix-method}
\alias{Dune,data.frame-method}
\alias{Dune,SummarizedExperiment-method}
\title{Dune}
\usage{
Dune(clusMat, ...)

\S4method{Dune}{matrix}(
  clusMat,
  unclustered = NULL,
  verbose = FALSE,
  parallel = FALSE,
  BPPARAM = BiocParallel::bpparam(),
  metric = "NMI"
)

\S4method{Dune}{data.frame}(
  clusMat,
  unclustered = NULL,
  verbose = FALSE,
  parallel = FALSE,
  BPPARAM = BiocParallel::bpparam(),
  metric = "NMI"
)

\S4method{Dune}{SummarizedExperiment}(
  clusMat,
  cluster_columns,
  unclustered = NULL,
  verbose = FALSE,
  parallel = FALSE,
  BPPARAM = BiocParallel::bpparam(),
  metric = "NMI"
)
}
\arguments{
\item{clusMat}{the matrix of samples by clustering labels.}

\item{...}{parameters including:}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}

\item{verbose}{Whether or not the print cluster merging as it happens.}

\item{parallel}{Logical, defaults to FALSE.
Set to TRUE if you want to parallellize the fitting.}

\item{BPPARAM}{object of class \code{bpparamClass} that specifies the
back-end to be used for computations.
See \code{bpparam} in \code{BiocParallel} package for details.
Won't be used if \code{parallel} is FALSE.}

\item{metric}{The metric that is tracked to decide which clusters to merge. For now,
either ARI and NMI are accepted. Default to NMI. See details.}

\item{cluster_columns}{if \code{clusMat} is a \code{\link{SummarizedExperiment}},
then this defines the columns of \code{colData} that are outputs from a clustering algorithm.}
}
\value{
A list with four components: the initial matrix of clustering labels,
the final matrix of clustering labels, the merge info matrix and the Metric
improvement vector.
}
\description{
Compute the Metric between every pair of clustering labels after merging every possible pair of clusters. Find the one that improves the Metric merging the most, merge the pair. Repeat until there is no improvement.
}
\details{
The Dune algorithm merges pairs of clusters in order to improve the
mean adjusted Rand Index or the mean normalized mutual information with other
clustering labels. It returns a list with five components.:
#' \itemize{
\item \code{initialMat}: The initial matrix of cluster labels
\item \code{currentMat}: The final matrix of cluster labels
\item \code{merges}: The step-by-step detail of the merges, recapitulating
which clusters where merged in which cluster label
\item \code{impMetric}: How much each merge improved the mean Metric between the
cluster label that has been merged and the other cluster labels.
\item \code{metric}: The metric that was used to find the merges.
}
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
# clusters 11 to 14 from cluster label 5 and 3 are subset of cluster 2 from
# other cluster labels. Designing cluster 2 as unclustered therefore means we
# do fewer merges.
merger2 <- Dune(clusMat = clusMat, unclustered = 2)
merger$merges
merger2$merges
}
\seealso{
clusterConversion ARIImp
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{NMIs}
\alias{NMIs}
\title{NMI Matrix}
\usage{
NMIs(clusMat, unclustered = NULL)
}
\arguments{
\item{clusMat}{The clustering matrix with a row per cell and a column per
clustering label type}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}
}
\value{
The NMI matrix
}
\description{
NMI Matrix
}
\details{
In the NMI matrix where each cell \strong{i,j} is the normalized mutual
information between columns \strong{i} and \strong{j} of the original \code{clusMat}.
If \code{unclustered} is not NULL, the cells which have been assigned to the
\code{unclustered} cluster will not be counted towards computing the NMI.
}
\examples{
data("clusMat", package = "Dune")
NMIs(clusMat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{ConfusionEvolution}
\alias{ConfusionEvolution}
\title{Plot the evolution of the ConfusionPlot as merging
happens}
\usage{
ConfusionEvolution(merger, unclustered = NULL, x, y, state_length = 1)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}

\item{x}{The name of the first cluster label to plot}

\item{y}{The name of the second cluster label to plot}

\item{state_length}{Time between steps. Default to 1. See \code{\link{transition_states}}
for details.}
}
\value{
a \code{gganim} object
}
\description{
Animated version of \code{\link{ConfusionPlot}}
}
\details{
See \code{\link{ConfusionPlot}} and \code{\link{animate}}.
}
\examples{
\dontrun{
  data("clusMat", package = "Dune")
  merger <- Dune(clusMat = clusMat)
  ConfusionEvolution(merger, x = "A", y = "B")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{ConfusionPlot}
\alias{ConfusionPlot}
\title{Plot confusion matrix}
\usage{
ConfusionPlot(x, y = NULL)
}
\arguments{
\item{x}{A vector of clustering labels or a matrix of clustering labels. See details.}

\item{y}{Optional. Another vector of clustering labels}
}
\value{
a \code{\link{ggplot}} object
}
\description{
A plot to visualize how alike two clustering labels are
}
\examples{
data("nuclei", package = "Dune")
ConfusionPlot(nuclei[, c("SC3", "Monocle")])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{NMIImp}
\alias{NMIImp}
\title{NMI improvement}
\usage{
NMIImp(merger, unclustered = NULL)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}
}
\value{
a vector with the mean NMI between methods at each merge
}
\description{
Compute the NMI improvement over the NMI merging procedure
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
plot(0:nrow(merger$merges), NMIImp(merger))
}
\seealso{
NMItrend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotPrePost}
\alias{plotPrePost}
\title{Plot the reduction in cluster size for an ARI merging with \code{Dune}}
\usage{
plotPrePost(merger)
}
\arguments{
\item{merger}{The output from an ARI merging, by calling \code{\link{Dune}}}
}
\value{
a \code{\link{ggplot}} object
#' @importFrom dplyr mutate
}
\description{
Plot the reduction in cluster size for an ARI merging with \code{Dune}
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
plotPrePost(merger)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{ARItrend}
\alias{ARItrend}
\title{ARI improvement plot}
\usage{
ARItrend(merger, unclustered = NULL)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{unclustered}{The value assigned to unclustered cells. Default to
\code{NULL}}
}
\value{
a \code{\link{ggplot}} object
}
\description{
A plot to see how ARI improves over merging
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
ARItrend(merger)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{whenToStop}
\alias{whenToStop}
\title{When to Stop}
\usage{
whenToStop(merger, p = 1, average_n = NULL)
}
\arguments{
\item{merger}{the result from having run \code{\link{Dune}}
on the dataset}

\item{p}{A value between 0 and 1. We stop when the metric used for merging has
improved by p of the final total improvement. Default to 1 (i.e running the full merging).}

\item{average_n}{Alternatively, you can specify the average number of clusters you
want to have.}
}
\value{
An integer giving the step where to stop.
}
\description{
When to Stop
}
\details{
The \code{\link{Dune}} process improves the metric. This return
the first merging step after which the metric has been improved by p of the
total. Setting p = 1 just return the number of merges.
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
whenToStop(merger, p = .5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.adjustedRandIndex}
\alias{.adjustedRandIndex}
\title{adjustedRandIndex}
\usage{
.adjustedRandIndex(tab)
}
\arguments{
\item{tab}{The confusion matrix}
}
\value{
The ARI
}
\description{
adjustedRandIndex
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotNMIs}
\alias{plotNMIs}
\title{Plot an heatmap of the NMI matrix}
\usage{
plotNMIs(clusMat, unclustered = NULL, values = TRUE, numericalLabels = FALSE)
}
\arguments{
\item{clusMat}{The clustering matrix with a row per cell and a column per
clustering label type}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}

\item{values}{Whether to also display the ARI values. Default to TRUE.}

\item{numericalLabels}{Whether labels are numerical values. Default to FALSE.}
}
\value{
a \code{\link{ggplot}} object
}
\description{
We can compute the NMI between pairs of cluster labels. This function plots
a matrix where a cell is the Normalized Mutual Information between cluster label of
row i and cluster label of column j.
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat, metric = "NMI")
plotNMIs(merger$initialMat)
plotNMIs(merger$currentMat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotARIs}
\alias{plotARIs}
\title{Plot an heatmap of the ARI matrix}
\usage{
plotARIs(clusMat, unclustered = NULL, values = TRUE, numericalLabels = FALSE)
}
\arguments{
\item{clusMat}{The clustering matrix with a row per cell and a column per
clustering label type}

\item{unclustered}{The value assigned to unclustered cells. Default to \code{NULL}}

\item{values}{Whether to also display the ARI values. Default to TRUE.}

\item{numericalLabels}{Whether labels are numerical values. Default to FALSE.}
}
\value{
a \code{\link{ggplot}} object
}
\description{
We can compute the ARI between pairs of cluster labels. This function plots
a matrix where a cell is the adjusted Rand Index between cluster label of
row i and cluster label of column j.
}
\examples{
data("clusMat", package = "Dune")
merger <- Dune(clusMat = clusMat)
plotARIs(merger$initialMat)
plotARIs(merger$currentMat)
}
